{* SaVcl 2.5.  SaDb.pas. Serge Gavrilov ( C ) 2003 - 2006 ******************* *}

unit SaDb;

{$i SaInc.inc}

{ unit desription

}
         
interface

uses
  Windows, Classes, SysUtils, Db,
{$ifdef D_67}
  Variants, FmtBcd,
{$else}
  SaAbsolute,
{$endif}  
  SaCommon, SaConsts, SaUtils, SaDbBase, SaDsBase;

type

{* ************************************************************************** *}

  TSaDb = class;
  TSaDs = class;
  TSaFlt = class;
  TSaOnWaitAction = ( waNoCancel, waCanCancel, waCancel );
  TSaOnWait = procedure(
      Sender : TSaDb;
      var Action : TSaOnWaitAction
    ) of object;
  TSaOnMessage = procedure(
      Sender : TSaDb;
      MsgType : TSaMsgType;
      Code : integer;
      Msg : string ) of object;
  TSaOnDebugMessage = procedure(
      Sender : TSaDb;
      DebMsg : string
    ) of object;

  TSaDb = class( TComponent )
  private
    oDsList : TList;
    mOnWait : TSaOnWait;
    mOnMessage : TSaOnMessage;
    mOnDebugMessage : TSaOnDebugMessage;
    mOnConnDropped : TNotifyEvent;
    mOnStart : TNotifyEvent;
    mOnFinish : TNotifyEvent;
    oDbBase : TSaDbBase;
    isStreamedActive : boolean;
    iGlInitBufCount : integer;
    procedure AddDs(
      oDsToAdd : TSaDs );
    procedure DelDs(
      oDsToDel : TSaDs );
    function Wait( isCanCancel : boolean ) : boolean;
    procedure DoOnMessage( iMsgType : TSaMsgType; iCode : integer;
      sMsg : string );
    procedure DoOnDebugMessage( const sDebMsg : string );
    procedure DoOnConnDropped;
    procedure DoOnStart;
    procedure DoOnFinish;
    procedure BeginSequence;
    procedure EndSequence;
    procedure SetActive( isVal : boolean );
    function GetActive : boolean;
    procedure SetSaSysDllPath( const sVal : string );
    function GetSaSysDllPath : string;
    procedure SetSharedName( const sVal : string );
    function GetSharedName : string;
    procedure SetLinkDllName( const sVal : string );
    function GetLinkDllName : string;
    procedure SetConParams( oVal : TSaConnectParams );
    function GetConParams : TSaConnectParams;
    procedure TransactAutoCommit;
    procedure TransactAutoRollBack;
    function GetSaVersion : TSaVersion;
    procedure SetSaVersion( iVal : TSaVersion );
    procedure SetOptions( iVal : TSaDbOptions );
    function GetOptions : TSaDbOptions;
    procedure SetTranslator( iVal : TSaTranslator );
    function GetTranslator : TSaTranslator;
    function GetIdentity : string;
  protected
    procedure Loaded; override;

  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
    function CancelRequest
        : boolean;
    procedure TransactStart;
    procedure TransactCommit;
    procedure TransactRollBack;
    procedure PutVeryLongData(
        oStream : TStream;
        isBinary : boolean;
        sTableName,
        sFieldName,
        sWhereCondition : string ); overload;
    procedure PutVeryLongData(
        s,
        sTableName,
        sFieldName,
        sWhereCondition : string ); overload;
    procedure PutVeryLongData(
        var pBuf;
        iBufLen : integer;
        sTableName,
        sFieldName,
        sWhereCondition : string ); overload;
    property Identity : string read GetIdentity;
    property LinkDllName : string read GetLinkDllName write SetLinkDllName;
  published
    property Active : boolean read GetActive write SetActive;
    property InitBufSize : integer  read iGlInitBufCount write iGlInitBufCount;
    property SaSysDllPath : string read GetSaSysDllPath write SetSaSysDllPath;
    property SharedName : string read GetSharedName write SetSharedName;
    property ConnectParams : TSaConnectParams read GetConParams
      write SetConParams;
    property Translator : TSaTranslator read GetTranslator write SetTranslator;
    property SaVersion : TSaVersion read GetSaVersion write SetSaVersion;
    property Options : TSaDbOptions read GetOptions write SetOptions;
    property OnWait : TSAOnWait read mOnWait write mOnWait;
    property OnMessage : TSaOnMessage read mOnMessage write mOnMessage;
    property OnDebugMessage : TSaOnDebugMessage read mOnDebugMessage
      write mOnDebugMessage;
    property OnConnDropped : TNotifyEvent read  mOnConnDropped
      write mOnConnDropped;
    property OnStartRequest : TNotifyEvent read mOnStart write mOnStart;
    property OnFinishRequest : TNotifyEvent read mOnFinish write mOnFinish;
  end;

{* ************************************************************************** *}
  TSaDsOpts = set of (
      soInsertInToCursor,
      soCheckReq,
      soResetValidChars,
      soAutoUnPrepare,
      soSequenced,
      soForReadOnly
    );
  TSaDsType = (
      dsQuery,
      dsTable
    );
  TSaCursorType = ( cDynScroll, cNoScroll, cInsensitive );
  TSaUpdateAction = (
      auaContinue,
      auaSkip,
      auaStop
    );
  TSaOnUpDateRecord = procedure(
      DataSet : TDataSet;
      UpdateDs : TSaDs;
      UpdateKind : TUpdateKind;
      var UpdateAction : TSaUpdateAction
    ) of object;

  TSaRecordSetType = (
      rsKeyScroll,
      rsBiScroll,
      rsScroll
    );
  TSaBaseDir = (
      abdDown,
      abdUp
    );
  TDoGlBuf = (
      dgClose,
      dgOpenFirst,
      dgOpenLast,
      dgNextDown,
      dgNextUp,
      dgGoFirst,
      dgGoLast,
      dgGlRefResh,
      dgGetAll,
      dgGoToBookMark,
      dgFindRecord,
      dgFltFindRecord,
      dgEnsureNotFirstOrLast
    );
  TDoCache = (
      dcActivate,
      dcDeActivate,
      dcClear,
      dcIsHasChanges,
      dcChangesToGlBuf,
      dcApply,
      dcUpdateRec,
      dcDeleteRec,
      dcInsertRec
    );
  TSaGlState = (
      ggsNull,
      ggsSkip,
      ggsData
    );
  TGlBuf = array[ 0..MAXWORD ] of PChar;
  TVLRecIndexes = array[ 0..MAXWORD ] of integer;
  TPVLRecIndexes = ^TVLRecIndexes;
  TPGlBuf = ^TGlBuf;
  TTitleGlobRec = record
    iOpts : integer;
    pCache : PChar;
  end;
  TPTitleGlobRec = ^TTitleGlobRec;

  TTitleLocRec = record
    iOpts : integer;
    iBookFlag : TBookmarkFlag;
    iAbsGlPos : integer;
  end;
  TPTitleLocRec = ^TTitleLocRec;

  TSaBookmark = record
    iAbsPos : integer;
    iVersion : integer;
  end;

  TSaDsTmpOpt = (
      stoOnResume,
      stoSimulateCacheHasChanges,
      stoForceExpandGlobalBuffer
    );
  TSaDsTmpOpts = set of TSaDsTmpOpt;


  TSaDs = class( TDataSet )
  private
    oSaDb : TSaDb;
    oUserParams : TList;
    iDsType : TSaDsType;
    iOpts : TSaDsOpts;
    iTmpOpts : TSaDsTmpOpts;
    isDbConnected : boolean;
    isParamsPrepared : boolean;
    sTblNames : string;
    sTblFields : string;
    sTblKeys : string;
    sTblWhere : string;
    sTblOrder : string;
    sTblNullSubs : string;
    sTblUpdNames : string;
    sTblNoUpdFields : string;
    sQuerySQL : string;
    oQuerySQLLines : TStrings;
    iCursorType : TSaCursorType;
    iRowAffected : integer;
    oExtFlt : TSaFlt;
    oSaMain : TSaDsBase;
    oParams  : TList;

    iGlDlBuf : integer;
    iGlLastFetchPos : integer;
    iGlInitBufCount : integer;
    iSaMainDir : TSaBaseDir;
    isOpenLast : boolean;

    evOnUpdateRecord : TSaOnUpDateRecord;

    isCursorIsOpen : boolean;
    (*
    isOnFilterRecord : boolean;
    *)
    isStreamedActive : boolean;
    isDoReFreshAfterPost : boolean;
    iBookMarkVersion : integer;
    isOpeningForInsert : boolean;

    pVLRecIndexes : TPVLRecIndexes;
    iVLRecIndexes : integer;
    pGlBuf : TPGlBuf;
    iGlBuf : integer;
    iGlFirstIndex : integer;
    iGlCurrentPos : integer;

    iRecSize : integer;
    iRecordCount : integer;
    isCached : boolean;
    oCacheList : TList;
    iUpdatingCacheRow : integer;
    oUpdateDs : TSaDs;
    pFldDescs : TPFldDescs;
    iFldDescs : integer;
    mOnRefresh : TNotifyEvent;
    oOrderList : TSaOrder;
    oNullSubs : TSaNullSubs;
    
    function GetModified : boolean;
    function GetGlState(
        iGlIndex : integer
      ) : TSaGlState;
    function GetGlRec(
        iGlIndex : integer
      ) : PChar;
    procedure DoAllocateRec(
        isGlobal : boolean;
        isAlloc : boolean;
        var p : PChar );
    procedure CopyRecord(
        pSource : PChar;
        pDest : PChar );
    procedure FreeVLRecs(
        pRec : PChar );

    function GetGlobTitle( p : PChar ) : TPTitleGlobRec;
    function GetLocTitle ( p : PChar ) : TPTitleLocRec;
    function DoGlBuf(
        iWhat : TDoGlBuf;
        isPar : boolean;
        iPar : integer;
        pPar : pointer
      ) : integer;
    procedure SetCached( isValue : boolean );
    function GetRecordSetType : TSaRecordSetType;
    procedure PrepareParams;
    procedure SetSaDb( oVal : TSaDb );
    procedure SetDbConnected( isVal : boolean );
    procedure SetQuerySQL( sVal : string );
    procedure SetTblNames( sVal : string );
    procedure SetTblFields( sVal : string );
    procedure SetTblKeys( sVal : string );
    procedure SetTblWhere( sVal : string );
    procedure SetTblOrder( sVal : string );
    procedure SetTblNullSubs( sVal : string );
    procedure SetTblNoUpdFields( sVal : string );
    procedure SetQuerySqlLines( oVal : TStrings );
    function GetQuerySql : string;

    procedure QuerySQLLinesChanged( Sender : TObject );

    procedure AssignNoUpdFields;
    procedure SetDsType( iVal : TSaDsType );
    procedure I_ExecSQL( isImm  : boolean );
    procedure SetCursorType( iVal : TSaCursorType );
    procedure SetParamValue( iIndex : integer; vValue : variant );
    function GetParamValue( iIndex : integer ) : variant;
    procedure FreeParams;
    function AllocateParams(
        sStatm : string
      ) : string;
    function InternalGetRecord( Buffer : PChar; GetMode : TGetMode;
      DoCheck : boolean ) : TGetResult;
    function GetBufferForData : PChar;
    function LocateToObject(
        sKeyFields : string;
        vKeyValues : variant;
        oObj : TObject;
        isFlt : boolean;
        iLocateOpts : TSaLocateOpts
      ) : pointer;
    property DbConnected : boolean read isDbConnected write SetDbConnected;
    function GetUserParam( iVal : integer ) : TSaParam;
    property Params[ iIndex : integer ] : TSaParam read GetUserParam;
    function DoCache(
        iWhat : TDoCache;
        isPar : boolean
      ) : integer;
    procedure SetNulls(
        isGlobal : boolean;
        p : PChar );
  protected
    procedure ResetBookMarks;

    function GetCanModify : boolean; override;
    procedure SetActive( isVal : boolean ); override;
    procedure SetFiltered( isVal : boolean ); override;
    procedure Loaded; override;
    procedure InternalOpen; override;
    procedure InternalClose; override;
    procedure InternalInitFieldDefs; override;
    function IsCursorOpen : boolean; override;
    function GetRecordSize : word; override;
    function AllocRecordBuffer : PChar; override;
    procedure FreeRecordBuffer( var Buffer : PChar ); override;
    function GetRecord( Buffer : PChar; GetMode : TGetMode;
      DoCheck : boolean ) : TGetResult; override;
    procedure InternalHandleException; override;
    procedure SetFieldData( Field : TField; Buffer : Pointer ); override;
    procedure InternalSetToRecord( Buffer : PChar ); override;
    procedure InternalFirst; override;
    procedure InternalLast; override;
    function GetBookmarkFlag( Buffer : PChar ) : TBookmarkFlag; override;
    procedure SetBookmarkFlag( Buffer : PChar; Value : TBookmarkFlag );override;
    procedure GetBookmarkData( Buffer : PChar; Data : Pointer ); override;
    procedure SetBookmarkData( Buffer : PChar; Data : Pointer ); override;
    procedure InternalGotoBookmark( Bookmark : Pointer ); override;
    procedure DataEvent( Event : TDataEvent; Info : Longint ); override;
    procedure InternalRefresh; override;
    procedure InternalPost; override;
    procedure DoAfterPost; override;
    procedure InternalDelete; override;
    procedure InternalAddRecord( Buffer : Pointer; Append : boolean ); override;
    procedure InternalInitRecord( Buffer : PChar ); override;
    function GetRecordCount : integer; override;
    function GetRecNo : integer; override;
{$ifndef D_67}
  {$ifndef D_5}
    function BCDToCurr(BCD: Pointer; var Curr: Currency): Boolean; override;
    function CurrToBCD(const Curr: Currency; BCD: Pointer; Precision,
      Decimals: Integer): Boolean; override;
  {$endif}
{$endif}
    property OrderList : TSaOrder read oOrderList;
  public
    constructor Create( AOwner : TComponent );  override;
    destructor Destroy; override;
    function IsExLocateAllowed : boolean;
    procedure ReOpen;
    procedure ExecuteSql;
    procedure ExecuteSqlImmediate;
    function GetParamIndex( sParamName : string ) : integer;
    procedure ApplyImmediate;
    function IsCacheHasChanges : boolean;
    procedure ApplyCachedUpdates;
    procedure ClearCache(
        isMoveUpdatesToBuf : boolean );
    procedure ApplyCache;
    procedure CancelCache;

    function BookmarkValid( Bookmark : TBookmark ) : boolean; override;
    function GetFieldData( Field  : TField; Buffer : Pointer ) : boolean;
      override;
    function CompareBookmarks( Bookmark1, Bookmark2 : TBookmark ) : integer;
      override;
    function CreateBlobStream( Field : TField; Mode : TBlobStreamMode ) :
      TStream; override;
    function IsSequenced : boolean; override;
    function GetFieldOrder( sFieldName : string ) : TSaOrderType;
    procedure FetchAll;
    procedure OpenForInsert;
    function Locate( const KeyFields: string; const KeyValues: Variant;
      Options : TLocateOptions ): boolean; override;
    function ExLocate(
        SaLocateOpts : TSaLocateOpts;
        const KeyFields : string;
        const KeyValues : Variant;
        oLocateFlt : TSaFlt = nil
      ) : boolean;
    function Resume : boolean;
    property ExtFlt : TSaFlt read oExtFlt;
    property Modified : boolean read GetModified write SetModified;
    property ParamValue[ iParamIndex : integer ] : variant
      read GetParamValue write SetParamValue;
    property RowAffected : integer read iRowAffected;
  published
    property SaDb : TSaDb read oSaDb write SetSaDb;
    property DsType : TSaDsType read iDsType write SetDsType;
    property QuerySql : string read sQuerySQL write SetQuerySql;
    property QuerySqlLines : TStrings read oQuerySqlLines
      write SetQuerySqlLines;
    property TblNames : string read sTblNames  write SetTblNames;
    property TblFields : string read sTblFields write SetTblFields ;
    property TblKeys : string read sTblKeys write SetTblKeys ;
    property TblWhere : string read sTblWhere write SetTblWhere;
    property TblOrder : string read sTblOrder write SetTblOrder;
    property TblNullSubs : string read sTblNullSubs write SetTblNullSubs;
    property TblUpdNames : string read sTblUpdNames write sTblUpdNames;
    property TblNoUpdFields : string read sTblNoUpdFields
      write SetTblNoUpdFields;
    property Options : TSaDsOpts read iOpts write iOpts;
    property CursorType : TSaCursorType read iCursorType write SetCursorType;
    property InitBufSize : integer  read iGlInitBufCount write iGlInitBufCount;
    property Cached : boolean read isCached write SetCached;
    property Active;
    property OnUpdateRecord : TSaOnUpDateRecord read evOnUpdateRecord
      write evOnUpdateRecord;
    property OnRefresh : TNotifyEvent read mOnRefresh write mOnReFresh;
    property Filtered;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
{$ifdef D_67}
    property BeforeRefresh;
    property AfterRefresh;
{$endif}    
    property OnDeleteError;
    property OnEditError;
    property OnNewRecord;
    property OnPostError;
    property OnFilterRecord;
  end;

{* ************************************************************************** *}
  TSaFltItemType = (
      aftAnd, aftOr,

      aftEq, aftEqPart, aftNotEqPart, aftEqNullSub, aftLess,
      aftEqLess, aftAbove, aftEqAbove, aftNotEq, aftString,
      aftLike
    );
  TSaFltNodeType = ( fnAnd, fnOr );
  TSaFltCondType = ( fcEq, fcLess, fcEqLess, fcAbove, fcEqAbove, fcNotEq,
    fcLike, fcString );
  TSaFltItem = class
  private
    iType : TSaFltItemType;
    oParent : TSaFltItem;
    oChild : TSaFltItem;
    oPrev : TSaFltItem;
    oNext : TSaFltItem;
    sValue : string;
    vValue : variant;
    pDesc : TPFldDesc;
    sParam : string;
  end;

  TSaFltNode = class( TSaFltItem );
  TSaFltCond = class( TSaFltItem );

  TSaFlt = class
  private
    oFirst : TSaFltItem;
    procedure FltToDsParams(
        oParamList : TList
      );
    function AddItem(
        oParent : TSaFltItem;
        iType : TSaFltItemType;
        sValue : string;
        vValue : variant
      ) : TSaFltItem;
    procedure DelItem( oItem : TSaFltItem );
    function GetFltString(
        oOrderList : TSaOrder;
        oNullSubs : TSaNullSubs;
        isInvert : boolean
      ) : string;
  public
    function CreateNode(
        oParentNode : TSaFltNode;
        iNodeType : TSaFltNodeType
      ) : TSaFltNode;
    procedure DestroyNode( oNode : TSaFltNode );
    function CreateCond(
        oParentNode : TSaFltNode;
        iCondType : TSaFltCondType;
        sCondField : string;
        vCondValue : variant
      ) : TSaFltCond;
    procedure DestroyCond( oCond : TSaFltCond );
    procedure Clear;
    function IsEmpty : boolean;
    destructor Destroy; override;
  end;

{* ************************************************************************** *}

procedure Register;

const
  SA_DEFSADSOPTS = [ soInsertInToCursor, soResetValidChars, soAutoUnPrepare ];

implementation

const
  SA_LOCATEOPT_PARTLEFT = 1;
  SA_LOCATEOPT_PARTANY = 2;
  SA_LOCATEOPT_UP = 4;
  SA_LOCATEOPT_DOWN = 8;
  SA_LOCATEOPT_NULLSTRINGS = 16;


const
  SA_RECOPT_NOCURRENTROW            =   1     ;
  SA_RECOPT_ROWUPDATED              =   2     ;
  SA_RECOPT_CACHEUPDATED            =   4     ;
  SA_RECOPT_CACHEINSERTED           =  8      ;
  SA_RECOPT_CACHEDELETED            =  16     ;
  SA_RECOPT_CACHENEEDTODELETE       =  32     ;
  SA_RECOPT_CACHED                  =  SA_RECOPT_CACHEUPDATED
                                         or  SA_RECOPT_CACHEINSERTED
                                         or  SA_RECOPT_CACHEDELETED;

  SA_MIN_BUF = 2;


{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor TSaDb.Create;
begin
  inherited;
  oDsList := TList.Create;
  oDbBase := TSaDbBase.Create;
  oDbBase.OnWait := Wait;
  oDbBase.OnMessage := DoOnMessage;
  oDbBase.OnDebugMessage := DoOnDebugMessage;
  oDbBase.OnConnDropped := DoOnConnDropped;
  oDbBase.OnStart := DoOnStart;
  oDbBase.OnFinish := DoOnFinish;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaDb.Destroy;
begin
  while oDsList.Count > 0
  do TSaDs( oDsList[ 0 ] ).SaDb := nil;
  oDbBase.Free;
  oDsList.Free;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetActive;
var
  i : integer;
begin
  if ( csReading in ComponentState )
  then  begin
        isStreamedActive := isVal;
        exit;
        end;
  if isVal = Active then exit;
  if not isVal
  then  begin
        for i := 0 to oDsList.Count - 1
        do  TSaDs( oDsList[ i ] ).DbConnected := false;
        end;
  oDbBase.Active := isVal;
  if oDbBase.Active
  then  begin
        for i := 0 to oDsList.Count - 1
        do  TSaDs( oDsList[ i ] ).DbConnected := true;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetActive;
begin
  result := oDbBase.Active;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetSaSysDllPath;
begin
  oDbBase.SaSysDllPath := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetSharedName;
begin
  result := oDbBase.SharedName;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetSharedName;
begin
  oDbBase.SharedName := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetLinkDllName;
begin
  result := oDbBase.LinkDllName;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetLinkDllName;
begin
  oDbBase.LinkDllName := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetSaSysDllPath;
begin
  result := oDbBase.SaSysDllPath;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetConParams;
begin
  oDbBase.ConnectParams := oVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetConParams;
begin
  result := oDbBase.ConnectParams;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.CancelRequest;
begin
  result := oDbBase.CancelRequest;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetSaVersion;
begin
  oDbBase.SaVersion := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetSaVersion;
begin
  result := oDbBase.SaVersion;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetOptions;
begin
  oDbBase.Options := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetOptions;
begin
  result := oDbBase.Options;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.SetTranslator;
begin
  oDbBase.Translator := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetTranslator;
begin
  result := oDbBase.Translator;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.GetIdentity;
begin
  result := oDbBase.GetIdentity;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.BeginSequence;
begin
  oDbBase.BeginSequence;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.EndSequence;
begin
  oDbBase.EndSequence;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.Loaded;
begin
  if isStreamedActive
  then  begin
        try
          Active := true;
        except
          on E : exception
          do  begin
              if not ( csDesigning in ComponentState ) then Raise;
              end;
        end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.AddDS;
begin
  oDsList.Add( oDsToAdd );
  TSaDs( oDSToAdd ).DbConnected := Active;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DelDS;
begin
  oDsList.Remove( oDsToDel );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDb.Wait;
var
  iWaitAction : TSAOnWaitAction;
begin
  result := false;
  if not Assigned( mOnWait ) then exit;
  if isCanCancel
  then  iWaitAction := waCanCancel
  else  iWaitAction := waNoCancel;
  result := iWaitAction = waCanCancel;
  mOnWait( Self, iWaitAction );
  result := result and ( iWaitAction = waCancel );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DoOnMessage;
begin
  if Assigned( mOnMessage ) then mOnMessage( Self, iMsgType, iCode, sMsg );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DoOnDebugMessage;
begin
  if Assigned( mOnDebugMessage ) then mOndebugMessage( Self, sDebMsg );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DoOnConnDropped;
begin
  if Assigned( mOnConnDropped ) then mOnConnDropped( Self );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DoOnStart;
begin
  if Assigned( mOnStart ) then mOnStart( Self );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.DoOnFinish;
begin
  if Assigned( mOnFinish ) then mOnFinish( Self );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.TransactStart;
begin
  oDbBase.TransactStart;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.TransactCommit;
begin
  oDbBase.TransactCommit;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.TransactRollBack;
begin
  oDbBase.TransactRollBack;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.TransactAutoCommit;
begin
  oDbBase.TransactAutoCommit;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.TransactAutoRollBack;
begin
  oDbBase.TransactAutoRollBack;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.PutVeryLongData(
        oStream : TStream;
        isBinary : boolean;
        sTableName : string;
        sFieldName : string;
        sWhereCondition : string );
begin
  oDbBase.PutVeryLongData( oStream, isBinary, -1, sTableName, sFieldName,
    sWhereCondition );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.PutVeryLongData(
        s : string;
        sTableName : string;
        sFieldName : string;
        sWhereCondition : string );
var
  oMemoryStream : TStream;
begin
  oMemoryStream := TMemoryStream.Create;
  try
    if s <> '' then oMemoryStream.Write( s[ 1 ], Length( s ) );
    oMemoryStream.Position := 0;
    PutVeryLongData( oMemoryStream, false, sTableName, sFieldName,
      sWhereCondition );
  finally
    oMemoryStream.Free;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDb.PutVeryLongData(
        var pBuf;
        iBufLen : integer;
        sTableName : string;
        sFieldName : string;
        sWhereCondition : string );
var
  oMemoryStream : TStream;
begin
  oMemoryStream := TMemoryStream.Create;
  try
    oMemoryStream.Write( pBuf, iBufLen );
    oMemoryStream.Position := 0;
    PutVeryLongData( oMemoryStream, true, sTableName, sFieldName,
      sWhereCondition );
  finally
    oMemoryStream.Free;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* TSaDb ( end ) ************************************************************ *}

{* TSABlobStream( begin ) ************************************************** *}
type
  TSABlobStream = class( TStream )
  private
    oField      : TBlobField;
    oDS         : TSaDs;
    iMode       : TBlobStreamMode;
    iSize       : integer;
    iPosition   : Longint;
  protected
    procedure SetSize( NewSize : longint ); override;
  public
    constructor Create( Field : TBlobField; MODe : TBlobStreamMode );
    function read( var Buffer; Count : Longint ) : Longint;         override;
    function write( const Buffer; Count : Longint ) : Longint;      override;
    function Seek ( Offset : Longint; Origin : Word ) : Longint;    override;
  end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSABlobStream.Seek( Offset : Longint; Origin : Word ) : Longint;
begin
  case Origin of
    0 : iPosition := Offset;
    1 : Inc( iPosition, Offset );
    2 : iPosition := iSize + Offset;
  end;
  result := iPosition;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSABlobStream.SetSize( NewSize : longint );
begin
  iSize := NewSize;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor TSABlobStream.Create;
var
  p : pointer;
begin
  inherited Create;
  oField := Field;
  oDS := TSaDs( oField.DataSet );
  iMode := Mode;
  if iMode = bmWrite
  then  begin
        Size := 0;
        p    := oDS.GetBufferForData;
        if p = nil then  exit;
        DoVLRec( dvlDestroy,
          @TByteArray( p^ )[ oDS.pFldDescs^[ oField.FieldNo - 1 ].
            iOffset + 1 ], 0, nil );
        TByteArray( p^ )[ oDS.pFldDescs^[ oField.FieldNo - 1 ].iOffset ] :=
          TByteArray( p^ )[ oDS.pFldDescs^[ oField.FieldNo - 1 ].iOffset ]
            or SA_SERV_ISNULL;
        end;
  if iMode = bmRead
  then  begin
        p := nil;
        Size := read( p, 0 );
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSABlobStream.read;
var
  iOfs    : integer;
  pVLRec  : TPVLRec;
  iPos    : integer;
  p       : pointer;
begin
  result := 0;
  p := oDS.GetBufferForData;
  if p = nil then    exit;
  iPos := Position;
  iOfs := oDS.pFldDescs^[ oField.FieldNo - 1 ].iOffset;
  if ( not ( ( TByteArray( p^ )[ iOfs ] and SA_SERV_VLREC ) <> 0 ) )
        or ( ( TByteArray( p^ )[ iOfs ] and SA_SERV_ISNULL ) <> 0 )
  then  exit;
  pVLRec := @TByteArray( p^ )[ iOfs + 1 ];
  if Count > 0
  then  begin
        if pVLRec^.iData - iPos < Count
        then    result := pVLRec^.iData - iPos
        else    result := Count;
        Move( TByteArray( pVLRec^.pData^ )[ iPos ], Buffer, result );
        Inc( iPosition, result );
        end
  else  result := pVLRec^.iData;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSABlobStream.write( const Buffer; Count : Longint ) : Longint;
var
    p           : pointer;
    pDesc       : TPFldDesc;
    iPos        : integer;
    pServeByte  : ^byte;
    pVLRec      : TPVLRec;
    iLen        : integer;
begin
  result := 0;
  p := oDS.GetBufferForData;
  if p = nil then    exit;
  iPos := Position;
  pDesc := @oDS.pFldDescs^[ oField.FieldNo - 1 ];
  pServeByte := @TByteArray( p^ )[ pDesc^.iOffset ];
  pVLRec := @TByteArray( p^ )[ pDesc^.iOffset + 1 ];
  result := Count;
  iLen := oDS.Fields[ pDesc^.iFldNum ].Size;
  if iLen < iPos + Count
  then  begin
        Count := iLen - iPos;
        if Count < 0 then Count := 0;
        end;

  if iPos = 0
  then  DoVLRec( dvlCreate, pVLRec, Count, @Buffer )
  else  begin
        if iPos + Count > pVLRec^.iData
        then  begin
              ReAllocMem( pVLRec^.pData, iPos + Count );
              ZeroMemory( @TByteArray( pVLRec^.pData^ )[ pVLRec^.iData ],
                          iPos + Count - pVLRec^.iData );
              pVLRec^.iData := iPos + Count;
              end;
        Move( Buffer, TByteArray( pVLRec^.pData^ )[ iPos ], Count );
        end;
  pServeByte^ := pServeByte^ and not SA_SERV_ISNULL;
  Inc( iPosition, Count );

  if result > 0
  then  oDS.DataEvent( deFieldChange, Longint( oField ) );

end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* TSABlobStream( end ) **************************************************** *}

{* TSaDs ( begin ) ********************************************************** *}

{ **************************************************************************** }
constructor TSaDs.Create;
begin
  inherited Create( AOwner );
  oOrderList := TSaOrder.Create;
  oExtFlt := TSaFlt.Create;
  oQuerySQLLines := TStringList.Create;
  TStringList( oQuerySQLLines ).OnChange := QuerySQLLinesChanged;
  oSaMain := TSaDsBase.Create;
  iOpts := SA_DEFSADSOPTS;
  BookMarkSize := sizeof( TSaBookmark );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaDs.Destroy;
begin
  Cached := false;
  oUpdateDs.Free; oUpdateDs := nil;
  SaDb := nil;
  oExtFlt.Free; oExtFlt := nil;
  oSaMain.Free; oSaMain := nil;
  oParams.Free;  oParams := nil;
  oUserParams.Free; oUserParams := nil;
  oOrderList.Free; oOrderList := nil;
  oNullSubs.Free; oNullSubs := nil;
  oQuerySQLLines.Free; oQuerySQLLines := nil;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetDbConnected;
begin
  if isVal = isDbConnected then exit;
  isDbConnected := isVal;
  if not isDbConnected
  then  begin
        try
          oSaMain.PrepStatm( '', false, 0, nil );
          Active := false;
        except
        end;
        FreeParams;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetSaDb;
begin
  if oSaDb = oVal then exit;
  DbConnected := false;
  if oSaDb <> nil then oSaDb.DelDS( Self );
  oSaDb := oVal;
  if oSaDb <> nil
  then  begin
        oSaDb.AddDS( Self );
        oSaMain.SaDbBase := oSaDb.oDbBase;
        end
  else  oSaMain.SaDbBase := nil;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function IndexOfParam( oParams : TList; sName : string ) : integer;
begin
  result := -1;
  if oParams = nil
  then    exit;
  for result := 0 to oParams.Count - 1
  do  if TSaParam( oParams[ result ] ).ParamName = sName
      then    exit;
  result := -1;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.AllocateParams;
  procedure CheckPar( sPar : string; iInd : integer; iOwn : TSaParamOwn );
  var
    i : integer;
  begin
    sPar := UpperCase( sPar );
    if oParams = nil then oParams := TList.Create;
    i := IndexOfParam( oParams, sPar );
    if ( i >= 0 )
    then  begin
          if i < iInd
          then  RSAE( eUser, 0, SA_S_ERR_DUPLICATEPARAMNAME )
          else  oParams.Exchange( i, iInd );
          end
    else  begin
          if iInd >= oParams.Count
          then  oParams.Add( TSaParam.Create( sPar, iOwn ) )
          else  oParams.Insert( iInd, TSaParam.Create( sPar, iOwn ) );
          end;
    end;
const
    PARAMNAME = [ '0'..'9', 'A'..'Z', '_', 'a'..'z' ];
var
  iPos : integer;
  iParams : integer;
  isStr : boolean;
  isParam : boolean;
  iBeg : integer;
  iCnt : integer;
  sPar : string;
  iOwn : TSaParamOwn;
  i : integer;
begin
  result := Trim( sStatm );
  iParams := 0;
  isStr := false;
  iPos := 0;
  iBeg := 1;
  iCnt := 0;
  isParam := false;
  iOwn := poUser;
  if result <> ''
  then  begin
        while iPos <= Length( result )
        do  begin
            Inc( iPos );
            if result[ iPos ] = ''''
            then  isStr := not isStr;
            if isStr then Continue;
            if isParam
            then  begin
                  isParam := result[ iPos ] in PARAMNAME;
                  if isParam
                  then  begin
                        Inc( iCnt );
                        Continue;
                        end
                  else  begin
                        sPar := Copy( result, iBeg, iCnt );
                        result[ iBeg ] := Chr( Ord( ' ' ) );
                        System.Delete( result, iBeg + 1, iCnt - 1 );
                        iPos := iBeg;
                        CheckPar( sPar, iParams, iOwn );
                        Inc( iParams );
                        end;
                  end;
            if result[ iPos ] = ':'
            then  begin
                  if ( iPos < Length( result ) )
                    and ( result[ iPos + 1 ] in PARAMNAME )
                  then  begin
                        isParam := true;
                        iBeg := iPos + 1;
                        iCnt := 0;
                        result[ iPos ] := '?';
                        if result[ iPos + 1 ] = '_'
                        then  iOwn := poInternal
                        else  iOwn := poUser;
                        end;
                  end;
            end;
        end;

  if oParams <> nil
  then  while iParams < oParams.Count
        do  begin
            TSaParam( oParams[ iParams ] ).Free;
            oParams.Delete( iParams );
            end;
  if oUserParams <> nil then oUserParams.Clear;
  if oParams <> nil
  then  for i := 0 to oParams.Count - 1
        do  begin
            if TSaParam( oParams[ i ] ).Owner = poUser
            then  begin
                  if oUserParams = nil then oUserParams := TList.Create;
                  oUserParams.Add( pointer( i ) );
                  end;
            end;
  isParamsPrepared := true;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetDsType;
begin
  if iVal = iDsType then exit;
  Active := false;
  FreeParams;
  iDsType := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.I_ExecSQL;
var
  sTmp : string;
begin
  if not isDbConnected then exit;
  Active := false;
  sTmp := GetQuerySql;
  if not isImm
  then  begin
        sTmp := AllocateParams( sTmp );
        try
          oSaMain.ExecStatm( sTmp, oParams );
          iRowAffected := oSaMain.RowAffected;
        finally
          if soAutoUnPrepare in iOpts
          then  oSaMain.PrepStatm( '', false, 0, nil );
        end;
        end
  else  oSaMain.ExecStatmImm( sTmp );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetQuerySQL;
begin
  if sVal = sQuerySQL then exit;
  Active := false;
  FreeParams;
  sQuerySQL := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetQuerySqlLines;
begin
  if ( oVal <> nil ) and ( oQuerySqlLines <> nil )
  then  oQuerySqlLines.Assign( oVal );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.QuerySQLLinesChanged;
begin
  Active := false;
  FreeParams;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetQuerySql;
begin
  result := Trim( sQuerySQL );
  if result = '' then result := oQuerySqlLines.Text;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ExecuteSQL;
begin
  I_ExecSQL( false );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ExecuteSQLImmediate;
begin
  I_ExecSQL( true );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.PrepareParams;
var
  sTmp : string;
begin
  if isParamsPrepared then exit;
  if iDsType = dsQuery
  then  sTmp := GetQuerySql
  else  sTmp := sTblWhere;
  AllocateParams( sTmp );
  isParamsPrepared := true;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblNames;
begin
  if sVal = sTblNames then exit;
  Active := false;
  sTblNames := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblFields;
begin
  if sVal = sTblFields then exit;
  Active := false;
  sTblFields := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblKeys;
begin
  if sVal = sTblKeys then exit;
  Active := false;
  sTblKeys  := sVal;
  sVal      := sTblOrder;
  sTblOrder := sTblOrder + '1';
  TblOrder  := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblWhere;
begin
  if sVal = sTblWhere then exit;
  Active := false;
  FreeParams;
  sTblWhere := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblOrder;
var
  i : integer;
begin
  if sVal = sTblOrder then exit;
  sVal := Trim( sVal );
  oOrderList.Clear;
  oOrderList.StringToSaOrder( sVal, false );
  oOrderList.StringToSaOrder( sTblKeys, true );
  if Active
  then  begin
        if csDesigning IN ComponentState
        then  begin
              Active := false;
              Active := true;
              end
        else  begin
              RefResh;
              for i := 0 to iFldDescs - 1
              do  oOrderList.SetFldNum( pFldDescs^[ i ].sFldName, i );
              end;
        end;
  sTblOrder := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblNullSubs;
begin
  sVal := Trim( sVal );
  if sVal = sTblNullSubs then exit;
  Active := false;
  sTblNullSubs := sVal;
  if oNullSubs = nil then oNullSubs := TSaNullSubs.Create;
  oNullSubs.AssignText( sTblNullSubs );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetTblNoUpdFields;
begin
  sVal := Trim( sVal );
  if sVal = sTblNoUpdFields then exit;
  sTblNoUpdFields := sVal;
  if sTblNoUpdFields <> ''
  then  Options := Options - [ soInsertInToCursor ]
  else  Options := Options + [ soInsertInToCursor ];
  if Active then AssignNoUpdFields;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.AssignNoUpdFields;
var
  oStrings : TStringList;
  i : integer;
begin
  if TblNoUpdFields <> ''
  then  begin
        oStrings := TStringList.Create;
{$ifdef D_67}
        oStrings.Delimiter := ',';
        oStrings.DelimitedText := TblNoUpdFields;
        oStrings.CaseSensitive := false;
        for i := 0 to oStrings.Count - 1
        do  oStrings[ i ] := Trim( oStrings[ i ] );
{$else}
        DelimitedTextToStrings( TblNoUpdFields, ',', oStrings );
{$endif}
        end
  else  oStrings := nil;
  try
    for i := 0 to iFldDescs - 1
    do  pFldDescs^[ i ].isNoUpdate := ( oStrings <> nil )
      and
{$ifdef D_67}
          ( oStrings.IndexOf( pFldDescs^[ i ].sFldName ) >= 0 )
{$else}
          ( NoCsIndexOf( oStrings, pFldDescs^[ i ].sFldName ) >= 0 )
{$endif}
                                                                    ;
  finally
    oStrings.Free;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetCursorType;
begin
  if iVal = iCursorType then exit;
  Active := false;
  oSaMain.PrepStatm( '', false, 0, nil );
  iCursorType := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetParamValue;
var
  oP : TSaParam;
begin
  if ( iIndex < 0 ) then exit;
  PrepareParams;
  if ( oUserParams = nil ) or ( iIndex >= oUserParams.Count ) then exit;
  oP := oParams[ integer( oUserParams[ iIndex ] ) ];
  if oP <> nil then oP.Value := vValue;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetParamValue;
var
  oP : TSaParam;
begin
  if ( iIndex < 0 ) then exit;
  PrepareParams;
  if ( oUserParams = nil ) or ( iIndex >= oUserParams.Count ) then exit;
  iIndex := integer( oUserParams[ iIndex ] );
  if ( oParams = nil )
    or ( iIndex < 0 )
    or ( iIndex >= oParams.Count )
  then exit;
  oP := oParams[ iIndex ];
  if oP <> nil then result := oP.Value;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetUserParam;
begin
  result := nil;
  PrepareParams;
  if ( iVal < 0 ) or ( iVal >= oUserParams.Count ) then exit;
  result := oParams[ integer( oUserParams[ iVal ] ) ];
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetParamIndex;
var
  oP  : TSaParam;
  i   : integer;
begin
  result := -1;
  PrepareParams;
  if oUserParams = nil then exit;
  for i := 0 to oUserParams.Count - 1
  do  begin
      oP := oParams[ integer( oUserParams[ i ] ) ];
      if CompareText( sParamName, oP.ParamName ) = 0
      then  begin
            result := i;
            exit;
            end;
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.FreeParams;
begin
  AllocateParams( '' );
  isParamsPrepared  := false;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetRecordSetType;
begin
  if ( DsType = dsTable )
    and ( CursorType = cDynScroll )
    and ( TblKeys <> '' )
  then  result := rsKeyScroll
  else  if ( CursorType = cNoScroll )
        then  result := rsScroll
        else  result := rsBiScroll;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetGlobTitle;
begin
  if p = nil
  then  result := nil
  else result := @( p[ iRecSize ] )
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetLocTitle;
begin
  if p = nil
  then  result := nil
  else  result := @( p[ iRecSize ] )
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetGlState;
begin
  if ( iGlIndex < 0 )
    or ( iGlIndex >= iGlBuf )
    or ( pGlBuf^[ iGlIndex ] = nil )
    then  result := ggsNull
  else  begin
        if ( GetGlobTitle( pGlBuf^[ iGlIndex ] )^.iOpts
          and SA_RECOPT_CACHEDELETED ) <> 0
        then  result := ggsSkip
        else  result := ggsData;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetGlRec;
begin
    if ( iGlIndex < 0 ) or ( iGlIndex >= iGlBuf )
    then    result := nil
    else    begin
            result := pGlBuf^[ iGlIndex ];
            if GetGlobTitle( result )^.pCache <> nil
            then    result := GetGlobTitle( result )^.pCache;
            end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.DoAllocateRec;
var
  i : integer;
begin
  if isAlloc
  then  begin
        if p <> nil then DoAllocateRec( isGlobal, false, p );
        if isGlobal
        then  i := sizeof( TTitleGlobRec ) + iRecSize + 1
        else  i := sizeof( TTitleLocRec  ) + iRecSize + 1;
        p := StrAlloc( i );
        ZeroMemory( p, i );
        for i := 0 to iVLRecIndexes - 1
        do  byte( p[ pFldDescs^[ pVLRecIndexes^[ i ] ].iOffset ] ) :=
                                                                SA_SERV_VLREC;

        end
  else  if p <> nil
        then  begin
              FreeVLRecs( p );
              StrDispose( p );
              p := nil;
              end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.CopyRecord;
var
  i       : integer;
  iTmp    : integer;
  pVLRecS : TPVLRec;
  pVLRecD : TPVLRec;
begin
  if ( pSource = nil ) or ( pDest = nil ) then exit;
  FreeVLRecs( pDest );
  Move( pSource^, pDest^, iRecSize );
  for i := 0 to iVLRecIndexes - 1
  do  begin
      iTmp := pFldDescs^[ pVLRecIndexes^[ i ] ].iOffset + 1;
      byte( pDest[ iTmp - 1 ] ) := byte( pSource[ iTmp - 1 ] );
      pVLRecD := @pDest  [ iTmp ];
      pVLRecS := @pSource[ iTmp ];
      ZeroMemory( pVLRecD, sizeof( pVLRecD^ ) );
      DoVLRec( dvlCreate, pVLRecD, pVLRecS^.iData, pVLRecS^.pData );
      end;

end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.FreeVLRecs;
var
  i : integer;
begin
  for i := 0 to iVLRecIndexes - 1
  do  DoVLRec ( dvlDestroy,
                @pRec[ pFldDescs^[ pVLRecIndexes^[ i ] ].iOffset + 1 ],
                0, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ResetBookMarks;
begin
  Inc( iBookMarkVersion );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.DoGlBuf;

  type
    TFillClearGlBufDir =  (
                            fgbNextDown,
                            fgbNextUp
                          );
  const
    AR_I_FILLCLEARGLBUF_DIR : array[ TFillClearGlBufDir ] of integer =( 1, -1 );
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure ClearGlBuf(
                        iWhat       :   TFillClearGlBufDir;
                        iBegPos     :   integer;
                        iCount      :   integer
                      );
  var
    i : integer;
  begin
    i := 0;
    while ( iBegPos >= 0 )
            and ( iBegPos < iGlBuf )
            and ( ( iCount < 0 ) or ( i < iCount ) )
    do  begin
        DoAllocateRec( true, false, pGlBuf^[ iBegPos ] );
        Inc( i );
        Inc( iBegPos, AR_I_FILLCLEARGLBUF_DIR[ iWhat ] );
        end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure DoAllocateGlBuf (
                              isAllocate :   boolean
                            );
  begin
    if isAllocate
    then  begin
          DoAllocateGlBuf( false );
          iGlBuf := iGlInitBufCount;
          if ( iGlBuf <= 0 ) and ( oSaDb <> nil )
          then  iGlBuf := oSaDb.iGlInitBufCount;
          if iGlBuf < SA_MIN_BUF then iGlBuf := SA_MIN_BUF;
          iGlDlBuf := iGlBuf div 2;
          GetMem( pGlBuf, sizeof( PChar ) * iGlBuf );
          ZeroMemory( pGlBuf, sizeof( PChar ) * iGlBuf );
          end
    else  if iGlBuf > 0
          then  begin
                ClearGlBuf( fgbNextDown, 0, -1 );
                FreeMem( pGlBuf, sizeof( PChar ) * iGlBuf );
                iGlBuf := 0;
                pGlBuf := nil;
                end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function GetStatement   (
                            isInvert : boolean;
                            oFlt     : TSaFlt
                         )  : string;
    function AddFilter( s : string; o : TSaFlt ) : string;
    var
      sTmp : string;
    begin
      result  := s;
      if o = nil then exit;
      sTmp    := o.GetFltString( oOrderList, oNullSubs, isInvert );
      if sTmp <> ''
      then  if result <> ''
            then  result := result + ' and ( ' + sTmp + ' ) '
            else  result := ' ( ' + sTmp + ' ) ';
    end;
  var
    sTmp, sTmp1 : string;
    iTmp : integer;
    oOI : TSAOrderItem;
  begin
    if DsType = dsQuery
    then  result := GetQuerySql
    else  begin
          result := 'SELECT ';
          if Trim( TblFields ) = ''
          then  result := result + '* '
          else  result := result + TblFields;
          if Trim( TblNames ) = ''
          then  result := result + ' FROM DUMMY'
          else  result := result + ' FROM ' + TblNames;
          sTmp := Trim( TblWhere );
          if sTmp <> ''
          then  sTmp := ' ( ' + sTmp + ' ) ';
          sTmp := AddFilter( sTmp, ExtFlt );
          if oFlt <> nil
          then  sTmp := AddFilter( sTmp, oFlt );
          if sTmp <> ''
          then  result := result + ' WHERE ' + sTmp;
          sTmp := '';
          for iTmp := 0 to oOrderList.ItemCount - 1
          do  begin
              oOI := oOrderList.Items[ iTmp ];
              if sTmp <> '' then sTmp := sTmp + ', ';
              sTmp1 := oOrderList.GetNullSubsByIndex( iTmp, oNullSubs );
              if sTmp1 <> ''
              then  sTmp := sTmp + 'ISNULL( ' + oOI.FieldName + ', ' +
                                                sTmp1 + ' )'
              else  sTmp := sTmp + oOI.FieldName;

              if ( isInvert
                    and ( not ( ofoDesc in oOI.Opts ) ) )
                or
                 ( not isInvert
                    and ( ofoDesc in oOI.Opts ) )
              then  sTmp := sTmp + ' DESC';
              end;
          if sTmp <> ''
          then  result := result + ' ORDER BY ' + sTmp;
          end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function GetCursorOpts : integer;
  const
    AR_I_CURSTYPE_TO_OPTS : array[ TSaCursorType ] of integer =
      ( SA_CURSOPT_DYNSCROLL, SA_CURSOPT_NOSCROLL, SA_CURSOPT_INSENSITIVE );
  begin
    result := AR_I_CURSTYPE_TO_OPTS[ CursorType ];
    if soForReadOnly in Options then result := result or SA_CURSOPT_FORREADONLY;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure CreateFlds( oBase : TSaDsBase );
  var
    i             : integer;
    oFldDef       : TFieldDef;
    pDesc         : TPFldDesc;
  begin
    iFldDescs := oSaMain.AllocateFldDescs( pFldDescs );
    if iFldDescs <= 0 then exit;
    AssignNoUpdFields;
    iRecSize := 0;
    FieldDefs.Clear;
    for i := 0 to iFldDescs - 1
    do  begin
        pDesc := @pFldDescs^[ i ];
        Inc( iRecSize, pDesc^.iBufLen );
        if pDesc^.isVLRec
        then  begin
              Inc( iVLRecIndexes );
              ReAllocMem( pVLRecIndexes, sizeof( integer ) * iVLRecIndexes );
              pVLRecIndexes^[ iVLRecIndexes - 1 ] := i;
              end;
        oOrderList.SetFldNum( pDesc^.sFldName, i );
        oFldDef := TFieldDef.Create( FieldDefs, pDesc^.sFldName,
                                     pDesc^.iFldType,
                                     pDesc^.iFldSize,
                                     ( soCheckReq in Options )
                                     and ( not pDesc^.isAllowNulls ),
                                     i + 1 );
        if pDesc^.iFldType in [ ftBCD {$ifdef D_67}, ftFmtBCD {$endif} ]
        then  oFldDef.Precision := pDesc^.iFldPrec;
        end;
    if DefaultFields
    then  begin
          CreateFields;
          if ( soResetValidChars in Options )
          then  for i := 0 to FieldCount - 1
                do  Fields[ i ].ValidChars := [ #0..#255 ];
          end;
    BindFields( true );
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function FillGlBuf  (
                        iWhat       :   TFillClearGlBufDir;
                        iBegPos     :   integer;
                        iCount      :   integer;
                        oSABase    :   TSaDsBase;
                        iSABaseDir :   TSaBaseDir
                      ) : integer;
  const
    AR_I_BASEDIR_TO_FETCHDIR :
                                array[ TSaBaseDir,
                                       TFillClearGlBufDir ]
    of integer = ( ( 1, -1 ), ( -1, 1 ) );
  var
    i, i2       : integer;
    iFetchRes   : TFetchResult;
  begin
    i := 0;
    result := 0;
    while ( iBegPos >= 0 )
      and ( iBegPos < iGlBuf )
      and ( ( iCount < 0 ) or ( i < iCount ) )
    do  begin
        iFetchRes := oSaMain.Fetch( true, false,
                      AR_I_BASEDIR_TO_FETCHDIR[ iSABaseDir, iWhat ] );
        iGlLastFetchPos := iBegPos;
        if ( not ( iFetchRes in [ frOK, frNoCurrentRow, frUpdateWarning ] ) )
        then  begin
              result := iBegPos + 1;
              break;
              end;
        DoAllocateRec( true, true, pGlBuf^[ iBegPos ] );
        for i2 := 0 to iFldDescs - 1
        do  oSaMain.GetSQLDABuf( true, true, i2, true,
                                @pGlBuf^[ iBegPos ][ pFldDescs^[ i2 ].iOffset ],
                                  nil, SaDb.oDbBase.TranslatorNum );
        WITH GetGlobTitle( pGlBuf^[ iBegPos ] )^
        do  begin
            if iFetchRes = frNoCurrentRow
            then  iOpts := iOpts or SA_RECOPT_NOCURRENTROW;
            if iFetchRes = frUpdateWarning
            then  iOpts := iOpts or SA_RECOPT_ROWUPDATED;
            end;
        Inc( i );
        Inc( iBegPos, AR_I_FILLCLEARGLBUF_DIR[ iWhat ] );
        end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function SetPosInBase (
                          oBase   : TSaDsBase;
                          iNewPos : integer
                        ) : boolean;
  var
    iFetchRes : TFetchResult;
  begin
    iFetchRes := oBase.Fetch( false, false, iNewPos - iGlLastFetchPos );
    result := ( iFetchRes in [ frOK, frNoCurrentRow, frUpdateWarning ] );
    if not result
    then  iGlLastFetchPos := iNewPos + SaDb.oDbBase.GetSQLCount
    else  iGlLastFetchPos := iNewPos;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function  OrderFldsToFlt  (
                              p       : PChar;
                              oFlt    : TSaFlt;
                              isNotEq : boolean
                            ) : pointer;
  var
    v           : variant;
    i1, i2      : integer;
    pDesc       : TPFldDesc;
    pOr         : pointer;
    pTmpAnd     : pointer;
    iCount      : integer;
  begin
    result := oFlt.AddItem( nil, aftAND, '', NULL );
    iCount := oOrderList.ItemCount - 1;
    if iCount >= 0
    then  v := VarArrayCreate( [ 0, 1 ], varInteger );
    if iCount > 0
    then  pOr := oFlt.AddItem( result, aftOr, '', NULL )
    else  pOr := result;
    for i1 := iCount downto 0
    do  begin
        if ( i1 > 0 )
        then  pTmpAnd := oFlt.AddItem( pOr, aftAnd, '', NULL )
        else  pTmpAnd := pOr;
        for i2 := 0 to i1
        do  begin
            pDesc := @pFldDescs^[ FieldByName( oOrderList.Items[ i2 ].
                                                FieldName ).
                                                FieldNo - 1 ];
            v[ 0 ] := integer( @p[ pDesc^.iOffset ] );
            v[ 1 ] := integer( pDesc );
            if i2 = i1
            then  begin
                  if i1 = iCount
                  then  begin
                        if isNotEq
                        then  oFlt.AddItem( pTmpAnd, aftNotEqPart,
                                            pDesc^.sFldName, v )
                        else  oFlt.AddItem( pTmpAnd, aftEqPart,
                                            pDesc^.sFldName, v );
                        end
                  else  oFlt.AddItem( pTmpAnd, aftNotEqPart,
                                      pDesc^.sFldName, v );
                  end
            else  oFlt.AddItem( pTmpAnd, aftEqNullSub, pDesc^.sFldName, v );
            end;
        end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure ExtendOrShiftGlBuf  (
                                    iWhat    : TFillClearGlBufDir;
                                    iDelta   : integer;
                                    isExtend : boolean
                                );
  var
    i : integer;
  begin
  if isExtend
  then  begin
        ReAllocMem( pGlBuf, ( iGlBuf + iDelta ) * sizeof( PChar ) );
        ZeroMemory( @pGlBuf^[ iGlBuf ], iDelta * sizeof( PChar ) );
        Inc( iGlBuf, iDelta );
        if iWhat = fgbNextUp
        then  begin
              for i := iGlBuf - 1 downto iDelta
              do  pGlBuf^[ i ] := pGlBuf^[ i - iDelta ];
              for i := 0 to iDelta - 1
              do  pGlBuf^[ i ] := nil;
              Dec ( iGlFirstIndex, iDelta );
              Inc ( iGlLastFetchPos, iDelta );
              Inc ( iGlCurrentPos, iDelta );
              end;
        end
  else  begin
        if iWhat = fgbNextDown
        then  begin
              ClearGlBuf( fgbNextUp, iDelta - 1, -1 );
              for i := 0 to iGlBuf - iDelta - 1
              do  begin
                  pGlBuf^[ i ] := pGlBuf^[ i + iDelta ];
                  pGlBuf^[ i + iDelta ] := nil;
                  end;
              Inc ( iGlFirstIndex, iDelta );
              Dec ( iGlLastFetchPos, iDelta );
              Dec ( iGlCurrentPos, iDelta );
              end
        else  begin
              ClearGlBuf( fgbNextDown, iGlBuf - iDelta, -1 );
              for i := iGlBuf - 1 downto iDelta
              do  begin
                  pGlBuf^[ i ] := pGlBuf^[ i - iDelta ];
                  pGlBuf^[ i - iDelta ] := nil;
                  end;
              Dec ( iGlFirstIndex, iDelta );
              Inc ( iGlLastFetchPos, iDelta );
              Inc ( iGlCurrentPos, iDelta );
              end;
        end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function IsExtendGlBuf
                            : boolean;
  begin
    result := ( stoForceExpandGlobalBuffer in iTmpOpts )
                or ( BufferCount > iGlBuf )
                or IsCacheHasChanges
                or ( CursorType = cNoScroll )
                or ( Assigned( OnFilterRecord ) and Filtered );
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure   OpenFromRecord  (
                                p       : PChar
                              );
  var
    oTmpFlt       : TSaFlt;
    iTmp          : integer;
    sStatement    : string;
  begin
        oTmpFlt := TSaFlt.Create;
        try
          iTmp := GetLocTitle( p )^.iAbsGlPos;
          OrderFldsToFlt  (
                            p,
                            oTmpFlt,
                            false
                          );
          DoGlBuf (
                    dgClose,
                    false,
                    0,
                    nil
                  );
          sStatement
                      := GetStatement( false,
                                       oTmpFlt );
          sStatement  := AllocateParams ( sStatement );
          if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
          oTmpFlt.FltToDSParams( oParams );
          oSaMain.OpenCurs( sStatement, GetCursorOpts, oParams );
          DoAllocateGlBuf( true );
          iGlFirstIndex := iTmp;
          iGlCurrentPos := 0;
          iSaMainDir   := abdDown;
          if FillGlBuf(
                        fgbNextDown,
                        0,
                        -1,
                        oSaMain, iSaMainDir ) = 1
          then  begin
                oSaMain.CloseCurs;
                sStatement
                      := GetStatement( true,
                                       oTmpFlt );
                sStatement  := AllocateParams ( sStatement );
                if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
                oTmpFlt.FltToDSParams( oParams );
                oSaMain.OpenCurs( sStatement, GetCursorOpts, oParams );
                iGlFirstIndex := iTmp - iGlBuf + 1;
                iGlCurrentPos := iGlBuf - 1;
                iSaMainDir   := abdUp;
                FillGlBuf( fgbNextUp, iGlBuf - 2, -1, oSaMain, iSaMainDir );
                end;
        finally
          oTmpFlt.Free;
        end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}


var
  sStatement      : string;
  iRecordSetType  : TSaRecordSetType;
  oTmpFlt         : TSaFlt;
  iTmp            : integer;
  pTmp            : pointer;
  isTmp           : boolean;
  iActAbsGlPos    : integer;
  pCharTmp        : PChar;
  oSAServ        : TSaDsBase;
  iLocateOpts : TSaLocateOpts;
begin
  result := SA_RES_OK;
  case iWhat of
  dgClose
  : begin
    DoAllocateGlBuf( false );
    DoCache( dcClear, false );
    if isPar
    then  begin
          if iFldDescs > 0
          then  begin
                for iTmp := 0 to iFldDescs - 1
                do  begin
                    pFldDescs^[ iTmp ].sFldName := '';
                    FreeMem(  pFldDescs^[ iTmp ].pAddData,
                              pFldDescs^[ iTmp ].iAddData );
                    end;
                FreeMem( pFldDescs, sizeof( TFLDDesc ) * iFldDescs );
                iFldDescs := 0;
                pFldDescs := nil;
                end;
          if iVLRecIndexes > 0
          then  begin
                FreeMem( pVLRecIndexes, iVLRecIndexes * sizeof( integer ) );
                iVLRecIndexes := 0;
                pVLRecIndexes := nil;
                end;
          BindFields( false );
          if DefaultFields then DestroyFields;
          iRecSize := 0;
          end;
    if not ( stoOnResume in iTmpOpts )
    then  begin
          oSaMain.CloseCurs;
          if isPar and ( soAutoUnPrepare in iOpts )
          then  oSaMain.PrepStatm( '', false, 0, nil );
          end;
    iGlFirstIndex   := 0;
    iGlCurrentPos   := 0;
    iGlLastFetchPos := 0;
    end;
  dgOpenFirst
  : begin
    DoGlBuf( dgClose, isPar, 0, nil );
    sStatement  := AllocateParams( GetStatement( false, nil ) );
    if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
    isTmp := ( not ( stoOnResume in iTmpOpts ) )
      and ( not Active )
      and ( soSequenced in Options )
        and ( not oSaMain.Prepared );
    if isTmp then SaDb.BeginSequence;
    try
      if stoOnResume in iTmpOpts
      then  oSaMain.OpenCursAfterResume
      else  oSaMain.OpenCurs( sStatement, GetCursorOpts, oParams );
    finally
      if isTmp then SaDb.EndSequence;
    end;
    iRecordCount := oSaMain.RecordCount;
    iSaMainDir  := abdDown;
    if isPar
    then  begin
          CreateFlds( oSaMain );
          end;
    DoAllocateGlBuf( true );
    if not isOpeningForInsert
    then  begin
          result := FillGlBuf( fgbNextDown, 1, -1, oSaMain, iSaMainDir );
          if result >= 0
          then  result := SA_RES_OK;
          end;
    iGlFirstIndex := 0;
    iGlCurrentPos := 0;
    if isOpenLast
    then  ResetBookMarks;
    isOpenLast := false;
    end;
  dgOpenLast
  : begin
    iRecordSetType := GetRecordSetType;
    DoGlBuf( dgClose, isPar, 0, nil );
    sStatement  := GetStatement( iRecordSetType = rsKeyScroll, nil );
    sStatement  := AllocateParams( sStatement );
    if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
    oSaMain.OpenCurs( sStatement, GetCursorOpts, oParams );
    if iRecordSetType = rsBiScroll
    then  begin
          oSaMain.Fetch( false, true,  -1 );
          oSaMain.Fetch( false, false,  1 );
          iSaMainDir := abdDown;
          end
    else  iSaMainDir := abdUp;
    if isPar
    then  begin
          CreateFlds( oSaMain );
          end;
    DoAllocateGlBuf( true );
    FillGlBuf( fgbNextUp, iGlBuf - 2, -1, oSaMain, iSaMainDir );
    iGlFirstIndex := abs( iRecordCount );
    iGlCurrentPos := iGlBuf - 1;
    if not isOpenLast
    then  ResetBookMarks;
    isOpenLast := true;
    end;
  dgNextDown, dgNextUp
  : begin

    if ( ( iWhat = dgNextDown ) and ( pGlBuf^[ iGlBuf - 1 ] = nil ) )
    then  begin
          result := DoGlBuf( dgGoLast, false, 0, nil );
          exit;
          end;
    if ( ( iWhat = dgNextUp ) and ( pGlBuf^[ 0 ] = nil ) )
    then  begin
          result := DoGlBuf( dgGoFirst, false, 0, nil );
          exit;
          end;


    case GetRecordSetType of
    rsBiScroll
    : begin
      if ( ( iWhat = dgNextDown )
        and ( not SetPosInBase( oSaMain, iGlBuf - 1 ) ) )
      then  begin
            ClearGlBuf( fgbNextDown, iGlLastFetchPos, -1 );
            iGlFirstIndex := iGlLastFetchPos;
            exit;
            end;
      if ( ( iWhat = dgNextUp )
        and ( not SetPosInBase( oSaMain, 0 ) ) )
      then  begin
            ClearGlBuf( fgbNextUp, iGlLastFetchPos, -1 );
            iGlFirstIndex := iGlLastFetchPos;
            exit;
            end;
      end;
    rsKeyScroll
    : begin
      if ( ( iWhat = dgNextDown ) and ( iSaMainDir = abdUp ) )
          or
         ( ( iWhat = dgNextUp ) and ( iSaMainDir = abdDown ) )
      then  begin
            oTmpFlt := TSaFlt.Create;
            try
              if iWhat = dgNextDown
              then  OrderFldsToFlt( pGlBuf^[ iGlBuf - 1 ], oTmpFlt, false )
              else  OrderFldsToFlt( pGlBuf^[ 0 ], oTmpFlt, false );
              oSaMain.CloseCurs;
              if iSaMainDir = abdDown
              then  iSaMainDir := abdUp
              else  iSaMainDir := abdDown;
              sStatement := AllocateParams( GetStatement( iSaMainDir = abdUp,
                                                          oTmpFlt ) );
              if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
              oTmpFlt.FltTodSParams( oParams );
              oSaMain.OpenCurs( sStatement, GetCursorOpts, oParams );
              oSaMain.Fetch( false, false, 1 );
            finally
              oTmpFlt.Free;
            end;
            end;
      end;
    end;
    if iWhat = dgNextDown
    then  begin
          ExtendOrShiftGlBuf ( fgbNextDown, iGlDlBuf, IsExtendGlBuf );
          result := FillGlBuf( fgbNextDown, iGlBuf - iGlDlBuf,
                               -1, oSaMain, iSaMainDir );
          end;
    if iWhat = dgNextUp
    then  begin
          ExtendOrShiftGlBuf ( fgbNextUp, iGlDlBuf, IsExtendGlBuf );
          result := FillGlBuf( fgbNextUp, iGlDlBuf - 1,
                               -1, oSaMain, iSaMainDir );
          end
    end;
  dgGoFirst
  : begin
    if ( pGlBuf^[ 0 ] = nil )
    then  begin
          iGlCurrentPos := 0;
          for iTmp := 1 to iGlBuf - 1
          do  if pGlBuf^[ iTmp ] <> nil
              then  begin
                    iGlCurrentPos := iTmp - 1;
                    break;
                    end;
          end
    else  begin
          if IsCacheHasChanges
          then  begin
                while ( DoGlBuf( dgNextUp, false, 0, nil ) = 0 )
                do  begin
                    end;
                DoGlBuf( dgGoFirst, false, 0, nil );
                end
          else  DoGlBuf( dgOpenFirst, false, 0, nil );
          end;
    end;
  dgGoLast
  : begin
    if pGlBuf^[ iGlBuf - 1 ] = nil
    then  begin
          iGlCurrentPos := iGlBuf - 1;
          for iTmp := iGlBuf - 2 downto 0
          do  if pGlBuf^[ iTmp ] <> nil
              then  begin
                    iGlCurrentPos := iTmp + 1;
                    break;
                    end;
          end
    else  begin
          if IsCacheHasChanges
            or ( not ( GetRecordSetType IN [ rsKeyScroll, rsBiScroll ] ) )
          then  begin
                while ( DoGlBuf( dgNextDown, false, 0, nil ) = 0 )
                do  ;
                DoGlBuf( dgGoLast, false, 0, nil );
                end
          else  DoGlBuf( dgOpenLast, false, 0, nil );
          end;
    end;
  dgGlReFresh
  : begin
    case GetRecordSetType of
      rsKeyScroll
      : begin
        OpenFromRecord( ActiveBuffer );
        end;

      else
        begin
        pTmp := GetBookMark;
        try
          if isOpenLast
          then  DoGlBuf ( dgOpenLast,
                          false,
                          0, nil
                        )
          else  DoGlBuf ( dgOpenFirst,
                          false,
                          0, nil
                        );
          InternalGoToBookmark( pTmp );
        finally
          FreeBookMark( pTmp );
        end;
        end;
      end;
    end;
  dgGetAll
  : begin
    if ( pGlBuf^[ iGlBuf - 1 ] = nil ) and ( pGlBuf^[ 0 ] = nil ) then exit;
    isTmp := stoSimulateCacheHasChanges in iTmpOpts;
    if not isTmp then Include( iTmpOpts, stoSimulateCacheHasChanges );
    try
      DoGlBuf( dgGoLast , false, 0, nil );
      DoGlBuf( dgGoFirst, false, 0, nil );
    finally
      if not isTmp then Exclude( iTmpOpts, stoSimulateCacheHasChanges );
    end;
    ReSync( [] );
    end;
  dgGoToBookMark
  : begin
    iGlCurrentPos := iPar - iGlFirstIndex;
    while iGlCurrentPos < 0
    do  begin
        iTmp := DoGlBuf( dgNextUp, false, 0, nil );
        if ( iGlCurrentPos < 0 ) and ( iTmp > 0 )
        then  DoGlBuf( dgGoFirst, false, 0, nil );
        end;
    while iGlCurrentPos >= iGlBuf
    do  begin
        iTmp := DoGlBuf( dgNextDown, false, 0, nil );
        if ( iGlCurrentPos >= iGlBuf ) and ( iTmp > 0 )
        then  DoGlBuf( dgGoLast, false, 0, nil );
        end;
    iActAbsGlPos := iGlCurrentPos;
    if pGlBuf^[ iGlCurrentPos ] = nil
    then  begin
          for iTmp := iGlCurrentPos - 1 downto 0
          do  if pGlBuf^[ iGlCurrentPos ] = nil
              then    Dec( iGlCurrentPos )
              else    break;
          if pGlBuf^[ iGlCurrentPos ] = nil
          then  begin
                iGlCurrentPos := iActAbsGlPos;
                for iTmp := iGlCurrentPos + 1 to iGlBuf - 1
                do  if pGlBuf^[ iGlCurrentPos ] = nil
                    then  Inc( iGlCurrentPos )
                    else    break;
                end;
          if pGlBuf^[ iGlCurrentPos ] = nil
          then  iGlCurrentPos := iActAbsGlPos;
          end;
    if iGlCurrentPos + iGlFirstIndex <> iPar
    then  result := SA_RES_NOTDATA
    else  ;
    end;
  dgFindRecord
  : begin
    result := SA_RES_NOTDATA;
    iLocateOpts := [];
    if ( iPar and SA_LOCATEOPT_NULLSTRINGS ) <> 0
    then  Include( iLocateOpts, loNullStrings );
    if ( iPar and SA_LOCATEOPT_PARTANY ) <> 0
    then  Include( iLocateOpts, loPartAny )
    else  if ( iPar and SA_LOCATEOPT_PARTLEFT ) <> 0
          then  Include( iLocateOpts, loPartLeft );

    if ( iPar and ( SA_LOCATEOPT_UP or SA_LOCATEOPT_DOWN ) ) = 0
    then  DoGlBuf( dgGoFirst, false, 0, nil );
    while true
      do  begin
          if ( iPar and SA_LOCATEOPT_UP ) <> 0
          then  begin
                Dec( iGlCurrentPos );
                if iGlCurrentPos < 0
                then  DoGlBuf( dgNextUp, false, 0, nil );
                end
          else  begin
                Inc( iGlCurrentPos );
                if iGlCurrentPos >= iGlBuf
                then  DoGlBuf( dgNextDown, false, 0, nil );
                end;
          case GetGlState( iGlCurrentPos ) of
            ggsNull : break;
            ggsData
              : if TSALocate( pPar ).CompareRec( GetGlRec( iGlCurrentPos ),
                                                 SaDb.oDbBase.CaseSensitive,
                                                 iLocateOpts )
                    or SaDb.Wait( true )
                then  begin
                      result := SA_RES_OK;
                      break;
                      end;
          end;
          end;
      ReSync( [] );
    end;
  dgFltFindRecord
  : begin
    result := SA_RES_NOTDATA;
    pCharTmp := nil;
    if ( ( iPar and ( SA_LOCATEOPT_UP or SA_LOCATEOPT_DOWN ) ) <> 0 )
          and ( ActiveBuffer <> nil )
    then  pTmp := OrderFldsToFlt( ActiveBuffer, TSaFlt( pPar ), true )
    else  pTmp := nil;
    try
      DoAllocateRec( false, true, pCharTmp );
      oSAServ := TSaDsBase.Create;
      try
        oSAServ.SaDbBase := SaDb.oDbBase;
        sStatement  := AllocateParams (
          GetStatement( ( iPar and SA_LOCATEOPT_UP ) <> 0, TSaFlt( pPar ) ) );
        if ExtFlt <> nil then ExtFlt.FltToDSParams( oParams );
        TSaFlt( pPar ).FltToDSParams( oParams );
        oSAServ.OpenCurs( sStatement, SA_CURSOPT_NOSCROLL, oParams );
          if ( oSAServ.Fetch( true, false, 1 ) in [ frOK, frUpdateWarning ] )
          then  begin
                result := SA_RES_OK;
                for iTmp := 0 to iFldDescs - 1
                do  oSAServ.GetSQLDABuf( true, true, iTmp, true,
                                        @pCharTmp[ pFldDescs^[ iTmp ].iOffset ],
                                        nil, SaDb.oDbBase.TranslatorNum );
                OpenFromRecord( pCharTmp );                                     
                ReSync( [] );
                end;
      finally
        oSAServ.Free;
        DoAllocateRec( false, false, pCharTmp );
      end;
    finally
      if pTmp <> nil then  TSaFlt( pPar ).DelItem( pTmp );
    end;
    end;
  dgEnsureNotFirstOrLast :
    begin
    result := SA_RES_OK;
    if ( iGlBuf = 0 ) or ( pGlBuf = nil ) then exit;
    if ( iGlCurrentPos = 0 ) and ( pGlBuf[ iGlCurrentPos ] <> nil )
    then  begin
          Include( iTmpOpts, stoForceExpandGlobalBuffer );
          try
            DoGlBuf( dgNextUp, false, 0, nil )
          finally
            Exclude( iTmpOpts, stoForceExpandGlobalBuffer );
          end;
          end;
    if ( iGlCurrentPos = iGlBuf - 1 ) and ( pGlBuf[ iGlCurrentPos ] <> nil )
    then  begin
          Include( iTmpOpts, stoForceExpandGlobalBuffer );
          try
            DoGlBuf( dgNextDown, false, 0, nil )
          finally
            Exclude( iTmpOpts, stoForceExpandGlobalBuffer );
          end;
          end;
    end;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.IsCacheHasChanges;
begin
  result := ( stoSimulateCacheHasChanges in iTmpOpts )
    or ( DoCache( dcIsHasChanges, false ) >0);
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetNulls;
var
  pGlobTitle  : TPTitleGlobRec;
  pLocTitle   : TPTitleLocRec;
  i           : integer;
begin
  FreeVLRecs( p );
  ZeroMemory( p, iRecSize );
  for i := 0 to iFldDescs - 1
  do  byte( p[ pFldDescs^[ i ].iOffset ] ) :=
      byte( p[ pFldDescs^[ i ].iOffset ] ) or SA_SERV_ISNULL;

  for i := 0 to iVLRecIndexes - 1
  do  byte( p[ pFldDescs^[ pVLRecIndexes^[ i ] ].iOffset ] ) :=
        SA_SERV_ISNULL or SA_SERV_VLREC;

  if isGlobal
  then  begin
        pGlobTitle := GetGlobTitle( p );
        pGlobTitle^.iOpts := 0;
        pGlobTitle^.pCache := nil;
        end
  else  begin
        pLocTitle := GetLocTitle( p );
        pLocTitle^.iOpts       := 0;
        pLocTitle^.iBookFlag   := bfCurrent;
        pLocTitle^.iAbsGlPos   := 0;
        end;

end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.DoCache;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure ClearEmptyRecs;
  var
    i       : integer;
    iShift  : integer;
  begin
    iShift := 0;
    i      := 0;
    while i + iShift < iGlBuf
    do  begin
        pGlBuf^[ i ] := pGlBuf^[ i + iShift ];
        if ( pGlBuf^[ i ] <> nil )
          and ( ( GetGlobTitle( pGlBuf^[ i ] )^.iOpts
                  and SA_RECOPT_CACHENEEDTODELETE ) <> 0 )
        then  begin
                    DoAllocateRec( true, false, pGlBuf^[ i ] );
                    Inc( iShift );
              end
        else  Inc( i );
        end;
    if iGlCurrentPos >= iGlBuf - iShift
    then  iGlCurrentPos := iGlBuf - iShift - 1;
    if iGlCurrentPos < 0
    then  iGlCurrentPos := 0;
    if iShift > 0
    then  begin
          Dec( iGlBuf, iShift );
          if iGlBuf <= 0
          then  begin
                FreeMem( pGlBuf );
                pGlBuf := nil;
                iGlBuf := 0;
                end
          else  ReAllocMem( pGlBuf, iGlBuf * sizeof( pChar ) );
          if iGlBuf < SA_MIN_BUF
          then  begin
                ReAllocMem( pGlBuf, SA_MIN_BUF * sizeof( pChar ) );
                ZeroMemory( @( pGlBuf^[ iGlBuf ] ),
                            sizeof( PChar ) * ( SA_MIN_BUF - iGlBuf ) );
                iGlBuf := SA_MIN_BUF;
                end;
          end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
var
  isHasDeleted  : boolean;
  i             : integer;
  pCacheRec      : PChar;
  pCacheTitle    : TPTitleGlobRec;
  pGlobRec      : PChar;
  pGlobTitle    : TPTitleGlobRec;
  pLocRec       : PChar;
  pLocTitle     : TPTitleLocRec;
  iAction       : TSaUpdateAction;
  iUpdateKind   : TUpdateKind;
begin
  result := SA_RES_OK;
  case iWhat of
    dcActivate
    : begin
      if oCacheList = nil
      then  oCacheList := TList.Create;
      end;
    dcDeActivate
    : begin
      if oCacheList <> nil
      then  begin
            DoCache( dcClear, isPar );
            oCacheList.Free;
            oCacheList := nil;
            end;
      end;
    dcClear
    : begin
      if oCacheList = nil then exit;
      isPar := isPar
                and IsCacheHasChanges
                and ( pGlBuf <> nil )
                and ( iGlBuf > 0 );
      isHasDeleted := false;
      if ( pGlBuf = nil ) or ( iGlBuf <= 0 )
      then  begin
            for i := 0 to oCacheList.Count - 1
            do  begin
                pCacheRec := PChar( oCacheList[ i ] );
                DoAllocateRec( true, false, pCacheRec );
                end;
            end
      else  begin
            for i := 0 to oCacheList.Count - 1
            do  begin
                pCacheRec := PChar( oCacheList[ i ] );
                if pCacheRec <> nil
                then  begin
                      pCacheTitle  := GetGlobTitle( pCacheRec );
                      pGlobRec    := pCacheTitle^.pCache;
                      pGlobTitle  := GetGlobTitle( pGlobRec );
                      if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEINSERTED ) <> 0
                      then  begin
                            pGlobTitle^.iOpts := pGlobTitle^.iOpts
                                                and not SA_RECOPT_CACHEINSERTED;
                            pGlobTitle^.iOpts := pGlobTitle^.iOpts
                                                or SA_RECOPT_CACHENEEDTODELETE;
                            isHasDeleted      := true;
                            end
                      else  pGlobTitle^.iOpts := pGlobTitle^.iOpts
                                                and not SA_RECOPT_CACHED;
                      pGlobTitle^.pCache := nil;
                      DoAllocateRec( true, false, pCacheRec );
                      end;
                end;
            end;
      oCacheList.Clear;
      if isHasDeleted then ClearEmptyRecs;
      if isPar
      then  begin
            UpdateCursorPos;
            ReSync( [] );
            end;
      end;
    dcIsHasChanges
    : if ( oCacheList <> nil ) and ( oCacheList.Count > 0 )
      then  result := 1
      else  result := 0;
    dcChangesToGlBuf
    : begin
      if oCacheList = nil then exit;
      isPar := isPar
                and IsCacheHasChanges
                and ( pGlBuf <> nil )
                and ( iGlBuf > 0 );
      isHasDeleted := false;
      if ( pGlBuf = nil ) or ( iGlBuf <= 0 )
      then  begin
            for i := 0 to oCacheList.Count - 1
            do  begin
                pCacheRec := PChar( oCacheList[ i ] );
                DoAllocateRec( true, false, pCacheRec );
                end;
            end
      else  begin
            for i := 0 to oCacheList.Count - 1
            do  begin
                pCacheRec := PChar( oCacheList[ i ] );
                if pCacheRec <> nil
                then  begin
                      pCacheTitle  := GetGlobTitle( pCacheRec );
                      pGlobRec    := pCacheTitle^.pCache;
                      pGlobTitle  := GetGlobTitle( pGlobRec );
                      if ( pGlobTitle^.iOpts
                        and ( SA_RECOPT_CACHEINSERTED
                            or SA_RECOPT_CACHEUPDATED ) ) <> 0
                      then  begin
                            CopyRecord( pCacheRec, pGlobRec );
                            pGlobTitle^.iOpts := pGlobTitle^.iOpts
                                                  and not SA_RECOPT_CACHED;
                            end
                      else  begin
                            pGlobTitle^.iOpts := ( pGlobTitle^.iOpts
                                                and not SA_RECOPT_CACHED )
                                                or SA_RECOPT_CACHENEEDTODELETE;
                            isHasDeleted      := true;
                            end;
                      pGlobTitle^.pCache := nil;
                      DoAllocateRec( true, false, pCacheRec );
                      end;
                end;
            end;
      oCacheList.Clear;
      if isHasDeleted then  ClearEmptyRecs;
      if isPar
      then  begin
            UpdateCursorPos;
            ReSync( [] );
            end;
      end;
    dcUpdateRec
    : begin
      pLocRec     := ActiveBuffer;
      pGlobRec    := pGlBuf^[ iGlCurrentPos ];
      if ( oCacheList = nil )
        or ( pLocRec = nil )
        or ( pGlobRec = nil )
      then  exit;
      pGlobTitle := GetGlobTitle( pGlobRec );
      pLocTitle  := GetLocTitle ( pLocRec );
      if  ( pLocTitle^.iAbsGlPos <> iGlFirstIndex + iGlCurrentPos )
      then    exit;
      if ( pGlobTitle^.pCache <> nil )
        and ( ( pGlobTitle^.iOpts and SA_RECOPT_CACHEDELETED ) <> 0 )
      then  exit;
      if ( pGlobTitle^.pCache <> nil )
      then  begin
            pCacheRec := pGlobTitle^.pCache;
            end
      else  begin
            pCacheRec := nil;
            DoAllocateRec( true, true, pCacheRec );
            pCacheTitle := GetGlobTitle( pCacheRec );
            pCacheTitle^.pCache := pGlobRec;
            oCacheList.Add( pCacheRec );
            pGlobTitle^.pCache    := pCacheRec;
            pGlobTitle^.iOpts := pGlobTitle^.iOpts or SA_RECOPT_CACHEUPDATED;
            end;
      CopyRecord( pLocRec, pCacheRec );
      end;
    dcDeleteRec
    : begin
      pLocRec     := ActiveBuffer;
      pGlobRec    := pGlBuf^[ iGlCurrentPos ];
      if ( oCacheList = nil )
        or ( pLocRec = nil )
        or ( pGlobRec = nil )
      then  exit;
      pGlobTitle := GetGlobTitle( pGlobRec );
      pLocTitle  := GetLocTitle ( pLocRec );
      if ( pLocTitle^.iAbsGlPos <> iGlFirstIndex + iGlCurrentPos )
      then  exit;
      if ( pGlobTitle^.pCache <> nil )
        and ( ( pGlobTitle^.iOpts and SA_RECOPT_CACHEDELETED ) <> 0 )
      then  exit;
      isHasDeleted  := false;
      if ( pGlobTitle^.pCache <> nil )
          and ( ( pGlobTitle^.iOpts
                and ( SA_RECOPT_CACHEINSERTED
                      or SA_RECOPT_CACHEUPDATED ) ) <> 0 )
      then  begin
            pCacheRec := pGlobTitle^.pCache;
            if ( ( pGlobTitle^.iOpts and SA_RECOPT_CACHEINSERTED ) <> 0 )
            then  begin
                  oCacheList.Delete( oCacheList.IndexOF( pCacheRec ) );
                  DoAllocateRec( true, false, pCacheRec );
                  isHasDeleted := true;
                  pGlobTitle^.pCache := nil;
                  pGlobTitle^.iOpts := ( pGlobTitle^.iOpts
                                          and not SA_RECOPT_CACHED )
                                          or SA_RECOPT_CACHENEEDTODELETE;
                  end
            else  begin
                  SetNulls( true, pCacheRec );
                  pCacheTitle := GetGlobTitle( pCacheRec );
                  pCacheTitle^.pCache := pGlobRec;
                  pGlobTitle^.iOpts := ( pGlobTitle^.iOpts
                                        and not SA_RECOPT_CACHED )
                                        or SA_RECOPT_CACHEDELETED;
                  end;
            end
      else  begin
            pCacheRec := nil;
            DoAllocateRec( true, true, pCacheRec );
            SetNulls( true, pCacheRec );
            pCacheTitle := GetGlobTitle( pCacheRec );
            pCacheTitle^.pCache := pGlobRec;
            oCacheList.Add( pCacheRec );
            pGlobTitle^.pCache := pCacheRec;
            pGlobTitle^.iOpts := pGlobTitle^.iOpts or SA_RECOPT_CACHEDELETED;
            end;
      if isHasDeleted then ClearEmptyRecs;
      end;
    dcInsertRec
    : begin
      pLocRec := ActiveBuffer;
      if ( oCacheList = nil )
        or ( pLocRec = nil )
      then  exit;
      Inc( iGlBuf );
      ReAllocMem( pGlBuf, iGlBuf * sizeof( PChar ) );
      pGlBuf^[ iGlBuf - 1 ] := nil;
      for i := iGlBuf - 1 downto iGlCurrentPos + 1
      do  pGlBuf^[ i ] := pGlBuf^[ i - 1 ];
      if ( iGlCurrentPos = 0 ) and ( pGlBuf^[ iGlCurrentPos ] = nil )
      then  Inc( iGlCurrentPos );
      pGlobRec := nil;
      DoAllocateRec( true, true, pGlobRec );
      pGlBuf^[ iGlCurrentPos ] := pGlobRec;
      pGlobTitle  := GetGlobTitle( pGlobRec );
      SetNulls( true, pGlobRec );
      pCacheRec    := nil;
      DoAllocateRec( true, true, pCacheRec );
      pCacheTitle := GetGlobTitle( pCacheRec );
      pCacheTitle^.pCache := pGlobRec;
      oCacheList.Add( pCacheRec );
      pGlobTitle^.pCache := pCacheRec;
      pGlobTitle^.iOpts := SA_RECOPT_CACHEINSERTED;
      CopyRecord( pLocRec, pCacheRec );
      end;
    dcApply
    : begin
      if ( oCacheList = nil ) or ( oCacheList.Count = 0 )
      then  exit;
      if oUpdateDs = nil then oUpdateDs := TSaDs.Create( nil );
      oUpdateDs.SaDb := SaDb;
      try
      try
        result := 1;
        for i := 0 to oCacheList.Count - 1
        do  begin
            iAction     := auaContinue;
            pCacheRec    := oCacheList[ i ];
            if pCacheRec = nil then    Continue;
            pGlobRec    := GetGlobTitle( pCacheRec )^.pCache;
            if pGlobRec = nil then    Continue;
            pGlobTitle  := GetGlobTitle( pGlobRec );
            iUpdateKind := ukModify;
            if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEINSERTED ) <> 0
            then  iUpdateKind := ukInsert;
            if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEDELETED ) <> 0
            then  iUpdateKind := ukDelete;
            iUpdatingCacheRow := i + 1;
            if Assigned( evOnUpdateRecord )
            then  begin
                  evOnUpdateRecord( Self, oUpdateDs, iUpdateKind, iAction );
                  end;
            if iAction = auaContinue
            then  ApplyImmediate
            else  if iAction = auaStop
                  then  begin
                        result := 0;
                        break;
                        end;
            end;
        SaDb.TransactAutoCommit;
      except
        SaDb.TransactAutoRollBack;
        raise;
      end;
      finally
        iUpdatingCacheRow := 0;
      end;
      end;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ApplyImmediate;
var
  pCacheRec        : PChar;
  pGlobRec        : PChar;
  pCacheTitle      : TPTitleGlobRec;
  pGlobTitle      : TPTitleGlobRec;
  oUI             : TList;
  oWI             : TList;
  i               : integer;
  sTmp            : string;
  sTmp1           : string;
  iSize           : integer;
  pGlobServeByte  : pointer;
  pCacheServeByte  : pointer;
  pCacheVLRec      : TPVLRec;
  pGlobVLRec      : TPVLRec;
  pGlobData       : pointer;
  pCacheData       : pointer;
  isChanged       : boolean;

  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function GetUpdNames : string;
  begin
    if Trim( TblUpdNames ) <> ''
    then  result := TblUpdNames
    else  result := TblNames;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  function GetWherePart : string;
  var
    i       : integer;
    iIndex  : integer;
  begin
    result := '';
    if ( DsType = dsQuery )
      or ( TblKeys = '' )
      or ( oOrderList.ItemCount = 0 )
    then  begin
          for i := 0 to iFldDescs - 1
          do  begin
              if pFldDescs^[ i ].isNoUpdate then Continue;
              if result <> '' then result := result + ' and ';
              if ( byte( pGlobRec[ pFldDescs^[ i ].iOffset ] )
                                                and SA_SERV_ISNULL ) <> 0
              then  result := result + pFldDescs^[ i ].sFldName + ' IS NULL'
              else  begin
                    result := result  + pFldDescs^[ i ].sFldName + ' = :W_'
                                      + pFldDescs^[ i ].sFldName;
                    oWI.Add( pointer( i ) );
                    end;
              end;
          end
    else  begin
          for i := 0 to oOrderList.ItemCount - 1
          do  begin
              if ofoKey in oOrderList.Items[ i ].Opts
              then  begin
                    iIndex := oOrderList.Items[ i ].FldNum;
                    if pFldDescs^[ iIndex ].isNoUpdate then Continue;
                    if result <> '' then result := result + ' and ';
                    if ( byte( pGlobRec[ pFldDescs^[ iIndex ].iOffset ] )
                                                and SA_SERV_ISNULL ) <> 0
                    then  result := result  + pFldDescs^[ iIndex ].sFldName
                                            + ' IS NULL'
                    else begin
                         result := result + pFldDescs^[ iIndex ].sFldName
                                    + ' = :W_' + pFldDescs^[ iIndex ].sFldName;
                         oWI.Add( pointer( iIndex ) );
                         end;
                    end;
              end;
          end;
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure SetUParams;
  var
    i : integer;
  begin
    for i := 0 to oUI.Count - 1
    do  oUpdateDs.Params[ i ].
        SetAsBuffer( @( pFldDescs^[ integer( oUI[ i ] ) ] ),
        @pCacheRec[ pFldDescs^[ integer( oUI[ i ] ) ].iOffSet ] );
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  procedure SetWParams( iBegin : integer );
  var
    i : integer;
  begin
    for i := iBegin to iBegin + oWI.Count - 1
    do  oUpdateDs.Params[ i ].
        SetAsBuffer( @( pFldDescs^[ integer( oWI[ i - iBegin ] ) ] ),
        @pGlobRec[ pFldDescs^[ integer( oWI[ i - iBegin ] ) ].iOffset ] );
  end;
  {* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

begin
  if ( ( DsType = dsQuery ) and ( TblUpdNames = '' ) )
    or ( iUpdatingCacheRow <= 0 )
    or ( oCacheList = nil )
  then  RSAE( eUser, 0, Format( SA_S_ERR_NOFUNCEXECUTING,
                                [ 'TSaDs.ApplyImmediate' ] ) );
  pCacheRec    := oCacheList[ iUpdatingCacheRow - 1 ];
  pCacheTitle  := GetGlobTitle( pCacheRec );
  pGlobRec    := pCacheTitle^.pCache;
  pGlobTitle  := GetGlobTitle( pGlobRec );

  if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEUPDATED ) <> 0
  then  begin
        oUI     := TList.Create;
        oWI     := TList.Create;
        sTmp    := '';
        try
          for i := 0 to iFldDescs - 1
          do  begin
              if pFldDescs^[ i ].isNoUpdate then Continue;
              iSize           := pFldDescs^[ i ].iBufLen;
              pGlobServeByte  := @pGlobRec[ pFldDescs^[ i ].iOffset     ];
              pGlobData       := @pGlobRec[ pFldDescs^[ i ].iOffset + 1 ];
              pCacheServeByte  := @pCacheRec[ pFldDescs^[ i ].iOffset     ];
              pCacheData       := @pCacheRec[ pFldDescs^[ i ].iOffset + 1 ];
              isChanged       := false;
              if ( byte( pGlobServeByte^ ) and SA_SERV_ISNULL )
                  <> ( byte( pCacheServeByte^ ) and SA_SERV_ISNULL )
              then  isChanged := true
              else  begin
                    if ( byte( pGlobServeByte^ ) and SA_SERV_VLREC ) = 0
                    then  isChanged := not CompareMem( pCacheData, pGlobData,
                                                       iSize - 1 )
                    else  begin
                          pCacheVLRec := pCacheData;
                          pGlobVLRec := pGlobData;
                          if ( ( pGlobVLRec =  nil ) and ( pCacheVLRec <> nil ))
                              or
                             ( ( pGlobVLRec <> nil ) and ( pCacheVLRec = nil ) )
                          then  isChanged := true
                          else  begin
                                if ( ( pGlobVLRec <> nil )
                                    and ( pCacheVLRec <> nil ) )
                                then  if pCacheVLRec^.iData <> pGlobVLRec^.iData
                                      then  isChanged := true
                                      else  isChanged :=
                                            not CompareMem( pGlobVLRec^.pData,
                                                            pCacheVLRec^.pData,
                                                    MinInt( pCacheVLRec^.iData,
                                                            pGlobVLRec^.iData));
                                end;
                          end;
                    end;
              if isChanged then  oUI.Add( pointer( i ) );
              end;
          if oUI.Count > 0
          then  begin
                sTmp := '';
                for i := 0 to oUI.Count - 1
                do  begin
                    if sTmp <> '' then    sTmp := sTmp + ', ';
                    sTmp := sTmp  + pFldDescs^[ integer( oUI[ i ] ) ].sFldName
                                  + ' = :U_'
                                  + pFldDescs^[ INteger( oUI[ i ] ) ].sFldName;
                    end;
                sTmp := 'UPDATE ' + GetUpdNames + ' SET ' + sTmp + ' WHERE ';
                sTmp := sTmp + GetWherePart;
                oUpdateDs.QuerySQL := sTmp;
                SetUParams;
                SetWParams( oUI.Count );
                oUpdateDs.ExecuteSQL;
                if oUpdateDs.iRowAffected = 0
                then  RSAE( eUser, 0, SA_S_ERR_NOTUPDATERECORD );
                if oUpdateDs.iRowAffected > 1
                then  RSAE( eUser, 0, SA_S_ERR_MULTYUPDATERECORD );
                end;
        finally
          oWI.Free;
          oUI.Free;
        end;
        end;

    if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEINSERTED ) <> 0
    then  begin
          if ( not ( soInsertInToCursor in Options ) )
          then  begin
                oUI     := TList.Create;
                oWI     := TList.Create;
                try
                  for i := 0 to iFldDescs - 1
                  do  if not pFldDescs^[ i ].isNoUpdate
                      then  oUI.Add( pointer( i ) );
                  if oUI.Count > 0
                  then  begin
                        sTmp  := '';
                        sTmp1 := '';
                        for i := 0 to oUI.Count - 1
                        do  begin
                            if sTmp <> '' then    sTmp := sTmp + ', ';
                            if sTmp1 <> '' then    sTmp1 := sTmp1 + ', ';
                            sTmp  := sTmp + pFldDescs^[ integer( oUI[ i ] ) ].
                                                                       sFldName;
                            sTmp1 := sTmp1 + ':U_'
                                    + pFldDescs^[ integer( oUI[ i ] ) ].
                                                                       sFldName;
                            end;
                        sTmp := 'INSERT INTO ' + GetUpdNames + ' ( '
                                        + sTmp + ' ) VALUES ( ' + sTmp1 + ' )';
                        oUpdateDs.QuerySQL := sTmp;
                        SetUParams;
                        oUpdateDs.ExecuteSQL;
                        end;
                finally
                  oWI.Free;
                  oUI.Free;
                end;
                end
          else  begin
                for i := 0 to iFldDescs - 1
                do  begin
                    oSaMain.SetSQLDABuf( true, i, true,
                                          @pCacheRec[ pFldDescs^[ i ].iOffset ],
                                          nil, SaDb.oDbBase.TranslatorNum );
                    end;
                oSaMain.PutRec;
                oSaMain.RestoreLength;
                for i := 0 to iFldDescs - 1
                do  begin
                    oSaMain.GetSQLDABuf( true, false, i, true,
                      @pCacheRec[ pFldDescs^[ i ].iOffset ],
                        nil, SaDb.oDbBase.TranslatorNum );
                    end;
                end;
          end;

    if ( pGlobTitle^.iOpts and SA_RECOPT_CACHEDELETED ) <> 0
    then  begin
          oWI     := TList.Create;
          try
            sTmp := 'DELETE FROM ' + GetUpdNames + ' WHERE ' + GetWherePart;
            oUpdateDs.QuerySQL := sTmp;
            SetWParams( 0 );
            oUpdateDs.ExecuteSQL;
            if oUpdateDs.iRowAffected = 0
            then  RSAE( eUser, 0, SA_S_ERR_NOTDELETERECORD );
            if oUpdateDs.iRowAffected > 1
            then  RSAE( eUser, 0, SA_S_ERR_MULTYDELETERECORD );
          finally
            oWI.Free;
          end;
          end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetCached;
begin
  if isValue = isCached then exit;
  if isCached
  then  DoCache( dcDeActivate, true )
  else  DoCache( dcActivate, false );
  isCached := isValue;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ApplyCachedUpdates;
begin
  CheckBrowseMode;
  if ( oCacheList = nil ) or ( oCacheList.Count = 0 ) then exit;
  DoCache( dcApply, false );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ClearCache;
begin
  if ( oCacheList = nil ) or ( oCacheList.Count = 0 ) then exit;
  if isMoveUpdatesToBuf
  then  DoCache( dcChangesToGlBuf, true )
  else  DoCache( dcClear, false );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ApplyCache;
begin
  CheckBrowseMode;
  if ( oCacheList = nil ) or ( oCacheList.Count = 0 ) then exit;
  ApplyCachedUpdates;
  DoCache( dcChangesToGlBuf, true );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.CancelCache;
begin
  CheckBrowseMode;
  if ( oCacheList = nil ) or ( oCacheList.Count = 0 ) then exit;
  DoCache( dcClear, true );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.InternalGetRecord;
var
  iGlState      : TSaGlState;
  pTmp          : PChar;
begin
  result := grOK;
  case GetMode of
    gmCurrent
    : begin
      iGlState := GetGlState( iGlCurrentPos );
      if iGlState in [ ggsSkip, ggsNull ]
      then  begin
            result := InternalGetRecord( Buffer, gmNext, DoCheck );
            exit;
            end;
      if Buffer <> nil
      then  begin
            pTmp := GetGlRec( iGlCurrentPos );
            CopyRecord( pTmp, Buffer );
            GetLocTitle( Buffer )^.iOpts := GetGlobTitle( pTmp )^.iOpts;
            GetLocTitle( Buffer )^.iAbsGlPos := iGlFirstIndex + iGlCurrentPos;
            SetBookmarkFlag( Buffer, bfCurrent );
            end;
      result := grOK;
      end;
    gmNext
    : begin
      while true
      do  begin
          while True
          do  begin
              if ( iGlCurrentPos = iGlBuf - 1 )
                or ( GetGlState( iGlCurrentPos + 1 ) <> ggsSkip )
              then  break
              else  Inc( iGlCurrentPos );
              end;
          if ( iGlCurrentPos = iGlBuf - 1 )
              and ( GetGlState( iGlCurrentPos ) <> ggsNull )
          then  DoGlBuf( dgNextDown, false, 0, nil )
          else  break;
          end;
      if ( GetGlState( iGlCurrentPos + 1 ) = ggsNull )
      then  begin
            SetBookmarkFlag( Buffer, bfEOF );
            result := grEOF;
            end
      else  begin
            Inc( iGlCurrentPos );
            result := InternalGetRecord( Buffer, gmCurrent, DoCheck );
            end;
      end;
    gmPrior
    : begin
      while true
      do  begin
          while True
          do  begin
              if ( iGlCurrentPos = 0 )
                or ( GetGlState( iGlCurrentPos - 1 ) <> ggsSkip )
              then  break
              else  Dec( iGlCurrentPos );
              end;
          if ( iGlCurrentPos = 0 )
              and ( GetGlState( iGlCurrentPos ) <> ggsNull )
          then  DoGlBuf( dgNextUp, false, 0, nil )
          else  break;
          end;
      if ( GetGlState( iGlCurrentPos - 1 ) = ggsNull )
      then  begin
            SetBookmarkFlag( Buffer, bfBOF );
            result := grBOF;
            end
      else  begin
            Dec( iGlCurrentPos );
            result := InternalGetRecord( Buffer, gmCurrent, DoCheck );
            end;
      end;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetBufferForData;
begin
  if State = dsFilter then result := TempBuffer
  else  begin
        if ( iUpdatingCacheRow > 0 )
        then  begin
              if State in [ dsNewValue, dsOldValue ]
              then  begin
                    if State = dsNewValue
                    then  result := oCacheList[ iUpdatingCacheRow - 1 ]
                    else  result := GetGlobTitle(
                                  oCacheList[ iUpdatingCacheRow - 1 ] )^.pCache;
                    end
              else  if IsEmpty
                    then  result := nil
                    else  result := ActiveBuffer;
              end
        else  if IsEmpty
              then  result := nil
              else  result := ActiveBuffer;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.BookmarkValid;
begin
  result := ( BookMark <> nil )
              and ( TSABookMark( BookMark^ ).iVersion = iBookmarkVersion );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalOpen;
begin
  Active := false;
  if ( SaDb = nil ) or ( not SaDb.Active )
  then  RSAE( eUser, 0, SA_S_ERR_NOTDBCONNECT );
  isCursorIsOpen := DoGlBuf( dgOpenFirst, true, 0, nil ) = SA_RES_OK;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalClose;
begin
  DoGlBuf( dgClose, true, 0, nil );
  isCursorIsOpen := false;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalInitFieldDefs;
begin
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.IsCursorOpen;
begin
  result := isCursorIsOpen;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetRecordSize;
begin
  result := iRecSize;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.AllocRecordBuffer;
begin
  result := nil;
  DoAllocateRec( false, true, result );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.FreeRecordBuffer;
begin
  DoAllocateRec( false, false, Buffer );
  Buffer := nil;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetRecord;
var
  isTmp : boolean;
  iOldState : TDataSetState;
begin
  result := grOK;
  while true
  do  begin
      result := InternalGetRecord( Buffer, GetMode, DoCheck );
      if ( result = grOK ) and Filtered and Assigned( OnFilterRecord ) 
      then  begin
            isTmp := true;
            iOldState := SetTempState( dsFilter );
            try
              OnFilterRecord( Self, isTmp );
            finally
              RestoreState( iOldState );
            end;
            if not isTmp
            then  begin
                  if GetMode = gmCurrent
                  then  GetMode := gmNext;
                  end
            else  break;
            end
      else  break;
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalHandleException;
begin
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetFieldData( Field : TField; Buffer : Pointer );
var
  i           : integer;
  iL          : integer;
  pBuf        : PChar;
  p           : pointer;
  isDataCh    : boolean;
  rcTimeStamp : TTimeStamp;
begin
  pBuf := GetBufferForData;
  if pBuf = nil then    Exit;
  i := Field.FieldNo - 1;
  if i < 0
  then  RSAE( eUser, 0, Format( SA_S_ERR_NOTSUPPORT,
    [ 'calculated and lookup fields' ] ) );
  p := @pBuf[ pFldDescs^[ i ].iOffSet ];
  if Buffer = nil
  then  begin
        isDataCh := ( byte( p^ ) and SA_SERV_ISNULL ) = 0;
        byte( p^ ) := byte( p^ ) or SA_SERV_ISNULL;
        end
  else  begin
        isDataCh := ( byte( p^ ) and SA_SERV_ISNULL ) <> 0;
        byte( p^ ) := byte( p^ ) and not SA_SERV_ISNULL;
        p := @TByteArray( p^ )[ 1 ];
        case pFldDescs^[ i ].iFldType of
          ftSmallInt
          : begin
            isDataCh := isDataCh or ( smallint( p^ ) <> smallint ( Buffer^ ) );
            smallint( p^ )  := smallint ( Buffer^ );
            end;
          ftInteger, ftAutoInc
          : begin
            isDataCh := isDataCh or ( integer ( p^ ) <> integer  ( Buffer^ ) );
            integer ( p^ )  := integer  ( Buffer^ );
            end;
          ftTime
          : begin
            isDataCh := isDataCh
             or ( longint( Buffer^ ) <> Frac( TDateTime( p^ ) ) * MSecsPerDay );
            TDateTime( p^ ) := longint( Buffer^ ) / MSecsPerDay;
            end;
          ftDate
          : begin
            isDataCh := isDataCh
             or ( longint( Buffer^ ) - DateDelta <> Int( TDateTime( p^ ) ) );
            TDateTime( p^ ) := longint( Buffer^ ) - DateDelta;
            end;
          ftDateTime
          : begin
            rcTimeStamp := MSecsToTimeStamp( TDateTimeRec( Buffer^ ).DateTime );
            isDataCh := isDataCh
              or ( rcTimeStamp.Time <>
                    DateTimeToTimeStamp( TDateTime( p^ ) ).Time )
              or ( rcTimeStamp.Date <>
                    DateTimeToTimeStamp( TDateTime( p^ ) ).Date );
            TDateTime( p^ ) := TimeStampToDateTime( rcTimeStamp );
            end;
          ftString
          : begin
            iL := byte( p^ );
            byte( p^ ) := Length( PChar( Buffer ) );
            isDataCh := isDataCh
              or ( iL <> byte( p^ ) )
              or ( not CompareMem( @TByteArray( p^ )[ 1 ],
                        Buffer, byte( p^ ) ) );
            Move( Buffer^, TByteArray( p^ )[ 1 ], byte( p^ ) );
            end;
          ftFloat
          : begin
            isDataCh := isDataCh or ( double( Buffer^ ) <> double( p^ ) );
            double( p^ ) := double( Buffer^ );
            end;
{$ifdef D_67}
          ftFmtBCD,
{$endif}
          ftBCD
          : begin
{$ifdef D_67}
            isDataCh := isDataCh
                        or ( BcdCompare( TBcd( p^ ), TBcd( Buffer^ ) ) <> 0 );
            TBcd( p^ ) := TBcd( Buffer^ );
{$else}
  {$ifdef D_5}
            isDataCh := isDataCh
              or ( not CompareMem( p, Buffer, sizeof( TBcd ) ) );
            TBcd( p^ ) := TBcd( Buffer^ );
  {$else}
            isDataCh := isDataCh
                        or ( currency( p^ ) <> currency( Buffer^ ) );
            currency( p^ ) := currency( Buffer^ );
  {$endif}
{$endif}
            end;
          ftLargeInt
          : begin
            isDataCh := isDataCh or ( Int64( p^ ) <> Int64( Buffer^ ) );
            Int64( p^ ) := Int64( Buffer^ );
            end;
          ftBoolean
          : begin
            isDataCh := isDataCh
              or ( WordBool( Buffer^ ) <> ( shortint( p^ ) <> 0 ) );
            if WordBool( Buffer^ )
            then  shortint( p^ ) := 1
            else  shortint( p^ ) := 0;
            end;
          ftWord
          : begin
            isDataCh := isDataCh or ( Word( p^ ) <> Word( Buffer^ ) );
            Word( p^ ) := Word( Buffer^ );
            end;
          ftVarBytes
          : begin
            iL := byte( p^ );
            if Word( Buffer^ ) > 255
            then  byte( p^ ) := 255
            else  byte( p^ ) := Word( Buffer^ );
            isDataCh := isDataCh
              or ( iL <> byte( p^ ) )
              or ( not CompareMem( @TByteArray( p^ )[ 1 ],
                @TByteArray( Buffer^ )[ 2 ],
                                   byte( p^ ) ) );
            Move( TByteArray( Buffer^ )[ 2 ],
              TByteArray( p^ )[ 1 ], byte( p^ ) );
            end;
        end;
        end;
  if isDataCh then DataEvent( deFieldChange, Longint( Field ) );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetFieldData    (
                                        Field  : TField;
                                        Buffer : Pointer
                                  ) : boolean;
var
    i       : integer;
    pBuf    : PChar;
    p       : pointer;
begin
  result := false;
  pBuf := GetBufferForData;
  if pBuf = nil then exit;
  i := Field.FieldNo - 1;
  if i < 0
  then  RSAE( eUser, 0, Format( SA_S_ERR_NOTSUPPORT,
    [ 'calculated and lookup fields' ] ) );
  p := @pBuf[ pFldDescs^[ i ].iOffSet ];
  result := ( byte( p^ ) and SA_SERV_ISNULL ) = 0;
  if ( not result ) or ( Buffer = nil ) then    exit;
  p := @TByteArray( p^ )[ 1 ];
  if ( i >= 0 )
  then  begin
        case pFldDescs^[ i ].iFldType of
          ftSmallInt
            : smallint ( Buffer^ ) := smallint( p^ );
          ftInteger, ftAutoInc
            : integer  ( Buffer^ ) := integer ( p^ );
          ftTime
            : longint( Buffer^ ) := DateTimeToTimeStamp( TDateTime( p^ ) ).Time;
          ftDate
            : longint( Buffer^ ) := DateTimeToTimeStamp( TDateTime( p^ ) ).Date;
          ftDateTime
            : TDateTimeRec( Buffer^ ).DateTime :=
              TimeStampToMSecs( DateTimeToTimeStamp( TDateTime( p^ ) ) );
          ftString
            : begin
              Move( TByteArray( p^ )[ 1 ], Buffer^, byte( p^ ) );
              TByteArray( Buffer^ )[ byte( p^ ) ] := 0;
              end;
          ftFloat
          : double( Buffer^ )   :=  double( p^ );
{$ifdef D_67}
          ftFmtBCD,
{$endif}
          ftBCD
{$ifdef D_67}
          : TBcd( Buffer^ ) := TBcd( p^ );
{$else}
  {$ifdef D_5}
          : TBcd( Buffer^ ) := TBcd( p^ );
  {$else}
          : currency( Buffer^ ) := currency( p^ );
  {$endif}
{$endif}
          ftLargeInt
            : Int64( Buffer^ ) := Int64( p^ );
          ftBoolean
            : WordBool( Buffer^ ) := shortint( p^ ) <> 0;
          ftWord
            : Word( Buffer^ ) := word( p^ );
          ftVarBytes
            : begin
              Word( Buffer^ ) := byte( p^ );
              Move( TByteArray( p^ )[ 1 ],
                TByteArray( Buffer^ )[ 2 ], byte( p^ ) );
              end;
        else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                    [ FieldTypeNames[ pFldDescs^[ i ].iFldType ] ] ) );
        end;
        end
  else  result := false;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalSetToRecord;
begin
  DoGlBuf( dgGoToBookMark, false, GetLocTitle( Buffer )^.iAbsGlPos, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalFirst;
begin
  DoGlBuf( dgGoFirst, false, 0, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalLast;
begin
  DoGlBuf( dgGoLast, false, 0, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetBookmarkFlag;
begin
  result := bfCurrent;
  if Buffer <> nil
  then  result := GetLocTitle( Buffer )^.iBookFlag;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetBookmarkFlag;
begin
  if Buffer <> nil then GetLocTitle( Buffer )^.iBookFlag := Value;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.GetBookmarkData;
begin
  TSaBookmark( Data^ ).iAbsPos  := GetLocTitle( Buffer )^.iAbsGlPos;
  TSaBookmark( Data^ ).iVersion := iBookMarkVersion;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetBookmarkData;
begin
  GetLocTitle( Buffer )^.iAbsGlPos := TSaBookmark( Data^ ).iAbsPos;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalGotoBookmark;
begin
  if Bookmark <> nil
  then  begin
        DoGlBuf( dgGoToBookMark, false, TSaBookmark( Bookmark^ ).iAbsPos, nil );
//        ReSync( [] );
        end;

end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function  TSaDs.CompareBookmarks;
begin
  if ( BookMark1 = nil ) or ( BookMark2 = nil )
  then  result := 0
  else  result := TSaBookmark( BookMark1^ ).iAbsPos -
                    TSaBookmark( BookMark2^ ).iAbsPos;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.CreateBlobStream;
begin
  result := TSABlobStream.Create( Field AS TBlobField, Mode );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.DataEvent;
begin
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.IsExLocateAllowed;
begin
  result := ( GetRecordSetType = rsKeyScroll )
    and ( not IsCacheHasChanges )
    and ( not ( Assigned( OnFilterRecord ) and Filtered ) );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.Locate;
begin
  if loPartialKey in Options
  then  result := ExLocate( [ loPartLeft, loNullStrings ], KeyFields, KeyValues,
    nil )
  else  result := ExLocate( [ loNullStrings ], KeyFields, KeyValues, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.ExLocate;

  function SALocateOptsToInt( SaLocateOpts : TSaLocateOpts ) : integer;
  begin
    result := 0;
    if loPartAny in SaLocateOpts
    then  result := result or SA_LOCATEOPT_PARTANY
    else  if loPartLeft in SaLocateOpts
          then  result := result or SA_LOCATEOPT_PARTLEFT;
    if loUp in SaLocateOpts
    then  result := result or SA_LOCATEOPT_UP
    else  if loDown in SaLocateOpts
          then  result := result or SA_LOCATEOPT_DOWN;
    if loNullStrings in SaLocateOpts
    then  result := result or SA_LOCATEOPT_NULLSTRINGS;
  end;
var
  pLocateNode   : pointer;
  isExFlt       : boolean;
  oLocate : TSALocate;
begin
  result := false;
  if not Active then exit;
  if ( oLocateFlt <> nil ) and not IsExLocateAllowed
  then  RSAE( eUser, 0, Format( SA_S_ERR_NOFUNCEXECUTING,
                        [ 'TSaDs.ExLocate' ] ) );
  if IsExLocateAllowed
  then  begin
        isExFlt := oLocateFlt <> nil;
        if not isExFlt then oLocateFlt := TSaFlt.Create;
        try
          pLocateNode := LocateToObject( KeyFields, KeyValues, oLocateFlt,
                                         true, SaLocateOpts );
          try
            DisableControls;
            try
              result := DoGlBuf( dgFltFindRecord, false,
                                 SALocateOptsToInt( SaLocateOpts ),
                                 oLocateFlt ) = SA_RES_OK;
            finally
              EnableControls;
            end;
          finally
            oLocateFlt.DelItem( pLocateNode );
          end;
        finally
          if not isExFlt then oLocateFlt.Free;
        end;
        end
  else  begin
        oLocate := TSALocate.Create;
        try
          LocateToObject( KeyFields, KeyValues, oLocate, false, SaLocateOpts );
          result := DoGlBuf( dgFindRecord, false,
                      SALocateOptsToInt( SaLocateOpts ),
                      oLocate ) = SA_RES_OK;
        finally
          oLocate.Free;
        end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.LocateToObject;
  function  ExtractSepString( const sSource : STRING;
                              const cSep    : char;
                              var   iPos    : Integer ) : STRING;
  var
    i : integer;
  begin
    i := iPos;
    while ( i <= Length( sSource ) ) and ( sSource[ i ] <> cSep ) do Inc( i );
    result := Trim( Copy( sSource, iPos, i - iPos ) );
    if ( i <= Length( sSource ) ) and ( sSource[ i ] = cSep ) then Inc( i );
    iPos := i;
  end;
var
  iPos    : integer;
  iCnt    : integer;
  iFldNum : integer;
  sFld    : string;
  i       : integer;
  vVal    : variant;
  pAddNode : pointer;
begin
  iPos := 1;
  iCnt := 0;
  sFld := ExtractSepString( sKeyFields, ';', iPos );
  if isFlt
  then  result := TSaFlt( oObj ).AddItem( nil, aftAnd, '', NULL )
  else  result := nil;
  while sFld <> ''
  do  begin
      iFldNum := -1;
      for i := 0 to iFldDescs - 1
      do  if CompareText( sFld, pFldDescs^[ i ].sFldName ) = 0
          then  begin
                iFldNum := i;
                break;
                end;
      if iFldNum >= 0
      then  begin
            if ( VarType( vKeyValues ) and varArray ) <> 0
            then  begin
                  if iCnt <= VarArrayHighBound( vKeyValues, 1 ) -
                             VarArrayLowBound( vKeyValues, 1 )
                  then    vVal := vKeyValues[ iCnt ]
                  else    vVal := NULL;
                  end
            else  vVal := vKeyValues;
            if isFlt
            then  begin
                  with pFldDescs^[ iFldNum ], TSaFlt( oObj )
                  do  begin
                      if ( FieldGroup( iFldType ) = fgString )
                        and ( ( [ loPartLeft, loPartAny, loNullStrings ]
                              * iLocateOpts ) <> [ ] )
                      then  begin
                            if ( loNullStrings in iLocateOpts )
                                and ( not VarIsNull( vVal ) )
                                and ( not VarIsEmpty( vVal ) )
                                and ( GSVVarToStr( vVal ) = '' )
                            then  begin
                                  pAddNode := AddItem( result, aftOr, '',NULL );
                                  AddItem( pAddNode, aftEq, sFld, NULL );
                                  end
                            else  pAddNode := result;
                            if loPartAny in iLocateOpts
                            then  AddItem( pAddNode, aftLike, sFld,
                                    '%' + GSVVarToStr( vVal ) + '%' )
                            else  if loPartLeft in iLocateOpts
                                  then  AddItem( pAddNode, aftLike, sFld,
                                          GSVVarToStr( vVal ) + '%' )
                                  else  AddItem( pAddNode, aftEq, sFld, vVal );
                            end
                      else  AddItem( result, aftEq, sFld, vVal );
                      end;
                  end
            else  begin
                  TSALocate( oObj ).Add( pFldDescs^[ iFldNum ].iOffset,
                                         pFldDescs^[ iFldNum ].iFldType,
                                         vVal );
                  end;
            end;
      Inc( iCnt );
      sFld := ExtractSepString( sKeyFields, ';', iPos );
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalRefresh;
begin
  if IsCacheHasChanges
  then  RSAE( eUser, 0, Format( SA_S_ERR_NOFUNCEXECUTING,
                        [ 'TSaDs.ReFresh' ] ) )
  else  begin
        DoGlBuf( dgGlReFresh, false, 0, nil );
        if Assigned( mOnRefresh ) then mOnRefresh( Self );
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalPost;
var
  pActBuf : PChar;
begin
  isDoReFreshAfterPost := false;
  if State = dsEdit
  then  begin
        pActBuf := ActiveBuffer;
        if pActBuf <> nil
        then  begin
              if GetLocTitle( pActBuf )^.iAbsGlPos <>
                                                iGlFirstIndex + iGlCurrentPos
              then  RSAE( eUser, 0, SA_S_ERR_NOTUPDATERECORD );
              if not isCached then  DoCache( dcActivate, false );
              try
                DoCache( dcUpdateRec, false );
                if not isCached
                then  begin
                      if ( DoCache( dcApply, false ) > 0 )
                      then  DoCache( dcChangesToGlBuf, false );
                      isDoReFreshAfterPost := ( DsType = dsTable )
                        and ( CursorType = cDynScroll )
                        and ( TblKeys <> '' );
                      end;
              finally
                if not isCached then DoCache( dcDeActivate, false );
              end;
              end;
        end
  else  begin
        pActBuf := ActiveBuffer;
        if pActBuf <> nil
        then  begin
              if not isCached then DoCache( dcActivate, false );
              try
                DoCache( dcInsertRec, false );
                if not isCached
                then  begin
                      if DoCache( dcApply, false ) > 0
                      then  DoCache( dcChangesToGlBuf, false );
                      isDoReFreshAfterPost := ( DsType = dsTable )
                                          and ( CursorType = cDynScroll )
                                          and ( TblKeys <> '' );
                      end;
              finally
                if not isCached then  DoCache( dcDeActivate, false );
              end;
              end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalDelete;
var
  pActBuf     : PChar;
begin
  pActBuf := ActiveBuffer;
  if pActBuf <> nil
  then  begin

        DoGlBuf( dgEnsureNotFirstOrLast, false, 0, nil );

        if GetLocTitle( pActBuf )^.iAbsGlPos <> iGlFirstIndex + iGlCurrentPos
        then  RSAE( eUser, 0, SA_S_ERR_NOTDELETERECORD );
        if not isCached then DoCache( dcActivate, false );
        try

          DoCache( dcDeleteRec, false );
          if not isCached
          then  begin

                while iGlBuf <= SA_MIN_BUF
                do  DoGlBuf( dgNextDown, false, 0, nil );

                if DoCache( dcApply, false ) > 0
                then  DoCache( dcChangesToGlBuf, false );
                end;
        finally
          if not isCached then DoCache( dcDeActivate, false );
        end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.DoAfterPost;
begin
  if isDoReFreshAfterPost then ReFresh;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalAddRecord;
begin
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.InternalInitRecord;
begin
  SetNulls( false, Buffer );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetRecordCount;
begin
  result := iRecordCount;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetRecNo;
var
  pActiveBuffer : PChar;
begin
  if Active
  then  pActiveBuffer := ActiveBuffer
  else  pActiveBuffer := nil;
  if pActiveBuffer = nil
  then  result := 0
  else  result := GetLocTitle( pActiveBuffer )^.iAbsGlPos;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.IsSequenced;
begin
  result := ( soSequenced in Options );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
{$ifndef D_67}
  {$ifndef D_5}
function TSaDs.BCDToCurr;
begin
  Curr := currency( BCD^ );
  result := true;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.CurrToBCD;
begin
  currency( BCD^ ) := AlignCurrBcd( Curr, Precision, Decimals );
  result := true;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
  {$endif}
{$endif}
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.FetchAll;
begin
  if Active
  then  DoGlBuf( dgGetAll, false, 0, nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.OpenForInsert;
begin
  Active := false;
  isOpeningForInsert := true;
  try
    Active := true;
  finally
    isOpeningForInsert := false;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetCanModify;
begin
  result := Active and ( ( DsType = dsTable ) or ( Cached ) );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetActive;
begin
  if ( csReading in ComponentState )
  then  isStreamedActive := isVal
  else  inherited SetActive( isVal );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.SetFiltered;
begin
  if Filtered = isVal then exit;
  if Active
  then  begin
        CheckBrowseMode;
        inherited SetFiltered( isVal );
        ReOpen;
        end
  else  inherited SetFiltered( isVal );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.Loaded;
begin
  if isStreamedActive
  then  begin
        try
          Active := true;
        except
          on E : exception
          do  begin
              if not ( csDesigning in ComponentState ) then Raise;
              end;
        end;
        end;
  inherited Loaded;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDs.ReOpen;
begin
  if not Active then exit;
  DoGlBuf( dgOpenFirst, false, 0, nil );
  ReSync( [] );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetModified;
begin
  result := ( inherited Modified )
    or ( Active and Cached and IsCacheHasChanges );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.GetFieldOrder;
begin
  result := oOrderList.GetFieldSort( sFieldName );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDs.Resume;
begin
  result := false;
  if not Active then exit;
  result := oSaMain.Resume;
  if not result then exit;
  Include( iTmpOpts, stoOnResume );
  try
    Active := false;
    Active := true;
  finally
    Exclude( iTmpOpts, stoOnResume );
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
const
  AR_S_SAFLTITEMTYPE : array[ TSaFltItemType ] of string =
  ( 'and', 'or', '=', '>=', '>', '=', '<', '<=', '>', '>=', '<>', 'string',
    ' LIKE ' );

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaFlt.Destroy;
begin
  Clear;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaFlt.Clear;
begin
  DelItem( nil );
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaFlt.IsEmpty;
begin
  result := oFirst = nil;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaFlt.DelItem;
begin
  if oItem = nil
  then  begin
        while oFirst <> nil
        do  begin
            oItem  := oFirst.oNext;
            DelItem( oFirst );
            oFirst := oItem;
            end;
        end
  else  begin
        while oItem.oChild <> nil
        do  DelItem( oItem.oChild );
        if ( oItem.oParent <> nil ) and ( oItem.oPrev = nil )
        then  oItem.oParent.oChild := oItem.oNext;
        if ( oItem.oPrev <> nil )
        then  oItem.oPrev.oNext := oItem.oNext;
        if ( oItem.oNext <> nil )
        then    oItem.oNext.oPrev := oItem.oPrev;
        oItem.sValue := '';
        oItem.sParam := '';
        VarClear( oItem.vValue );
        if oItem = oFirst
        then  oFirst := oItem.oNext;
        oItem.Free;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

var
  _iNextPar : integer = 0;

function TSaFlt.AddItem;

  function GetNextPar : integer;
  begin
    if InterlockedIncrement( _iNextPar ) < 0
    then  InterlockedExchange( _iNextPar, 0 );
    result := _iNextPar;
  end;

  procedure SetValLength( var vValue : variant; iLen : integer );
  var
    s : string;
  begin
    s := vValue;
    SetLength( s, iLen );
    vValue := s;
  end;
var
  iTmp : integer;
begin
  result := nil;
  if ( not ( iType in [ aftAnd..aftOr, aftString ] ) )
    and ( Trim( sValue ) = '' )
  then  exit;
  if ( iType = aftString ) and ( VarIsNull( vValue ) or VarIsEmpty( vValue ) )
  then  exit;  
  if ( oParent <> nil ) and ( not ( oParent.iType in [ aftAnd..aftOr ] ) )
  then  exit;
  if iType in [ aftAnd..aftOr ]
  then  result := TSaFltNode.Create
  else  result := TSaFltCond.Create;
  result.iType := iType;
  result.oParent := oParent;
  if oParent <> nil
  then  begin
        if oParent.oChild <> nil
        then  oParent.oChild.oPrev := result;
        result.oNext   := oParent.oChild;
        oParent.oChild := result;
        end
  else  begin
        result.oNext  := oFirst;
        if oFirst <> nil
        then    oFirst.oPrev := result;
        oFirst := result;
        end;
  if ( iType in [ aftAnd..aftOr ] )
  then  begin
        end
  else  begin
        result.sValue := sValue;
        if ( iType = aftLike )
          and ( VarType( vValue ) = varString )
          and ( Length( vValue ) > 126 )
        then  SetValLength( vValue, 126 );
        if VarType( vValue ) = varArray + varInteger
        then  begin
              iTmp := vValue[ 1 ];
              result.pDesc := pointer( iTmp );
              iTmp := vValue[ 0 ];
              result.vValue := iTmp;
              if ( byte( ( pointer( iTmp ) )^ ) and SA_SERV_ISNULL ) <> 0
              then  result.sParam := ''
              else  result.sParam := '_' + IntToStr( GetNextPar );
              end
        else  begin
              result.vValue := vValue;
              if
                  ( VarIsEmpty( vValue ) or VarIsNull( vValue ) )
              then  result.sParam := ''
              else  result.sParam := '_' + IntToStr( GetNextPar );
              end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaFlt.GetFltString;
  function I_GetFltString(
      oItem : TSaFltItem
    ) : string;
  const
    AR_I_PARTRES : array[ aftEqPart..aftEqNullSub,
        boolean,
        boolean
      ] of TSaFltItemType
    = (
        ( ( aftEqAbove, aftEqLess ), ( aftEqLess, aftEqAbove ) ),
        ( ( aftAbove, aftLess ), ( aftLess, aftAbove ) ),
        ( ( aftEq, aftEq ), ( aftEq, aftEq ) )
      );

    AR_I_NULLS : array[ TSaFltItemType ] of byte =
      ( 0, 0, 1, 3, 2, 1, 2, 3, 2, 3, 2, 0, 1 );
    AR_S_NULLS : array[ 0..4 ] of string
      = ( 'error', ' is null', ' is not null', '11=11', '11=3' );

    function GetStrNull( sValue : string; iType : TSaFltItemType ) : string;
    var
      i : integer;
    begin
      i := AR_I_NULLS[ iType ];
      if i >= 3
      then  result := AR_S_NULLS[ i ]
      else  result := sValue + ' ' + AR_S_NULLS[ i ];
    end;
  var
    oFirst : TSaFltItem;
    iSign : TSaFltItemType;
    sTmp : string;
    isFirst : boolean;
  begin
    result := '';
    if oItem = nil
    then  begin
          iSign := aftAnd;
          oFirst := Self.oFirst;
          end
    else  begin
          iSign := oItem.iType;
          oFirst := oItem.oChild;
          end;
    if iSign in [ aftAnd..aftOr ]
    then  begin
          isFirst := true;
          while oFirst <> nil
          do  begin
              sTmp := I_GetFltString( oFirst );
              if sTmp <> ''
              then  begin
                    if result <> ''
                    then  begin
                          if isFirst
                          then  result := '(' + result + ')';
                          isFirst := false;
                          result  := result + ' ' +
                              AR_S_SAFLTITEMTYPE[ iSign ] + ' (' + sTmp + ')';
                          end
                    else  result := sTmp;
                    end;
              oFirst := oFirst.oNext;
              end;
          end
    else  begin
          if oItem.iType = aftString
          then  result := GsvVarToStr( oItem.vValue )
          else  begin
                if ( oItem.iType in [ aftEqPart, aftNotEqPart, aftEqNullSub ] )
                then  begin
                      sTmp := oOrderList.GetNullSubsByName( oItem.sValue,
                        oNullSubs );
                      if oItem.sParam = ''
                      then  begin
                            if sTmp <> ''
                            then  result := 'isnull( ' + oItem.sValue + ', ' +
                                                       sTmp + ' ) ' +
                                  AR_S_SAFLTITEMTYPE[
                                    AR_I_PARTRES[ oItem.iType, isInvert,
                                      oOrderList.GetFieldSort( oItem.sValue ) =
                                        otDesc ] ] + sTmp
                            else  result := GetStrNull( oItem.sValue,
                                                        oItem.iType );
                            end
                      else  begin
                            if sTmp <> ''
                            then  result := 'isnull( ' + oItem.sValue + ', ' +
                                                         sTmp + ' ) ' +
                                  AR_S_SAFLTItemTYPE[
                                    AR_I_PARTRES[ oItem.iType, isInvert,
                                      oOrderList.GetFieldSort( oItem.sValue ) =
                                        otDesc ] ] + ':' + oItem.sParam
                            else  result := oItem.sValue +
                                  AR_S_SAFLTITEMTYPE[
                                    AR_I_PARTRES[ oItem.iType, isInvert,
                                      oOrderList.GetFieldSort( oItem.sValue ) =
                                        otDesc ] ] + ':' + oItem.sParam;
                            end;
                      end
                else  begin
                      if oNullSubs = nil
                      then  sTmp := ''
                      else  sTmp := oNullSubs.GetNullSubs( oItem.sValue );

                      if sTmp <> ''
                      then  sTmp := 'isnull( ' + oItem.sValue + ', ' +
                                                       sTmp + ' ) '
                      else  sTmp := oItem.sValue;

                      if oItem.sParam = ''
                      then  result := GetStrNull( sTmp, oItem.iType )
                      else  result := sTmp +
                                  AR_S_SAFLTITEMTYPE[ oItem.iType ] + ':' +
                                  oItem.sParam;
                      end;
                end;
          end;
  end;
begin
  result := I_GetFltString( nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaFlt.FltToDSParams;

  procedure I_FltToTblParams(
                              oItem : TSaFltItem
                            );
  var
    oFirst  : TSaFltItem;
    isNode  : boolean;
    oParam  : TSaParam;
    p       : pointer;
    i       : integer;
  begin
    if oItem = nil
    then  begin
          isNode := true;
          oFirst := Self.oFirst;
          end
    else  begin
          isNode := oItem.iType in [ aftAnd..aftOr ];
          oFirst := oItem.oChild;
          end;
    if isNode
    then  begin
          while oFirst <> nil
          do  begin
              I_FltToTblParams( oFirst );
              oFirst := oFirst.oNext;
              end;
          end
    else  begin
          if oItem.sParam <> ''
          then  begin
                oParam := nil;
                for i := 0 to oParamList.Count - 1
                do  if CompareText( TSaParam( oParamList[ i ] ).ParamName,
                                    oItem.sParam ) = 0
                    then  begin
                          oParam := oParamList[ i ];
                          Break;
                          end;
                if oParam = nil then  exit;
                if oItem.pDesc <> nil
                then  begin
                      p := pointer( integer( oItem.vValue ) );
                      oParam.SetAsBuffer( oItem.pDesc, p );
                      end
                else  oParam.Value := oItem.vValue;
                end;
          end;
  end;
begin
  if oParamList = nil then exit;
  I_FltToTblParams( nil );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaFlt.CreateNode;
const
  AI_NODE_TO_ITEM : array[ TSaFltNodeType ] of TSaFltItemType =
    ( aftAnd, aftOr );
begin
  result := AddItem( oParentNode, AI_NODE_TO_ITEM[ iNodeType ], '', NULL ) as
    TSaFltNode;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaFlt.DestroyNode;
begin
  DelItem( oNode );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaFlt.CreateCond;
const
  AI_COND_TO_ITEM : array[ TSaFltCondType ] of TSaFltItemType =
    ( aftEq, aftLess, aftEqLess, aftAbove, aftEqAbove, aftNotEq, aftLike,
      aftString );
begin
  result := AddItem( oParentNode, AI_COND_TO_ITEM[ iCondType ],
    sCondField, vCondValue ) as TSaFltCond;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaFlt.DestroyCond;
begin
  DelItem( oCond );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
{* ************************************************************************** *}

procedure Register;
begin
  RegisterComponents( SA_S_VCLPAGENAME, [ TSaDb, TSaDs ] );
end;

end.



