{* SaVcl 2.5.  SaDsBase.pas. Serge Gavrilov ( C ) 2003 - 2006 *************** *}

unit SaDsBase;

{$i SaInc.inc}

{ unit description

}

interface

uses
  Windows, Classes, SysUtils, Db, 
{$ifdef D_67}
  Variants, FmtBcd,
{$else}
  SaAbsolute,
{$endif}
  SaCommon, SaConsts, SaUtils, SaDbBase;

type

{* ************************************************************************** *}

  TSaDsBase = class
  private
    iCursOpts : integer;
    oSaDbBase : TSaDbBase;
    iStatm : Ta_sql_statement_number;
    sStatm : string;
    sCursName : string;
    pOutput : TPSqldaRec;
    pInput : TPSqldaRec;
    isCursOpen : boolean;
    pLength : pointer;
    iRowAffected : integer;
    iRecordCount : integer;
    procedure FreeOutput;
    procedure ParamsToBind( oParams : TList );
    procedure ParamsByOut( oParams : TList );
    procedure SetSaDbBase( oVal : TSaDbBase );
    function GetIsPrepared : boolean;
    procedure StoreLength;
  public
    destructor Destroy; override;
    procedure RestoreLength;
    function AllocateFldDescs(
        var pFldDescs : TPFldDescs
      ) : integer;
    procedure OpenCurs(
        sStatm : string;
        iCursOpts : integer;
        oParams : TList
      );
    procedure OpenCursAfterResume;
    procedure CloseCurs;
    procedure ExecStatmImm( sStatm: string );
    procedure ExecStatm(
        sStatm : string;
        oParams : TList );
    procedure PutRec;
    property SaDbBase : TSaDbBase read oSaDbBase write SetSaDbBase;
    procedure PrepStatm(
        sStatm : string;
        isCursor : boolean;
        iCursOpts : integer;
        oParams : TList );
    function Fetch(
        isFetch : boolean;
        isAbs : boolean;
        iAbsRel : integer
      ) : TFetchResult;
    procedure GetSQLDABuf (
        isOut : boolean;
        isAfterFetch : boolean;
        iInd : integer;
        isBuf : boolean;
        pBuf : pointer;
        pv : PVariant;
        iTransl : integer );
    procedure SetSQLDABuf( isOut : boolean;
        iInd : integer;
        isBuf : boolean;
        pBuf : pointer;
        pv : PVariant;
        iTransl : integer );
    function Resume : boolean;
    property RowAffected : integer read iRowAffected;
    property RecordCount : integer read iRecordCount;
    property Prepared : boolean read GetIsPRepared;
  end;

{* ************************************************************************** *}

type

  TDoVLRec = (
      dvlCreate,
      dvlDestroy
    );
  procedure DoVLRec(
      iWhat : TDoVLRec;
      pRec : TPVLRec;
      iData : integer;
      pData : pointer
    ); forward;

implementation

function GetGenueType( p : TPSqlVarRec ) : integer;
begin
  result := p^.rcSqlName.iLength;
end;

type
  TFldDescAddData = record
    iSQLType : integer;
    iSQLLen : integer;
  end;
  TPFldDescAddData = ^TFldDescAddData;



{* ************************************************************************** *}
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure DoVLRec;
begin
  if iWhat = dvlCreate
  then  begin
        if ( iData <= 0 ) or ( pData = nil )
        then  DoVLRec( dvlDestroy, pRec, 0, nil )
        else  begin
              ReAllocMem( pRec^.pData, iData );
              pRec^.iData := iData;
              Move( pData^, pRec^.pData^, iData );
              end;
        end
  else  begin
        if ( pRec^.iData > 0 ) and ( pRec^.pData <> nil )
        then  FreeMem( pRec^.pData, pRec^.iData );
        ZeroMemory( pRec, sizeof( TVLRec ) );
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function GetBufSize(
    p : TPSqldaRec;
    iInd : integer
  ) : integer;
  function IsVLRec(
      iActType : integer;
      iGenueActType : integer;
      pRec : TPSqlVarRec
    ) : boolean;
  begin
    result :=
      ( ( ( iActType = DT_VARCHAR ) or ( iActType = DT_BINARY ) )
        and ( pRec^.iSqlLen - 3 > 255 ) )
      or ( ( iActType = DT_FIXCHAR ) and ( pRec^.iSqlLen > 255 ) )
      or ( ( iActType = DT_STRING ) and ( pRec^.iSqlLen > 255 ) )
      or ( iActType = DT_LONGVARCHAR )
      or ( iActType = DT_LONGBINARY )
      or ( iGenueActType = DT_LONGVARCHAR )
      or ( iGenueActType = DT_LONGBINARY );
  end;
var
  iActType, iGenueActType : integer;
  pRec : TPSqlVarRec;
begin
  result := 0;
  if ( p = nil ) or ( iInd >= p^.iSqld ) or ( iInd < 0 ) then exit;
    pRec     := @p^.arSqlVar[ iInd ];
    iActType := pRec^.iSqlType and SQL_NULL_MASK;
    iGenueActType := GetGenueType( pRec ) and SQL_NULL_MASK;
  case iActType of
    DT_SMALLINT, DT_TINYINT
      : result := sizeof( smallint );
    DT_INT
      : result := sizeof( integer  );
    DT_UNSINT
      : result := sizeof( LongWord  );
    DT_DECIMAL
      : result :=
{$ifdef D_67}
        sizeof( TBcd );
{$else}
        DecimalBufSize( pRec );
{$endif}        
    DT_FLOAT, DT_DOUBLE
      : result := sizeof( double );
    DT_TIMESTAMP_STRUCT
      : result := sizeof( TDateTime );
    DT_VARCHAR, DT_BINARY, DT_FIXCHAR,
    DT_STRING,
    DT_LONGVARCHAR, DT_LONGBINARY
      : if IsVLRec( iActType, iGenueActType, pRec )
        then  result := - sizeof( TVLRec )
        else  result :=   sizeof( byte )
                                        + pRec^.iSqlLen;
                                        
    DT_BIGINT, DT_UNSBIGINT
      : result := sizeof( Int64 );
    DT_BIT
      : result := sizeof( shortint );
    DT_UNSSMALLINT
      : result := sizeof( Word );
  else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                                [ 'SQLType=' + IntToStr( iActType ) ] ) );
  end;
  if result < 0
  then  result := result - sizeof( byte )
  else  result := result + sizeof( byte );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

procedure GetDataCallBack(
    pDest : pointer;
    pData : pointer;
    iData : integer;
    iOffs : integer;
    iTotal : integer;
    iTransl : integer
  ); stdcall;
begin
  if iTotal < 0
  then  begin
        TVLRec( pDest^ ).iData := SA_VLREC_ISNULL;
        exit;
        end;
  if TVLRec( pDest^ ).pData = nil
  then  begin
        TVLRec( pDest^ ).iData := iTotal;
        GetMem( TVLRec( pDest^ ).pData, TVLRec( pDest^ ).iData );
        end;
    if ( iTransl > 0 )
    then  Translate( pData, iData, tdToApp, iTransl );
    Move( pData^, TByteArray( TVLRec( pDest^ ).pData^ )[ iOffs ], iData );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.GetSQLDABuf;
var
  iActType : integer;
  pRec : TPSqlVarRec;
  pGenBuf : pointer;
  i : integer;
  sTmp : string;
  dTmp : TDateTime;
  rVLRec : TVLRec;
  p : TPSqldaRec;
{$ifdef D_567}
  rBCD : TBcd;
  iBCDBytes : integer;
{$endif}
begin
  if isOut
  then  p := pOutput
  else  p := pInput;
  if isBuf
  then  begin
        if ( pBuf = nil ) or ( p = nil ) or ( iInd >= p^.iSqld ) or ( iInd < 0 )
        then  exit;
        end
  else  begin
        if ( pv = nil ) or ( p = nil ) or ( iInd >= p^.iSqld ) or ( iInd < 0 )
        then  exit
        end;
  pRec     := @p^.arSqlVar[ iInd ];
  pGenBuf  := pBuf;
  iActType := pRec^.iSqlType and SQL_NULL_MASK;

  if ( ( iActType = DT_BINARY ) or ( iActType = DT_VARCHAR ) ) and isAfterFetch
  then  begin
        i := GetGenueType( pRec ) and SQL_NULL_MASK;
        if ( i = DT_LONGBINARY ) or ( i = DT_LONGVARCHAR ) then iActType := i;
        end;

  if isBuf
  then  begin
        if ( pRec^.pSqlInd = nil ) or ( pRec^.pSqlInd^ < 0 )
        then  begin
              TByteArray( pBuf^ )[ 0 ] :=
                TByteArray( pBuf^ )[ 0 ] or SA_SERV_ISNULL;
              exit;
              end;
        TByteArray( pBuf^ )[ 0 ] :=
          TByteArray( pBuf^ )[ 0 ] and not SA_SERV_ISNULL;
        pBuf := @TByteArray( pBuf^ )[ 1 ];
        end
  else  begin
        if ( pRec^.pSqlInd = nil ) or ( pRec^.pSqlInd^ < 0 )
        then  begin
              pv^ := NULL;
              exit;
              end;
        end;

  case iActType of
    DT_DECIMAL
      : begin
{$ifdef D_567}
  {$ifdef D_5}
        if DecimalIsFloat( pRec )
        then  GetDecimalSQLDA( pRec, isBuf, pBuf, pv )
        else  begin
  {$endif}
        ZeroMemory( @rBCD, sizeof( rBCD ) );
        rBCD.Precision := LOBYTE( pRec^.iSqlLen );
        rBCD.SignSpecialPlaces := HIBYTE( pRec^.iSqlLen );
        if not Odd( rBCD.Precision )
        then  begin
              iBCDBytes := ( rBCD.Precision + 1 ) div 2 + 1;
              for i := 0 to iBCDBytes - 2
              do  rBCD.Fraction[ i ] :=
                  ( TByteArray( pRec^.pSqlData^ )[ i ] and $0F ) shl 4 +
                  ( TByteArray( pRec^.pSqlData^ )[ i + 1 ] and $F0 ) shr 4;
              end
        else  begin
              iBCDBytes := rBCD.Precision div 2 + 1;
              Move( pRec^.pSqlData^, rBCD.Fraction, iBCDBytes );
              end;
        if ( TByteArray( pRec^.pSqlData^ )[ iBCDBytes - 1 ] and $0F ) = $0D
        then  Inc( rBCD.SignSpecialPlaces, 1 shl 7 );
        if isBuf
        then  TBcd( pBuf^ ) := rBCD
        else
    {$ifndef D_5}
        VarFMTBcdCreate( pv^, rBCD );
    {$else}
        pv^ := BcdToVar( rBCD );
    {$endif}

  {$ifdef D_5}
              end;
  {$endif}
{$else}
        GetDecimalSQLDA( pRec, isBuf, pBuf, pv );
{$endif}
        end;
  DT_BIT
  : if isBuf
    then  shortint( pBuf^ ) := shortint( pRec^.pSqlData^ )
    else  pv^ := VarAsType( shortint( pRec^.pSqlData^ ), varBoolean );
  DT_FLOAT
  : if isBuf
    then  double  ( pBuf^ ) := single  ( pRec^.pSqlData^ )
    else  pv^ := VarAsType( single( pRec^.pSqlData^ ), varSingle );
  DT_DOUBLE
  : if isBuf
    then  double  ( pBuf^ ) := double  ( pRec^.pSqlData^ )
    else  pv^ := VarAsType( double( pRec^.pSqlData^ ), varDouble );
  DT_TIMESTAMP_STRUCT
  : begin
    WITH TSADTRec( pRec^.pSqlData^ )
    do  begin
        dTmp := EncodeDate( iYear, Succ( bMonth ), bDay );
        if dTmp < 0
        then  dTmp := dTmp - EncodeTime( bHour, bMinute, bSecond, 0 )
        else  dTmp := dTmp + EncodeTime( bHour, bMinute, bSecond, 0 );
        if isBuf
        then  TDateTime( pBuf^ ) := dTmp
        else  pv^ := VarAsType( dTmp, varDate );
        end;
    end;
  DT_FIXCHAR, DT_STRING
  : begin
    if ( iActType = DT_FIXCHAR )
    then  begin
          i := pRec^.iSqlLen - 1;
          while ( i >= 0 )
                       and ( ( PChar( pRec^.pSqlData )[ i ] = ' ' )
                               or
                             ( PChar( pRec^.pSqlData )[ i ] = #0 ) )
          do  Dec( i );
          Inc( i );
          end
    else  i := StrLen( PChar( pRec^.pSqlData ) );
    if isBuf
    then  begin
          if ( TByteArray( pGenBuf^ )[ 0 ] and SA_SERV_VLREC ) <> 0
          then  begin
                DoVLRec( dvlCreate, pBuf, i, pRec^.pSqlData );
                Translate( TVLRec( pBuf^ ).pData, TVLRec( pBuf^ ).iData,
                           tdToApp, iTransl );
                end
          else  begin
                byte( pBuf^ ) := byte( i );
                pBuf := @TByteArray( pBuf^ )[ 1 ];
                Move( pRec^.pSQLData^, pBuf^, i );
                Translate( pBuf, i, tdToApp, iTransl );
                end;
          end
    else  begin
          SetLength( sTmp, i );
          Move( pRec^.pSQLData^, sTmp[ 1 ], i );
          pv^ := VarAsType( sTmp, varString );
          end;
    end;
  DT_VARCHAR, DT_BINARY
  : begin
    if isBuf
    then  begin
          if ( TByteArray( pGenBuf^ )[ 0 ] and SA_SERV_VLREC ) <> 0
          then  begin
                DoVLRec(  dvlCreate, pBuf,
                          TVarCharRec( pRec^.pSqlData^ ).iLen,
                          @TVarCharRec( pRec^.pSqlData^ ).sChar );
                if ( iActType = DT_VARCHAR )
                then  Translate( TVLRec( pBuf^ ).pData, TVLRec( pBuf^ ).iData,
                                 tdToApp, iTransl );
                end
          else  begin
                i := byte( TVarCharRec( pRec^.pSqlData^ ).iLen );
                byte( pBuf^ ) := byte( i );
                pBuf := @TByteArray( pBuf^ )[ 1 ];
                Move( TVarCharRec( pRec^.pSqlData^ ).sChar, pBuf^, i );
                if iActType = DT_VARCHAR
                then  Translate( pBuf, i, tdToApp, iTransl );
                end;
          end
    else  begin
          i := TVarCharRec( pRec^.pSqlData^ ).iLen;
          SetLength( sTmp, i );
          Move( TVarCharRec( pRec^.pSqlData^ ).sChar, sTmp[ 1 ], i );
          if iActType = DT_VARCHAR
          then  Translate( @sTmp[ 1 ], i, tdToApp, iTransl );
          pv^ := VarAsType( sTmp, VARString );
          end;
    end;
  DT_LONGVARCHAR, DT_LONGBINARY
  : begin
    if oSaDbBase = nil then exit;
    if not isAfterFetch then exit;
    if iActType = DT_LONGBINARY then iTransl := 0;
    ZeroMemory( @rVLRec, sizeof( rVLRec ) );
    try
      oSaDbBase.SaGetData( PChar( sCursName ),
                           iInd + 1,
                           @rVLRec,
                           @GetDataCallBack,
                           iTransl
                         );
      if rVLRec.iData = SA_VLREC_ISNULL
      then  begin
            rVLRec.iData := 0;
            byte( pGenBuf^ ) := byte( pGenBuf^ ) or SA_SERV_ISNULL;
            end;
      if isBuf
      then  TVLRec( pBuf^ ) := rVLRec
      else  if ( rVLRec.iData > 0 ) and ( rVLRec.pData <> nil )
            then  begin
                  SetLength( sTmp, rVLRec.iData );
                  Move( rVLRec.pData^, sTmp[ 1 ], rVLRec.iData );
                  pv^ := sTmp;
                  end
            else  pv^ := NULL;
    finally
      if not isBuf
      then  DoVLRec( dvlDestroy, @rVLRec, 0, nil );
    end;
    end;
    DT_BIGINT, DT_UNSBIGINT
    : begin
      if isBuf
      then  Int64( pBuf^ ) := Int64( pRec^.pSqlData^ )
      else  pv^ :=
{$ifdef D_67}
                    VarAsType( Int64( pRec^.pSqlData^ ), varInt64 );
{$else}
                    Int64ToVar( Int64( pRec^.pSqlData^ ) );
{$endif}
      end;
    DT_INT
    : if isBuf
      then  integer ( pBuf^ ) := integer ( pRec^.pSqlData^ )
      else  pv^ := VarAsType( integer( pRec^.pSqlData^ ), varInteger );
    DT_SMALLINT
    : if isBuf
      then  smallint( pBuf^ ) := smallint( pRec^.pSqlData^ )
      else  pv^ := VarAsType( smallint( pRec^.pSqlData^ ), varSmallInt );
    DT_TINYINT
    : if isBuf
      then  smallint( pBuf^ ) := shortint( pRec^.pSqlData^ )
      else  pv^ := VarAsType( shortint( pRec^.pSqlData^ ), varShortInt );
    DT_UNSINT
    : begin
      if isBuf
      then  LongWord( pBuf^ ) := LongWord( pRec^.pSqlData^ )
{$ifdef D_5}
      else  pv^ := VarAsType( LongInt( pRec^.pSqlData^ ), varLongWord );
{$else}
      else  pv^ := VarAsType( LongWord( pRec^.pSqlData^ ), varLongWord );
{$endif}
      end;
    DT_UNSSMALLINT
    : begin
      if isBuf
      then  Word( pBuf^ ) := Word( pRec^.pSqlData^ )
      else  pv^ := VarAsType( Word( pRec^.pSqlData^ ), varWord );
      end;
  else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                                [ 'SQLType=' + IntToStr( iActType ) ] ) );
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
{$ifdef D_567}
function  BCDToBCD        (
                            const rBCD  : TBCD;
                            iPrec       : byte;
                            iScale      : byte
                          ) : TBCD;
  function MinByte( i1, i2 : byte ) : byte;
  begin
    if i1 > i2
    then  result := i2
    else  result := i1;
  end;
type
  TBytes  = array[ 0..0 ] of byte;
  TPBytes = ^TBytes;
  function GetNibble( pNibbles : TPBytes; iIndex : byte ) : byte;
  var
    iTmp : byte;
  begin
    iTmp := iIndex SHR 1;
    if Odd( iIndex )
    then  result := pNibbles[ iTmp ] and $0F
    else  result := ( pNibbles[ iTmp ] and $F0 ) SHR 4;
  end;
  procedure SetNibble( pNibbles : TPBytes; iIndex : byte; iNibble : byte );
  var
    iTmp : byte;
  begin
    iTmp := iIndex SHR 1;
    if Odd( iIndex )
    then  pNibbles[ iTmp ] := ( pNibbles[ iTmp ] and $F0 ) + ( iNibble and $0F )
    else  pNibbles[ iTmp ] := ( pNibbles[ iTmp ] and $0F ) +
          ( ( iNibble and $0F ) SHL 4 );
  end;
var
  iBCDSN  : byte;
  iResSN  : byte;
  iBCDS   : byte;
  iTmp    : integer;
begin
  iBCDS := rBCD.SignSpecialPlaces and $3F;
  if ( iBCDS = iScale ) and ( iPrec = rBCD.Precision )
  then  begin
        result := rBCD;
        exit;
        end;
  with result
  do  begin
      SignSpecialPlaces := ( rBCD.SignSpecialPlaces and ( 1 SHL 7 ) ) + iScale;
      Precision := iPrec;
      ZeroMemory( @Fraction, sizeof( Fraction ) );
      end;
  iBCDSN  := rBCD.Precision - iBCDS;
  iResSN  := iPrec - iScale;
  for iTmp := 0 to MinByte( iScale, iBCDS ) - 1
  do  SetNibble( @result.Fraction, iResSN + iTmp,
                  GetNibble( @rBCD.Fraction, iBCDSN + iTmp ) );
  for iTmp := 1 to MinByte( iResSN, iBCDSN )
  do  SetNibble( @result.Fraction, iResSN - iTmp,
                  GetNibble( @rBCD.Fraction, iBCDSN - iTmp ) );
end;
{$endif}
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.SetSQLDABuf;
var
  i, iTmp, iActType : integer;
  pRec        : TPSqlVarRec;
  pGenBuf     : pointer;
  w1, w2,
  w3, w4      : word;
  pVLRec      : TPVLRec;
  p           : TPSqldaRec;
{$ifdef D_567}
  rBCD : TBCD;
  iBCDBytes : integer;
  iBCDSign : byte;
{$endif}
begin

  if isOut
  then  p := pOutput
  else  p := pInput;

  if ( p = nil ) or ( iInd >= p^.iSqld ) or ( iInd < 0 ) then  exit;
  pRec  := @p^.arSqlVar[ iInd ];

  if ( isBuf
    and ( ( pBuf = nil ) or ( ( byte( pBuf^ ) and SA_SERV_ISNULL  ) <> 0 ) ) )
    or
     ( not isBuf
    and ( ( pv = nil ) or ( VarIsEmpty( pv^ ) ) or ( VarIsNull( pv^ ) ) ) )
  then  begin
        if ( pRec^.pSqlInd <> nil ) then pRec^.pSqlInd^ := -1;
        exit;
        end;
        
  iActType := pRec^.iSqlType and SQL_NULL_MASK;
  pRec^.pSqlInd^ :=  0;
  pGenBuf := pBuf;
  pBuf := @TByteArray( pBuf^ )[ 1 ];

  case iActType of
    DT_DECIMAL
    : begin
{$ifdef D_567}
  {$ifdef D_5}
      if DecimalIsFloat( pRec )
      then  SetDecimalSQLDA( pRec, isBuf, pBuf, pv )
      else  begin
  {$endif}
      if isBuf
      then  rBCD := TBCD( pBuf^ )
      else  rBCD := VarToBcd( pv^ );
      rBCD := BCDToBCD( rBCD, LOBYTE( pRec^.iSqlLen ), HIBYTE( pRec^.iSqlLen ));
      if not Odd( rBCD.Precision )
      then  begin
            iBCDBytes := ( rBCD.Precision + 1 ) div 2 + 1;
            if iBCDBytes > 0
            then  TByteArray( pRec^.pSqlData^ )[ 0 ] :=
                              ( rBCD.Fraction[ 0 ] and $F0 ) SHR 4;
            for i := 1 to iBCDBytes - 1
            do  TByteArray( pRec^.pSqlData^ )[ i ] :=
                  ( rBCD.Fraction[ i - 1 ] and $0F ) SHL 4 +
                  ( rBCD.Fraction[ i ] and $F0 ) SHR 4;
            end
      else  begin
            iBCDBytes := rBCD.Precision div 2 + 1;
            Move( rBCD.Fraction, pRec^.pSqlData^, iBCDBytes );
            end;


      iBCDSign := TByteArray( pRec^.pSqlData^ )[ iBCDBytes - 1 ] and $F0;
      if ( rBCD.SignSpecialPlaces and ( 1 SHL 7 ) ) <> 0
      then  Inc( iBCDSign, $0D )
      else  Inc( iBCDSign, $0C );
      TByteArray( pRec^.pSqlData^ )[ iBCDBytes - 1 ] := iBCDSign;
  {$ifdef D_5}
            end;
  {$endif}
{$else}
      SetDecimalSQLDA( pRec, isBuf, pBuf, pv );
{$endif}
      end;
  DT_BIT
  : if isBuf
    then  shortint( pRec^.pSqlData^ ) := shortint( pBuf^ )
    else  begin
          if pv^ = TRUE
          then  shortint( pRec^.pSqlData^ ) := 1
          else  shortint( pRec^.pSqlData^ ) := 0
          end;
  DT_BIGINT, DT_UNSBIGINT
  : if isBuf
    then  Int64( pRec^.pSqlData^ ) := Int64( pBuf^ )
    else  Int64( pRec^.pSqlData^ ) :=
{$ifdef D_67}
                                      pv^;
{$else}
                                      VarToInt64( pv^ );
{$endif}
  DT_INT
  : if isBuf
    then  integer ( pRec^.pSqlData^ ) := integer ( pBuf^ )
    else  integer ( pRec^.pSqlData^ ) := integer( pv^ );
  DT_SMALLINT
  : if isBuf
    then  smallint( pRec^.pSqlData^ ) := smallint( pBuf^ )
    else  smallint( pRec^.pSqlData^ ) := smallint( pv^ );
  DT_TINYINT
  : if ISBuf
    then  shortint( pRec^.pSqlData^ ) := smallint( pBuf^ )
    else  if ( oSaDbBase.SaVersion = sa50 ) and ( VarType( pv^ ) = varBoolean )
          then  begin
                if pv^ = true
                then  shortint( pRec^.pSqlData^ ) := 1
                else  shortint( pRec^.pSqlData^ ) := 0;
                end
          else  shortint( pRec^.pSqlData^ ) := shortint( pv^ );
  DT_UNSINT
  : if isBuf
    then  LongWord( pRec^.pSqlData^ ) := LongWord( pBuf^ )
    else  LongWord( pRec^.pSqlData^ ) := LongWord( pv^ );
  DT_UNSSMALLINT
  : if isBuf
    then  Word( pRec^.pSqlData^ ) := Word( pBuf^ )
    else  Word( pRec^.pSqlData^ ) := Word( pv^ );
  DT_FLOAT
  : if isBuf
    then    single  ( pRec^.pSqlData^ ) := double ( pBuf^ )
    else    single( pRec^.pSqlData^ ) := single( pv^ );
  DT_DOUBLE
  : if isBuf
    then    double  ( pRec^.pSqlData^ ) := double ( pBuf^ )
    else    double  ( pRec^.pSqlData^ ) := double( pv^ );
  DT_TIMESTAMP_STRUCT
  : begin
    ZeroMemory( pRec^.pSqlData, sizeof( TSADTRec ) );
    if isBuf
    then    DecodeDate( TDateTime( pBuf^ ), w1, w2, w3 )
    else    DecodeDate( TDateTime( pv^ ), w1, w2, w3 );
    TSADTRec( pRec^.pSqlData^ ).iYear      := w1;
    TSADTRec( pRec^.pSqlData^ ).bMonth     := Pred( w2 );
    TSADTRec( pRec^.pSqlData^ ).bDay       := w3;
    if isBuf
    then    DecodeTime( TDateTime( pBuf^ ), w1, w2, w3, w4 )
    else    DecodeTime( TDateTime( pv  ^ ), w1, w2, w3, w4 );
    TSADTRec( pRec^.pSqlData^ ).bHour      := w1;
    TSADTRec( pRec^.pSqlData^ ).bMinute    := w2;
    TSADTRec( pRec^.pSqlData^ ).bSecond    := w3;
    end;
  DT_FIXCHAR, DT_STRING
  : begin

    ZeroMemory( pRec^.pSqlData, pRec^.iSqlLen );
    if isBuf
    then  begin
          if ( byte( pGenBuf^ ) and SA_SERV_VLREC ) <> 0
          then  begin
                pVLRec := pBuf;
                if iActType = DT_STRING
                then  begin
                      iTmp := pVLRec^.iData;
                      if iTmp >= pRec^.iSqlLen then iTmp := pRec^.iSqlLen - 1;
                      end
                else  begin
                      pRec^.iSqlLen := pVLRec^.iData;
                      iTmp := pRec^.iSqlLen;
                      end;
                if iTmp > 0 then Move( pVLRec^.pData^, pRec^.pSqlData^, iTmp );
                end
          else  begin
                if iActType = DT_STRING
                then  begin
                      iTmp := byte( pBuf^ );
                      if iTmp >= pRec^.iSqlLen then iTmp := pRec^.iSqlLen - 1;
                      end
                else  begin
                      pRec^.iSqlLen := byte( pBuf^ );
                      iTmp := pRec^.iSqlLen;
                      end;
                pBuf := @TByteArray( pBuf^ )[ 1 ];
                if iTmp > 0 then Move( pBuf^, pRec^.pSqlData^, iTmp );
                end;
          end
    else  begin
          iTmp := Length( pv^ );
          if iActType = DT_STRING
          then  begin
                if iTmp >= pRec^.iSqlLen then iTmp := pRec^.iSqlLen - 1;
                end
          else  begin
                if iTmp > pRec^.iSqlLen then iTmp := pRec^.iSqlLen;
                pRec^.iSqlLen := iTmp;
                end;
          if iTmp > 0 then Move( string( pv^ )[ 1 ], pRec^.pSqlData^, iTmp );
          end;
    Translate( pRec^.pSqlData, iTmp, tdToDB, iTransl );
    end;
  DT_VARCHAR, DT_BINARY
  : begin
    if isBuf
    then  begin
          if ( byte( pGenBuf^ ) and SA_SERV_VLREC ) <> 0
          then  begin
                pVLRec := pBuf;
                TVarCharRec( pRec^.pSqlData^ ).iLen := pVLRec^.iData;
                Move( pVLRec^.pData^, TVarCharRec( pRec^.pSqlData^ ).sChar,
                      pVLRec^.iData );
                end
          else  begin
                TVarCharRec( pRec^.pSqlData^ ).iLen := byte( pBuf^ );
                pBuf := @TByteArray( pBuf^ )[ 1 ];
                Move( pBuf^, TVarCharRec( pRec^.pSqlData^ ).sChar,
                      TVarCharRec( pRec^.pSqlData^ ).iLen );
                end;
          end
    else  begin
          i := Length( pv^ );
          if i > pRec^.iSqlLen - 3
          then    i := pRec^.iSqlLen - 3;
          TVarCharRec( pRec^.pSqlData^ ).iLen := i;
          Move( string( pv^ )[ 1 ], TVarCharRec( pRec^.pSqlData^ ).sChar,
                                    TVarCharRec( pRec^.pSqlData^ ).iLen );
          end;
    if iActType = DT_VARCHAR
    then    Translate( @TVarCharRec( pRec^.pSqlData^ ).sChar,
                       TVarCharRec( pRec^.pSqlData^ ).iLen, tdToDB, iTransl );
    end;
  else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                                [ 'SQLType=' + IntToStr( iActType ) ] ) );
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaDsBase.Destroy;
begin
  PrepStatm( '', false, 0, nil );
  INHERITED;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.FreeOutput;
begin
  if ( pOutput <> nil )
  then  begin
        FreeMem( pLength ); pLength := nil;
        oSaDbBase.SaFreeFilledDa( pOutput );
        pOutput := nil;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
var
  _iStatmCount : integer = 0;
procedure TSaDsBase.PrepStatm;


  procedure CheckBindField( i : integer );
  var
    oP : TSaParam;
    iL : integer;
{$ifdef D_67}
    rBcd : TBcd;
{$endif}
  begin
    oP := TSaParam( oParams[ i ] );
    if ( ( pInput^.arSqlVar[ i ].pSqlInd^ and DT_PARAM_INPUT ) <> 0 )
        and
        ( ( pInput^.arSqlVar[ i ].pSqlInd^ and DT_PROCEDURE_OUT ) <> 0 )
    then  oP.ParamType := ptInputOutput
    else  if ( ( pInput^.arSqlVar[ i ].pSqlInd^ and DT_PARAM_INPUT ) <> 0 )
          then  oP.ParamType := ptInput
          else  if( ( pInput^.arSqlVAR[ i ].pSqlInd^ and DT_PROCEDURE_OUT )<> 0)
                then  oP.ParamType := ptOutput
                else  oP.ParamType := ptRESULT;

    if ( ( pInput^.arSqlVar[ i ].pSqlInd^ and DT_DESCRIBE_INPUT ) <> 0 )
        and ( pInput^.arSqlVar[ i ].iSqlLen = 0 )
    then  begin
          if ( oP.Desc <> nil ) and ( oP.Data <> nil )
          then  begin
                pInput^.arSqlVar[ i ].iSqlLen  :=
                          TFldDescAddData( oP.Desc^.pAddData^ ).iSQLLen;
                pInput^.arSqlVar[ i ].iSqlType :=
                          TFldDescAddData( oP.Desc^.pAddData^ ).iSQLType;
                end
          else  begin
                case VarType( oP.Value ) of
                  varEmpty, varNull : ;
{$ifdef D_67}
                  varShortInt,
{$endif}
                  varSmallint, varInteger
                  : begin
                    pInput^.arSqlVar[ i ].iSqlLen  := 4;
                    pInput^.arSqlVar[ i ].iSqlType := DT_INT;
                    end;
{$ifdef D_67}
                  varLongWord, varWord,
{$endif}
                  varByte
                  : begin
                    pInput^.arSqlVar[ i ].iSqlLen  := 4;
                    if Self.oSaDbBase.SaVersion = sa50
                    then  pInput^.arSqlVar[ i ].iSqlType := DT_INT
                    else  pInput^.arSqlVar[ i ].iSqlType := DT_UNSINT;
                    end;
{$ifdef D_67}
                  varInt64
                  : begin
                    pInput^.arSqlVar[ i ].iSqlLen  := 8;
                    pInput^.arSqlVar[ i ].iSqlType := DT_BIGINT;
                    end;
{$endif}
                  varBoolean
                  : begin
                    if Self.oSaDbBase.SaVersion = sa50
                    then  begin
                          pInput^.arSqlVar[ i ].iSqlLen  := 1;
                          pInput^.arSqlVar[ i ].iSqlType := DT_TINYINT
                          end
                    else  begin
                          pInput^.arSqlVar[ i ].iSqlLen  := 1;
                          pInput^.arSqlVar[ i ].iSqlType := DT_BIT;
                          end;
                    end;
                  varSingle, varDouble, varCurrency
                  : begin
                    pInput^.arSqlVar[ i ].iSqlLen  := 8;
                    pInput^.arSqlVAR[ i ].iSqlType := DT_DOUBLE;
                    end;
                  varDate
                  : begin
                    pInput^.arSqlVar[ i ].iSqlLen  := 14;
                    pInput^.arSqlVar[ i ].iSqlType := DT_TIMESTAMP_STRUCT;
                    end;
                  varString,
                  varOLEStr
                  : begin
                    iL := Length( oP.Value );
                    if iL = 0 then iL := 1;
                    if iL > SA_MAXDATA_LEN
                    then  begin
                          pInput^.arSqlVar[ i ].iSqlLen  := SA_MAXDATA_LEN;
                          pInput^.arSqlVar[ i ].iSqlType := DT_VARCHAR;
                          end
                    else  begin
                          pInput^.arSqlVar[ i ].iSqlLen  := iL;
                          if oSaDbBase.UseDtString
                          then  pInput^.arSqlVar[ i ].iSqlType := DT_STRING
                          else  pInput^.arSqlVar[ i ].iSqlType := DT_FIXCHAR;
                          end;
                    end;
                  else  begin
{$ifdef D_67}
                        if VarIsFMTBcd( oP.Value )
                        then  begin
                              rBCD := VarToBcd( oP.Value );

                              pInput^.arSqlVar[ i ].iSqlLen  :=
                                MAKEWORD( rBCD.Precision,
                                          rBCD.SignSpecialPlaces and $3F );
                              pInput^.arSqlVar[ i ].iSqlType := DT_DECIMAL;
                              end
                        else
{$endif}
                              RSAE( eInternal, 0,
                                   Format( SA_S_ERR_NOTSUPFLDTYPE,
                      [ 'VarType=' + IntToHex( VarType( oP.Value ), 8 ) ] ) );
                        end;
                end;
                end;
          pInput^.arSqlVar[ i ].iSqlType := pInput^.arSqlVar[ i ].iSqlType
                                                    or ( not SQL_NULL_MASK );
          end;
  end;
var
  i, iParamCount : integer;
begin

  if oSaDbBase = nil then exit;
  if ( pInput <> nil )
  then  begin
        oSaDbBase.SaFreeFilledDa( pInput );
        pInput := nil;
        end;

  if ( sStatm <> Self.sStatm ) or ( iCursOpts <> Self.iCursOpts )
  then  begin
        if Self.sStatm <> ''
        then  begin
              CloseCurs;
              FreeOutput;
              oSaDbBase.SaUnPrepare( iStatm );
              iStatm := 0;
              sCursName := '';
              end;

        Self.sStatm := '';
        Self.iCursOpts := iCursOpts;

        if ( sStatm <> '' )
        then  begin
              oSaDbBase.SaPrepare( PChar( sStatm ), @iStatm );
              if isCursor
              then  begin
                    randomize;
                    InterlockedIncrement( _iStatmCount );
                    sCursName := 's'
                      + IntToStr( _iStatmCount )
                        + IntToStr( GetTickCount );
                    oSaDbBase.SaDeclare( PChar( sCursName ), iStatm,
                      Self.iCursOpts );
                    end;
              Self.sStatm := sStatm;
              end;
        end;
  if ( Self.sStatm <> '' )
  then  begin
        oSaDbBase.SaDescribe( true, nil, iStatm, @pInput );
        if ( oParams <> nil )
        then  iParamCount := oParams.Count
        else  iParamCount := 0;
        if iParamCount > pInput^.iSqld
        then  iParamCount := pInput^.iSqld;
        for i := 0 to iParamCount - 1 do CheckBindField( i );
        if pInput^.iSqld > 0 then oSaDbBase.SaFillDa( pInput );
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.ParamsToBind;
var
  i   : integer;
  oP  : TSAParam;
begin
  if ( oParams <> nil ) and ( pInput <> nil ) and ( oSaDbBase <> nil )
  then  begin
          for i := 0 to pInput^.iSqld - 1
          do  if ( i < oParams.Count )
              then  begin
                    oP := TSAParam( oParams[ i ] );
                    if ( oP.ParamType IN [ ptInput, ptInputOutput ] )
                    then  begin
                          if ( oP.Desc <> nil ) and ( oP.Data <> nil )
                          then  SetSQLDABuf( FALSE, i, TRUE, oP.Data, nil,
                                  oSaDbBase.TranslatorNum )
                          else  SetSQLDABuf( FALSE, i, FALSE, nil, @oP.Value,
                                  oSaDbBase.TranslatorNum );
                          end;
                    end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.ParamsByOut;
var
  i, iInc : integer;
  oP : TSaParam;
begin
  if ( oParams <> nil ) and ( pOutput <> nil ) and ( oSaDbBase <> nil )
  then  begin
        iInc := 0;
        for i := 0 to oParams.Count - 1
        do  begin
            oP := TSaParam( oParams[ i ] );
            if ( oP.ParamType in [ ptOutput, ptInputOutput ] )
              and ( iInc < pOutput^.iSqld )
            then  begin
                  GetSQLDABuf( true, false, iInc, FALSE, nil,
                    @oP.Value, oSaDbBase.TranslatorNum );
                  Inc( iInc );
                  end;
            end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.ExecStatmImm;
begin
  if oSaDbBase = nil then exit;
  CloseCurs;
  oSaDbBase.SQLM_PreparedQuery( Self, pInput, sStatm );
  try
    oSaDbBase.SaExecImm( PChar( sStatm ) );
    oSaDbBase.TransactAutoCommit;
  except
    oSaDbBase.TransactAutoRollBack;
    raise;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.ExecStatm;
begin
  if oSaDbBase = nil then exit;
  CloseCurs;
  oSaDbBase.SQLM_Query( Self, sStatm );
  PrepStatm( sStatm, false, 0, oParams );
  ParamsToBind( oParams );
  iRowAffected := 0;
  oSaDbBase.SQLM_PreparedQuery( Self, pInput, sStatm );
  if pOutput = nil
  then  begin
        oSaDbBase.SaDescribe( false, nil, iStatm, @pOutput );
        oSaDbBase.SaFillDa( pOutput );
        StoreLength;
        end;
  try
    oSaDbBase.SaExec( iStatm, pInput, pOutput );
    iRowAffected := oSaDbBase.GetSQLCount;
    ParamsByOut( oParams );
    oSaDbBase.TransactAutoCommit;
  except
    oSaDbBase.TransactAutoRollBack;
    raise;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.OpenCurs;
var
  isTmp : boolean;
begin
  if oSaDbBase = nil then exit;
  CloseCurs;
  isTmp := ( sStatm <> Self.sStatm );
  oSaDbBase.SQLM_Query( Self, sStatm );
  PrepStatm( sStatm, true, iCursOpts, oParams );
  ParamsToBind( oParams );
  oSaDbBase.SQLM_PreparedQuery( Self, pInput, sStatm );
  oSaDbBase.SaOpen( PChar( sCursName ), pInput );

  iRecordCount := oSaDbBase.GetSQLCount;
  if pOutput = nil
  then  begin
        oSaDbBase.SaDescribe( false, PChar( sCursName ), iStatm, @pOutput );
        oSaDbBase.SaFillDa( pOutput );
        StoreLength;
        end;
  if ( not isTmp ) and ( ( iCursOpts and SA_CURSOPT_NOSCROLL ) = 0 )
  then  begin
        oSaDbBase.SaFetch( PChar( sCursName ), nil, TRUE, 0 );
        if ( iCursOpts and SA_CURSOPT_INSENSITIVE ) <> 0
        then  oSaDbBase.SaFetch( PChar( sCursName ), nil, TRUE, 0 );
        end;

  isCursOpen := TRUE;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.OpenCursAfterResume;
begin
  if oSaDbBase = nil then exit;
  FreeOutput;
  oSaDbBase.SaDescribe( false, PChar( sCursName ), iStatm, @pOutput );
  oSaDbBase.SaFillDa( pOutput );
  StoreLength;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.CloseCurs;
begin
  if oSaDbBase = nil then exit;
  if isCursOpen
  then  begin
        isCursOpen := FALSE;
        iRecordCount := 0;
        oSaDbBase.SaClose( PChar( sCursName ) );
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
type
  TFldInfo = record
    iFldLen : integer;
    pFldPtr : pointer;
  end;
  TFldInfos = record
    iLongDataCount : integer;
    arInfos : array[ 0..0 ] of TFldInfo;
  end;
procedure TSaDsBase.StoreLength;
var
  i : integer;
  iGenueType : integer;
begin
  if ( pLength <> nil ) or ( pOutput = nil ) or ( pOutput^.Isqld <= 0 )
  then  exit;
  GetMem( pLength, sizeof( TFldInfos ) +
    ( pOutput^.Isqld - 1 ) * sizeof( TFldInfo ) );
  TFldInfos( pLength^ ).iLongDataCount := 0;
  for i := 0 to pOutput^.Isqld - 1
  do  with TFldInfos( pLength^ ).arInfos[ i ], pOutput^.arSqlVar[ i ]
      do  begin
          iGenueType := GetGenueType( @pOutput^.arSqlVar[ i ] )
            and SQL_NULL_MASK;
          if ( iGenueType = DT_LONGVARCHAR ) or ( iGenueType = DT_LONGBINARY )
          then  begin
                pFldPtr := pSqlData;
                Inc( TFldInfos( pLength^ ).iLongDataCount );
                end
          else  pFldPtr := nil;
          if ( iGenueType = DT_FIXCHAR ) or ( iGenueType = DT_STRING )
          then  iFldLen := iSqlLen
          else  iFldLen := -1;
          end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.RestoreLength;
var
  i : integer;
begin
  if ( pLength = nil ) or ( pOutput = nil ) then exit;
  for i := 0 to pOutput^.iSqld - 1
  do  with TFldInfos( pLength^ ).arInfos[ i ]
      do  if iFldLen >= 0 then pOutput^.arSqlVar[ i ].iSqlLen := iFldLen;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.SetSaDbBase;
begin
  if oVal = oSaDbBase then exit;
  PrepStatm( '', false, 0, nil );
  oSaDbBase := oVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDsBase.Fetch;

  procedure LockFetchLongData;
  var
    i : integer;
  begin
    if ( pOutput = nil )
      or ( pLength = nil )
        or ( TFldInfos( pLength^ ).iLongDataCount = 0 )
    then  exit;
    for i := 0 to pOutput^.iSqld - 1
    do  with TFldInfos( pLength^ ).arInfos[ i ]
        do  if pFldPtr <> nil then pOutput^.arSqlVar[ i ].pSqlData := nil;
  end;

  procedure UnLockFetchLongData;
  var
    i : integer;
  begin
    if ( pOutput = nil )
      or ( pLength = nil )
        or ( TFldInfos( pLength^ ).iLongDataCount = 0 )
    then  exit;
    for i := 0 to pOutput^.iSqld - 1
    do  with TFldInfos( pLength^ ).arInfos[ i ]
        do  if pFldPtr <> nil then pOutput^.arSqlVar[ i ].pSqlData := pFldPtr;
  end;

begin
  if ( oSaDbBase = nil ) or not isCursOpen
  then  result := frNotFound
  else  if isFetch
        then  begin
              LockFetchLongData;
              try
                result := oSaDbBase.SaFetch( PChar( sCursName ), pOutput,
                            isAbs, iAbsRel );
              finally
                UnLockFetchLongData;
              end;
              end
        else  result := oSaDbBase.SaFetch( PChar( sCursName ), nil,
          isAbs, iAbsRel );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDsBase.GetIsPrepared;
begin
  result := iStatm <> 0;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDsBase.Resume;
begin
  result := false;
  if not isCursOpen or ( oSaDbBase = nil ) then exit;
  result := oSaDbBase.SaResume( PChar( sCursName ) );
  if result then iRecordCount := oSaDbBase.GetSQLCount;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaDsBase.PutRec;
begin
  if not isCursOpen or ( oSaDbBase = nil ) then exit;
  iRowAffected := 0;
  try
    oSaDbBase.SaChange( PChar( sCursName ), SA_CURSCHANGE_INSERT, pOutput,
      pOutput );
    iRowAffected := oSaDbBase.GetSQLCount;
    oSaDbBase.TransactAutoCommit;
  except
    oSaDbBase.TransactAutoRollBack;
    raise;
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaDsBase.AllocateFldDescs;
var
  i : integer;
  p : TPSqlVarRec;
  pDesc : TPFldDesc;
  iGenueSqlType : integer;
  iSqlType : integer;
  iOffs : integer;
  iTmp : integer;
begin
  result := 0;
  if ( pOutput = nil ) or ( pOutput^.iSqld <= 0 ) then exit;
  for i := 0 to pOutput^.iSqld - 1
  do  if ( pOutput^.arSqlVar[ i ].pSqlInd <> nil )
            and ( ( pOutput^.arSqlVAR[ i ].pSqlInd^ and DT_PROCEDURE_OUT ) <> 0)
      then  exit;
  result := pOutput^.iSqld;
  GetMem    ( pFldDescs, sizeof( TFLDDesc ) * result );
  ZeroMemory( pFldDescs, sizeof( TFLDDesc ) * result );
  iOffs := 0;
  for i := 0 to result - 1
  do  begin
      p                     := @pOutput^.arSqlVar[ i ];
      pDesc                 := @pFldDescs^[ i ];
      pDesc^.iAddData       := sizeof( TFldDescAddData );
      GetMem( pDesc^.pAddData, pDesc^.iAddData );
      pDesc^.iFldNum        := i;
      pDesc^.sFldName       := StrPas( @p^.rcSqlName.sData );
      iGenueSqlType := GetGenueType( p );
      iSqlType := p^.iSqlType;
      TFldDescAddData( pDesc^.pAddData^ ).iSQLType := iSqlType;
      TFldDescAddData( pDesc^.pAddData^ ).iSQLLen  := p^.iSqlLen;
      pDesc^.iOffset        := iOffs;
      pDesc^.isAllowNulls   := ( p^.iSqlType and ( not SQL_NULL_MASK ) ) <> 0;

      iTmp                  := GetBufSize( pOutput, i );
      if iTmp < 0
      then  begin
            pDesc^.isVLRec := TRUE;
            pDesc^.iBufLen := -iTmp;
            end
      else  pDesc^.iBufLen := iTmp;
      pDesc^.iFldSize      := 0;
      pDesc^.iFldType      := ftUnKnown;
      Inc( iOffs, pDesc^.iBufLen );

      case ( iGenueSQLType and SQL_NULL_MASK ) of
        DT_FIXCHAR
          : begin
            if ( iSqlType and SQL_NULL_MASK ) = DT_STRING
            then  begin
                  pDesc^.iFldSize := p^.iSqlLen - 1;
                  if p^.iSqlLen > 255
                  then  pDesc^.iFldType := ftMemo
                  else  pDesc^.iFldType := ftString;
                  end
            else  begin
                  pDesc^.iFldSize := p^.iSqlLen;
                  if p^.iSqlLen > 255
                  then  pDesc^.iFldType := ftMemo
                  else  pDesc^.iFldType := ftString;
                  end;
            end;
        DT_VARCHAR
          : begin
            pDesc^.iFldSize := p^.iSqlLen - 3;
            if pDesc^.iFldSize > 255
            then  pDesc^.iFldType := ftBlob
            else  pDesc^.iFldType := ftString;
            end;
        DT_BINARY
          : begin
            pDesc^.iFldSize := p^.iSqlLen - 3;
            if pDesc^.iFldSize > 255
            then  pDesc^.iFldType := ftBlob
            else  pDesc^.iFldType := ftVarBytes;
            end;
        DT_SMALLINT, DT_TINYINT
          : pDesc^.iFldType := ftSmallInt;
        DT_INT
          : pDesc^.iFldType := ftInteger;
        DT_BIGINT, DT_UNSBIGINT
          : pDesc^.iFldType := ftLargeInt;
        DT_UNSINT
          : pDesc^.iFldType := ftInteger;
        DT_UNSSMALLINT
          : pDesc^.iFldType := ftWord;
        DT_BIT
          : pDesc^.iFldType := ftBoolean;
        DT_FLOAT, DT_DOUBLE
          : pDesc^.iFldType := ftFloat;
        DT_DATE
          : pDesc^.iFldType := ftDate;
        DT_TIME
          : pDesc^.iFldType := ftTime;
        DT_TIMESTAMP
          : pDesc^.iFldType := ftDateTime;
        DT_DECIMAL
          : begin
            if ( HIBYTE( p^.iSqlLen ) <= 4 ) and ( LOBYTE( p^.iSqlLen ) <= 19 )
            then  begin
                  pDesc^.iFldType := ftBCD;
                  pDesc^.iFldSize := HIBYTE( p^.iSqlLen );
                  pDesc^.iFldPrec := LOBYTE( p^.iSqlLen );
                  end
            else  begin
{$ifdef D_67}
                  pDesc^.iFldType := ftFmtBCD;
                  pDesc^.iFldPrec := LOBYTE( p^.iSQLLen );
{$else}
                  pDesc^.iFldType := ftFloat;
{$endif}
                  end;
            end;
        DT_LONGBINARY
          : begin
            pDesc^.iFldType := ftBlob;
            pDesc^.iFldSize := MAXWORD;
            end;
        DT_LONGVARCHAR
          : begin
            pDesc^.iFldType := ftMemo;
            pDesc^.iFldSize := MAXWORD;
            end;
      else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
           [ 'SQLType=' + IntToStr( ( iGenueSQLType and SQL_NULL_MASK ) ) ] ) );
      end;
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

end.


