{* SaVcl 2.5.  SaUtils.pas. Serge Gavrilov ( C ) 2003 - 2006 **************** *}

unit SaUtils;

{$i SaInc.inc}

interface

uses
  Windows, Messages, Classes, SysUtils, Db,
{$ifdef D_67}
  Variants, FmtBcd,
{$else}
  SaAbsolute,
{$endif}  
  SaCommon, SaConsts;

type

{* ************************************************************************** *}
  TSaExceptionType = (
      eSyBase,
      eUser,
      eInternal
    );
  SaException = class( EDatabaseError )
  private
    iExcType : TSaExceptionType;
    iSQLCode : integer;
    sDebugMsg : string;
  public
    property ExcType : TSaExceptionType read iExcType;
    property SQLCode : integer read iSQLCode;
    property DebugMsg : string read sDebugMsg;
    constructor Create(
        iExcType : TSaExceptionType;
        iSQLCode : integer;
        sMsg : string );
  end;

{* ************************************************************************** *}

  THWNDConnectOnRecv = procedure(
      oSender : TObject;
      pData : pointer;
      iLen : integer
    ) of object;
  THWNDConnect = class( TComponent )
  private
    isActive : boolean;
    isServer : boolean;
    iFileMap : THandle;
    sFileMap : string;
    pFileMap : pointer;
    iMaxSize : integer;
    iServerHWND : HWND;
    evOnRecv : THWNDConnectOnRecv;
    procedure SetServer( isVal : boolean );
    procedure SetFileMap( sVal : string );
    procedure SetMaxSize( iVal : integer );
    procedure SetActive( isVal : boolean );
    function WndProc(
        iMsg : integer;
        wParam : integer;
        lParam : integer;
        var iRes : integer
      ) : boolean;
  public
    function Send ( pData : pointer; iLen : integer ) : integer;
    constructor Create( AOwner : TComponent ); override;
  published
    property Server : boolean read isServer write SetServer;
    property FileMap : string read sFileMap write SetFileMap;
    property MaxSize : integer read iMaxSize write SetMaxSize;
    property Active : boolean read isActive write SetActive;
    property OnServerRecv : THWNDConnectOnRecv read evOnRecv write evOnRecv;
  end;

{* ************************************************************************** *}
  TSaParamType = ( ptInput, ptOutput, ptInputOutput, ptResult );
  TSaParamOwn = (
      poUser,
      poInternal
    );
  TSaParam = class
  private
    sName : string;
    vValue : variant;
    iOwner : TSaParamOwn;
    iType : TSaParamType;
    pDesc : TPFldDesc;
    pData : pointer;
    procedure SetValue( vValue : variant );
  public
    procedure SetAsBuffer(
        pDesc : TPFldDesc;
        pData : pointer
      );
    property Value : variant read vValue write SetValue;
    property ParamName : string read sName;
    property Owner : TSaParamOwn read iOwner;
    property ParamType : TSaParamType read iType write iType;
    property Desc : TPFldDesc read pDesc;
    property Data : pointer read pData;
    constructor Create(
        sName : string;
        iOwner : TSaParamOwn );
    end;

{* ************************************************************************** *}
  TSaOrderItem = class
  private
    iOpts : TSaOrderFldOpts;
    sFieldName : string;
    sNullSubs : string;
    iFldNum : integer;
    property NullSubs : string read sNullSubs;
  public
    property FieldName : string read sFieldName;
    property Opts : TSaOrderFldOpts read iOpts;
    property FldNum : integer read iFldNum;
  end;

  TSaNullSubs = class;
  TSaOrderType = ( otNone, otAsc, otDesc );
  TSaOrder = class
  private
    oItems : TList;
    function GetItem( iIndex : integer ) : TSaOrderItem;
    function GetIndexByName( sFldName : string ) : integer;
    function GetItemCount : integer;
  public
    constructor Create;
    destructor Destroy; override;
    function SetFldNum( sFldName : string; iNum : integer ) : boolean;
    function GetNullSubsByName( sFldName : string;
      oNullSubs : TSaNullSubs ) : string;
    function GetNullSubsByIndex( iIndex : integer;
      oNullSubs : TSaNullSubs ) : string;
    function GetFieldSort( sFldName : string ) : TSaOrderType;
    procedure Clear;
    procedure StringToSaOrder( s : string;
        isKey : boolean
      );
    property ItemCount : integer read GetItemCount;
    property Items[ iIndex : integer ] : TSaOrderItem read GetItem;
  end;

{* ************************************************************************** *}

  TSaNullSubs = class
  private
    oItems : TStrings;
    procedure Clear;
  public
    constructor Create;
    destructor Destroy; override;
    function GetNullSubs( sFieldName : string ) : string;
    procedure AssignText( sText : string );
  end;

{* ************************************************************************** *}

  TSaLocateRec = record
    iOffset : integer;
    iFldType : TFieldType;
    vValue : variant;
  end;
  TPSaLocateRec = ^TSaLocateRec;
  TSaLocateRecs = array[ 0..0 ] of TSaLocateRec;
  TPSaLocateRecs = ^TSaLocateRecs;
  TSaLocate = class
  private
    pRecs : TPSaLocateRecs;
    iRecs : integer;
    function BufferToVar(
        pBuf : pointer;
        iFldType : TFieldType;
        iStrLen : integer
      ) : variant;
  public
    function CompareRec(
        pBuf : pointer;
        isDBCaseSens : boolean;
        iLocateOpts : TSALocateOpts
      ) : boolean;
    procedure Add(
        iOffset : integer;
        iFldType : TFieldType;
        vVal : variant
      );
  public
    destructor Destroy; override;
  end;

{* ************************************************************************** *}

  TTranslDirection = ( tdToDB, tdToApp );
  TFieldGroup = ( fgUnknown,
    fgString, fgBytes, fgInt, fgFloat,
    fgBcd, fgBool, fgDate );
    
  procedure Translate(
      pData : pointer;
      iData : integer;
      iDir : TTranslDirection;
      iTransl : word );

  function FieldGroup( oField : TField ) : TFieldGroup; overload;
  function FieldGroup( iFieldType : TFieldType ) : TFieldGroup; overload;
  function GsvVarToStr( vValue : variant ) : string;
  function MinInt( i1, i2 : integer ) : integer;
  procedure RSAE(
      iExcType : TSaExceptionType;
      iSQLCode : integer;
      sExcMsg : string );


implementation

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor SaException.Create;
begin
  Self.iExcType := iExcType;
  Self.iSQLCode := iSQLCode;
  inherited Create( sMsg );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure  RSAE;
begin
  raise SaException.Create( iExcType, iSQLCode, sExcMsg );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure   Translate;
const
    CharByDBOrd : array[ byte ] of Char =
    (
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 0   - 9   }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 10  - 19  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 20  - 29  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 30  - 39  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 40  - 49  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 50  - 59  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 60  - 69  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 70  - 79  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 80  - 89  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 90  - 99  }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 100 - 109 }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 110 - 119 }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , '', '',   { 120 - 129 }
        '', '', '', '', '', '', '', '', '', '',   { 130 - 139 }
        '', '', '', '', '', '', '', '', '', '',   { 140 - 149 }
        '', '', '', '', '', '', '', '', '', '',   { 150 - 159 }
        '', '', '', '', '', '', '', '', '', '',   { 160 - 169 }
        '', '', '', '', '', '', #0 , #0 , #0 , #0 ,   { 170 - 179 }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 180 - 189 }
        '', '', #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 190 - 199 }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 200 - 209 }
        #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 210 - 219 }
        #0 , #0 , #0 , #0 , '', '', '', '', '', '',   { 220 - 229 }
        '', '', '', '', '', '', '', '', '', '',   { 230 - 239 }
        '', '', #0 , #0 , #0 , #0 , #0 , #0 , #0 , #0 ,   { 240 - 249 }
        #0 , #0 , #0 , #0 , #0 , #0                         { 250 - 255 }
    );
    OrdByIntOrd : array[ byte ] of byte =
    (
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 0   - 9   }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 10  - 19  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 20  - 29  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 30  - 39  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 40  - 49  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 50  - 59  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 60  - 69  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 70  - 79  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 80  - 89  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 90  - 99  }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 100 - 109 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 110 - 119 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 120 - 129 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 130 - 139 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 140 - 149 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 150 - 159 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 240, 0  ,   { 160 - 169 }
        0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  , 0  ,   { 170 - 179 }
        0  , 0  , 0  , 0  , 241, 0  , 0  , 0  , 0  , 0  ,   { 180 - 189 }
        0  , 0  , 128, 129, 130, 131, 132, 133, 134, 135,   { 190 - 199 }
        136, 137, 138, 139, 140, 141, 142, 143, 144, 145,   { 200 - 209 }
        146, 147, 148, 149, 150, 151, 152, 153, 154, 155,   { 210 - 219 }
        156, 157, 158, 159, 160, 161, 162, 163, 164, 165,   { 220 - 229 }
        166, 167, 168, 169, 170, 171, 172, 173, 174, 175,   { 230 - 239 }
        224, 225, 226, 227, 228, 229, 230, 231, 232, 233,   { 240 - 249 }
        234, 235, 236, 237, 238, 239                        { 250 - 259 }
    );
var
  i : integer;
  b : byte;
begin
  if iTransl <> 1 then exit;
  if iDir = tdToApp
  then  begin
        for i := 0 to iData - 1
        do  begin
            b := Ord( CharByDBOrd[ TByteArray( pData^ )[ i ] ] );
            if b <> 0 then TByteArray( pData^ )[ i ] := b;
            end;
        end
  else  begin
        for i := 0 to iData - 1
        do  begin
            b := OrdByIntOrd[ TByteArray( pData^ )[ i ] ];
            if b <> 0 then TByteArray( pData^ )[ i ] := b;
            end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function MinInt;
begin
  if i1 < i2
  then  result := i1
  else  result := i2;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor THWNDConnect.Create;
begin
  inherited;
  sFileMap := 'HWNDConnect';
  MaxSize  := 1024;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure THWNDConnect.SetServer;
begin
  if isVal = isServer then exit;
  if isActive
  then  begin
        Active := false;
        isServer := isVal;
        Active := true;
        end
  else  isServer := isVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure THWNDConnect.SetFileMap;
begin
  if sVal = sFileMap then exit;
  if isActive
  then  begin
        Active := false;
        sFileMap := sVal;
        Active := true;
        end
  else  sFileMap := sVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure THWNDConnect.SetMaxSize;
begin
  if ( iVal = iMaxSize ) or ( not isServer ) then exit;
  if isActive
  then  begin
        Active := false;
        iMaxSize := iVal;
        Active := true;
        end
  else  iMaxSize := iVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function I_WndProc( iWnd : HWND; iMsg : UINT; iwParam : WPARAM; ilParam : LPARAM
  ) : LRESULT;  stdcall;
var
  oConnect : THWNDConnect;
begin
  oConnect := pointer( GetWindowLong( iWnd, GWL_USERDATA ) );
  if ( oConnect = nil )
    or not oConnect.WndProc( iMsg, iwParam, ilParam, result )
  then  result := CallWindowProc( @DefWindowProc, iWnd, iMsg, iwParam,ilParam );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

var
  _iHWNDConnectWndClassAtom : ATOM = 0;
const
  I_S_WNDCLASSNAME = 'HWNDConnectFileMapWndClass';
  I_S_WNDNAME = 'HWNDConnectFileMapWnd';
  I_I_SendMSG = WM_USER + 1;
type
  TFileMapHeader = packed record
    iServerHWND : HWND;
    iMaxSize    : integer;
  end;
  TPFileMapHeader = ^TFileMapHeader;
  TMapData = packed record
    rHeader : TFileMapHeader;
    iData   : byte;
  end;
  TPMapData = ^TMapData;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function THWNDConnect.WndProc;
begin
  if iMsg = I_I_SendMSG
  then  begin
        if Assigned( evOnRecv )
        then  evOnRecv( Self, @( TPMapData( pFileMap )^.iData ), lParam );
        result := true;
        end
  else  result := false;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure THWNDConnect.SetActive;

  procedure FreeRes;
  begin
    if pFileMap <> nil
    then  begin
          UnmapViewOfFile( pFileMap );
          pFileMap := nil;
          end;
    if iFileMap <> 0
    then  begin
          CloseHandle( iFileMap );
          iFileMap := 0;
          end;
    if isServer and ( iServerHWND <> 0 )
    then  begin
          DestroyWindow( iServerHWND );
          iServerHWND := 0;
          end;
  end;
var
  rCl : TWndClass;
  rHeader : TFileMapHeader;
begin
  if ( isVal = isActive ) then exit;
  if isServer
  then  begin
        if isVal
        then  begin
              iFileMap := CreateFileMapping( $FFFFFFFF, nil, PAGE_READWRITE,
                0, iMaxSize + sizeof( rHeader ), PChar( sFileMap ) );
              if ( iFileMap = 0 )
              then  begin
                    FreeRes;
                    exit;
                    end;
              pFileMap := MapViewOfFile( iFileMap, FILE_MAP_ALL_ACCESS,
                                         0, 0, 0 );
              if pFileMap = nil
              then  begin
                    FreeRes;
                    exit;
                    end;
              if _iHWNDConnectWndClassAtom = 0
              then  begin
                    ZeroMemory( @rCl, sizeof( rCl ) );
                    rCl.lpfnWndProc := @I_WndProc;
                    rCl.hInstance := hInstance;
                    rCl.hCursor := LoadCursor( 0, IDC_ARROW );
                    rCl.hbrBackground := GetStockObject( WHITE_BRUSH );
                    rCl.lpszClassName := PChar( I_S_WNDCLASSNAME );
                    _iHWNDConnectWndClassAtom := Windows.RegisterClass( rCl );
                    if _iHWNDConnectWndClassAtom = 0
                    then  begin
                          FreeRes;
                          exit;
                          end;
                    end;
              iServerHWND := CreateWindow( PChar( I_S_WNDCLASSNAME ),
                                           PChar( I_S_WNDNAME ),
                                           WS_OVERLAPPED,
                                           0, 0, 0, 0,
                                           0, 0, hInstance, Self );
              if iServerHWND = 0
              then  begin
                    FreeRes;
                    exit;
                    end;
              SetWindowLong( iServerHWND, GWL_USERDATA, integer( Self ) );
              rHeader.iServerHWND := iServerHWND;
              rHeader.iMaxSize    := iMaxSize;
              TPFileMapHeader( pFileMap )^ := rHeader;
              end
        else  FreeRes;
        isActive := isVal;
        end
  else  begin
        if isVal
        then  begin
              iFileMap := OpenFileMapping( FILE_MAP_ALL_ACCESS, false,
                PChar( sFileMap ) );
              if ( iFileMap = 0 )
              then  begin
                    FreeRes;
                    exit;
                    end;
              pFileMap := MapViewOfFile( iFileMap, FILE_MAP_ALL_ACCESS,
                                         0, 0, 0 );
              if pFileMap = nil
              then  begin
                    FreeRes;
                    exit;
                    end;
              rHeader := TPFileMapHeader( pFileMap )^;
              iServerHWND := rHeader.iServerHWND;
              iMaxSize    := rHeader.iMaxSize;
              end
        else  FreeRes;
        isActive := isVal;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function THWNDConnect.Send;
begin
  if not Active or isServer
  then  begin
        result := -1;
        exit;
        end;
  if iLen > MaxSize
  then  iLen := iMaxSize;
  if iLen < 0
  then  iLen := 0;
  if ( pData <> nil ) and ( iLen > 0 )
  then  Move( pData^, TPMapData( pFileMap )^.iData, iLen );
  if not IsWindow( iServerHWND )
  then  begin
        Active := false;
        result := -1;
        exit;
        end;
  SendMessage( iServerHWND, I_I_SendMSG, 0, iLen );
  result := iLen;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor TSaParam.Create;
begin
  Self.sName := sName;
  Self.iOwner := iOwner;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaParam.SetValue;
begin
  pDesc := nil;
  pData := nil;
  Self.vValue := vValue;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaParam.SetAsBuffer;
begin
  Value := Null;
  Self.pDesc := pDesc;
  Self.pData := pData;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function GsvVarToStr( vValue : variant ) : string;
begin
  if VarIsNull( vValue )
  then  result := ''
  else  begin
        case VarType( vValue ) of
          varString : result := StrPas( TVarData( vValue ).VString );
        else
{$ifdef D_67}
              if VarIsFmtBcd( vValue )
              then  result := BCDToStr( VarToBcd( vValue ) )
              else
{$endif}
                    result := VarToStr( vValue );
        end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function FieldGroup( oField : TField ) : TFieldGroup;
begin
  if oField = nil
  then  result := fgUnknown
  else  result := FieldGroup( oField.DataType );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function FieldGroup( iFieldType : TFieldType ) : TFieldGroup;
begin
  if ( iFieldType in [ ftString, ftMemo ] )
  then  begin
        result := fgString;
        exit;
        end;
  if ( iFieldType in [ ftSmallint, ftInteger, ftWord, ftLargeInt ] )
  then  begin
        result := fgInt;
        exit;
        end;
  if ( iFieldType in [ ftDate, ftTime, ftDateTime ] )
  then  begin
        result := fgDate;
        exit;
        end;
  if ( iFieldType in [ ftBCD {$ifdef D_67}, ftFMTBcd {$endif} ] )
  then  begin
        result := fgBcd;
        exit;
        end;
  if ( iFieldType in [ ftBoolean ] )
  then  begin
        result := fgBool;
        exit;
        end;
  if ( iFieldType in [ ftFloat ] )
  then  begin
        result := fgFloat;
        exit;
        end;
  if ( iFieldType in [ ftVarBytes, ftBlob ] )
  then  begin
        result := fgBytes;
        exit;
        end;
  result := fgUnknown;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{* ************************************************************************** *}

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor TSaOrder.Create;
begin
  oItems := TList.Create;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaOrder.Destroy;
begin
  Clear;
  oItems.Free;
  inherited;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetNullSubsByName;
var
  i : integer;
begin
  i := GetIndexByName( sFldName );
  if i >= 0
  then  result := Items[ i ].NullSubs
  else  result := '';
  if ( result = '' ) and ( oNullSubs <> nil )
  then result := oNullSubs.GetNullSubs( sFldName );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetNullSubsByIndex;
begin
  result := '';
  if ( iIndex < 0 ) or ( iIndex >= ItemCount ) then exit;
  result := Items[ iIndex ].NullSubs;
  if ( result = '' ) and ( oNullSubs <> nil )
  then result := oNullSubs.GetNullSubs( Items[ iIndex ].FieldName );
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetItem;
begin
  if ( iIndex >= oItems.Count ) or ( iIndex < 0 )
  then  result := nil
  else  result := oItems[ iIndex ];
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetItemCount;
begin
  result := oItems.Count;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.SetFldNum;
var
  i : integer;
begin
  i := GetIndexByName( sFldName );
  result := i >= 0;
  if result then  Items[ i ].iFldNum := iNum;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetFieldSort;
const
  AI_SORT : array[ boolean ] of TSaOrderType = ( otAsc, otDesc );
var
  i : integer;
begin
  i := GetIndexByName( sFldName );
  if i >= 0
  then  result := AI_SORT[ ofoDesc in Items[ i ].iOpts ]
  else  result := otNone;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaOrder.GetIndexByName;
var
  i : integer;
begin
  result := -1;
  for i := 0 to ItemCount - 1
  do  if CompareText( Items[ i ].FieldName, sFldName ) = 0
      then  begin
            result := i;
            break;
            end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaOrder.StringToSaOrder;

    function ExtSepStr(
        const sSource : string;
        const cSep : char;
        var iPos : Integer;
        isSepReq : boolean
      ) : string;
    const
      AS_BEGS : array [ 0..2 ] of Char =( '"', '''', '(' );
      AS_ENDS : array [ 0..2 ] of Char =( '"', '''', ')' );
    var
      i, iL   : integer;
      sLocks  : string;
      function GetIndex( c : Char; ar : array of Char ) : integer;
      var
        i : integer;
      begin
        result := Low( ar ) - 1;
        for i := Low( ar ) to High( ar )
        do  if ar[ i ] = c
            then  begin
                  result := i;
                  exit;
                  end;
      end;
      function GetLocks( c : Char ) : boolean;
      var
        i : integer;
      begin
        result := sLocks <> '';
        if result
        then  begin
              i := GetIndex( c, AS_ENDS );
              if i >= Low( AS_ENDS )
              then  begin
                    if sLocks[ Length( sLocks ) ] = c
                    then  begin
                          System.Delete( sLocks, Length( sLocks ), 1 );
                          result := sLocks <> '';
                          exit;
                          end;
                    end;
              end;
        i := GetIndex( c, AS_BEGS );
        if i >= Low( AS_BEGS )
        then  begin
              sLocks := sLocks + c;
              result := true;
              end;
      end;
    begin
      sLocks := '';
      i := iPos; iL := Length( sSource );
      while ( i <= iL )
      do  begin
          if not GetLocks( sSource[ i ] )
          then  if sSource[ i ] = cSep
                then    break;
          Inc( i );
          end;
      result := Trim( Copy( sSource, iPos, i - iPos ) );
      if isSepReq
      then  begin
            if ( i > iL ) or ( sSource[ i ] <> cSep )
            then    result := '';
            end
      else  if ( i <= iL ) and ( sSource[ i ] = cSep )
            then Inc( i );
      iPos := i;
    end;
var
  iFldPos : integer;
  iTmp : integer;
  sFld : string;
  sFldName : string;
  sSubs : string;
  isDesc : boolean;
  oNewItem : TSaOrderItem;
begin
  s := Trim( UpperCase( s ) );
  if s <> ''
  then  begin
        iFldPos := 1;
        sFld := Trim( ExtSepStr( s, ',', iFldPos, false ) );
        while sFld <> ''
        do  begin
            isDesc := false;
            iTmp := 1;
            sFldName := ExtSepStr( sFld, ' ', iTmp, false );
            if sFldName <> ''
            then  begin
                  System.Delete( sFld, 1, Length( sFldName ) );
                  sFld := Trim( sFld );
                  iTmp := 1;
                  sSubs := ExtSepStr( sFld, ' ', iTmp, false );
                  if sSubs <> ''
                  then  begin
                        System.Delete( sFld, 1, Length( sSubs ) );
                        if ( CompareText( Trim( sSubs ), 'DESC' ) = 0 )
                        then  begin
                              isDesc  := true;
                              sSubs   := sFld;
                              end;
                        end;
                  iTmp := GetIndexByName( sFldName );
                  if iTmp < 0
                  then  begin
                        oNewItem := TSaOrderItem.Create;
                        oNewItem.sFieldName := Trim( sFldName );
                        oNewItem.sNullSubs := Trim( sSubs );
                        if isDesc
                        then oNewItem.iOpts := oNewItem.iOpts + [ ofoDesc ];
                        iTmp := oItems.Add( oNewItem );
                        end;
                  if isKey
                  then  with Items[ iTmp ]
                        do  iOpts := iOpts + [ ofoKey ];
                  end;
            sFld := Trim( ExtSepStr( s, ',', iFldPos, false ) );
            end;
        end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaOrder.Clear;
var
  i : integer;
begin
  for i := 0 to oItems.Count - 1
  do  TObject( oItems.List^[ i ] ).Free;
  oItems.Count := 0;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{ **************************************************************************** }

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
constructor TSaNullSubs.Create;
begin
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaNullSubs.Destroy;
begin
  Clear;
  oItems.Free;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaNullSubs.AssignText;
  function Extract( sText : string; var sResult : string;
    var iPos : integer ) : boolean;
  var
    i : integer;
    iOnQuaters : integer;
  begin
    sResult := '';
    iOnQuaters := 0;
    for i := iPos to Length( sText )
    do  begin
        if sText[ i ] = '(' then  Inc( iOnQuaters );
        if sText[ i ] = ')' then  Dec( iOnQuaters );
        if iOnQuaters <> 0 then Continue;
        if ( sText[ i ] = ',' ) or ( i = Length( sText ) )
        then  begin
              if sText[ i ] = ','
              then  sResult := Copy( sText, iPos, i - iPos )
              else  sResult := Copy( sText, iPos, i - iPos + 1 );
              iPos := i + 1;
              break;
              end;
        end;
    sResult := Trim( sResult );
    result := sResult <> '';
  end;
var
  sPair : string;
  iPos1 : integer;
  iPos2 : integer;
  sFieldName : string;
  sNullSubs : string;
begin
  Clear;
  iPos1 := 1;
  while Extract( sText, sPair, iPos1 )
  do  begin
      iPos2 := Pos( ' ', sPair );
      if iPos2 = 0 then continue;
      sFieldName := Trim( Copy( sPair, 1, iPos2 - 1 ) );
      sNullSubs := Trim( Copy( sPair, iPos2 + 1, Length( sPair ) - iPos2 ) );
      if ( sFieldName <> '' ) and ( sNullSubs <> '' )
      then  begin
            if oItems = nil then oItems := TStringList.Create;
            oItems.Add( sFieldName );
            oItems.Add( sNullSubs );
            end;
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaNullSubs.Clear;
begin
  if oItems = nil then exit;
  oItems.Clear;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaNullSubs.GetNullSubs;
var
  iIndex : integer;
begin
  result := '';
  if oItems = nil then exit;
  iIndex := oItems.IndexOf( sFieldName );
  if ( iIndex < 0 ) or ( Odd( iIndex ) ) then exit;
  result := oItems[ iIndex + 1 ];
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{ **************************************************************************** }

{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
destructor TSaLocate.Destroy;
var
  i : integer;
begin
  if pRecs <> nil
  then  begin
        for i := 0 to iRecs - 1
        do  begin
            pRecs^[ i ].vValue    := NULL;
            end;
        FreeMem( pRecs, sizeof( TSaLocateRec ) * iRecs );
        iRecs := 0;
        pRecs := nil;
        end;
  inherited Destroy;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaLocate.BufferToVar;
var
  isVLR : boolean;
  sTmp : string;
begin
  result := Null;
  if ( pBuf = nil ) or ( ( byte( pBuf^ ) and SA_SERV_ISNULL ) <> 0 )
  then  exit;
  isVLR := ( ( byte( pBuf^ ) and SA_SERV_VLREC ) <> 0 );
  pBuf := @( TByteArray( pBuf^ )[ 1 ] );
  case iFldType of
    ftSmallInt  : result := smallint ( pBuf^ );
    ftInteger : result := integer ( pBuf^ );
    ftTime, ftDate, ftDateTime : result := TDateTime( pBuf^ );
    ftFloat : result := double( pBuf^ );
    ftLargeInt : result :=
{$ifdef D_67}
      Int64( pBuf^ );
{$else}
      Int64ToVar( Int64( pBuf^ ) );
{$endif}
    ftBoolean : result := VarAsType( shortint( pBuf^ ) <> 0, varBoolean );
    ftWord : result := Word( pBuf^ );
    ftBCD {$ifdef D_67}, ftFmtBCD {$endif} :
      result :=
{$ifdef D_67}
        VarFMTBcdCreate( TBcd( pBuf^ ) );
{$else}
  {$ifdef D_5}
        BcdToVar( TBcd( pBuf^ ) );
  {$else}
        currency( pBuf^ );
  {$endif}
{$endif}
    ftString, ftVarBytes, ftMemo, ftBlob :
      begin
      if isVLR
      then  begin
            if ( iStrLen < 0 ) or ( TVLRec( pBuf^ ).iData < iStrLen )
            then  iStrLen := TVLRec( pBuf^ ).iData;
            SetLength( sTmp, iStrLen );
            Move( TVLRec( pBuf^ ).pData^, sTmp[ 1 ], iStrLen );
            end
      else  begin
            if ( iStrLen < 0 ) or ( byte( pBuf^ ) < iStrLen )
            then  iStrLen := byte( pBuf^ );
            SetLength( sTmp, iStrLen );
            Move( TByteArray( pBuf^ )[ 1 ], sTmp[ 1 ], iStrLen );
            end;
      result := sTmp;
      end;
    else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                                      [ FieldTypeNames[ iFldType ] ] ) );
  end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
function TSaLocate.CompareRec;
var
  i : integer;
{$ifdef D_67}
  Buffer : pointer;
{$endif}
  iRecLen : integer;
  pServByte : pointer;
  vTmp : variant;
  iFldGrp : TFieldGroup;
begin
  result := true;
  for i := 0 to iRecs - 1
  do  begin
      iFldGrp := FieldGroup( pRecs^[ i ].iFldType );
      pServByte := @TByteArray( pBuf^ )[ pRecs^[ i ].iOffset ];
      if ( byte( pServByte^ ) and SA_SERV_ISNULL ) <> 0
      then  begin
            if ( iFldGrp = fgString ) and ( loNullStrings in iLocateOpts )
            then  result := GsvVarToStr( pRecs^[ i ].vValue ) = ''
            else  result := pRecs^[ i ].vValue = NULL;
            end
      else  begin
            result := pRecs^[ i ].vValue <> NULL;
            if result
            then  begin
                  case iFldGrp of
{$ifndef D_67}
                    fgBcd,
{$endif}
                    fgInt, fgFloat, fgBool, fgDate :
                      begin
                      vTmp := BufferToVar( pServByte,
                                           pRecs^[ i ].iFldType, -1 );
                      result := vTmp = pRecs^[ i ].vValue;
                      end;
{$ifdef D_67}
                    fgBcd :
                      begin
                      Buffer := @TByteArray( pBuf^ )[ pRecs^[ i ].iOffset + 1 ];
                      if VarIsFMTBcd( pRecs^[ i ].vValue )
                      then  result := BcdCompare( TBcd( Buffer^ ),
                                      VarToBcd( pRecs^[ i ].vValue ) ) = 0
                      else  result := BcdToDouble( TBcd( Buffer^ ) ) =
                                      pRecs^[ i ].vValue;
                      end;                      
{$endif}
                    fgBytes :
                      begin
                      vTmp := BufferToVar( pServByte, pRecs^[ i ].iFldType, -1);
                      result := CompareStr( vTmp,
                        GsvVarToStr( pRecs^[ i ].vValue ) ) = 0;
                      end;
                    fgString :
                      begin
                      if loPartAny in iLocateOpts
                      then  begin
                            vTmp := BufferToVar( pServByte,
                              pRecs^[ i ].iFldType, -1 );
                            if ( not isDBCaseSens )
                            then  result := AnsiPos(
                              AnsiUpperCase( vTmp ),
                              AnsiUpperCase( GsvVarToStr( pRecs^[ i ].vValue ) )
                                    ) > 0
                            else  result := AnsiPos( string( vTmp ),
                                    GsvVarToStr( pRecs^[ i ].vValue ) ) > 0
                            end
                      else  begin
                            iRecLen :=
                              Length( GsvVarToStr( pRecs^[ i ].vValue ) );
                            if loPartLeft in iLocateOpts
                            then  vTmp := BufferToVar( pServByte,
                              pRecs^[ i ].iFldType, iRecLen )
                            else  vTmp := BufferToVar( pServByte,
                              pRecs^[ i ].iFldType, -1 );
                            if ( not isDBCaseSens )
                            then  result := CompareText( vTmp,
                                        GsvVarToStr( pRecs^[ i ].vValue ) ) = 0
                            else  result := CompareStr(  string( vTmp ),
                                        GsvVarToStr( pRecs^[ i ].vValue ) ) = 0;
                            end;
                      end;
                  else  RSAE( eInternal, 0, Format( SA_S_ERR_NOTSUPFLDTYPE,
                                 [ FieldTypeNames[ pRecs^[ i ].iFldType ] ] ) );
                  end;
                  end;
            end;
      if not result then exit;
      end;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}
procedure TSaLocate.Add;
var
  pRec : TPSaLocateRec;
begin
  Inc( iRecs );
  ReAllocMem( pRecs, sizeof( TSaLocateRec ) * iRecs );
  pRec := @pRecs^[ iRecs - 1 ];
  ZeroMemory( pRec, sizeof( TSaLocateRec ) );
  pRec^.iOffset := iOffset;
  pRec^.iFldType := iFldType;
  pRec^.vValue := vVal;
end;
{* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *}

{ **************************************************************************** }

end.


